/* packet-smb.c
 * Routines for smb packet dissection
 * Copyright 1999, Richard Sharpe <rsharpe@ns.aus.com>
 *
 * $Id: packet-smb.c,v 1.113 2001/09/11 05:31:45 guy Exp $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * Copied from packet-pop.c
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <time.h>
#include <string.h>
#include <glib.h>
#include <ctype.h>
#include "packet.h"
#include "smb.h"
#include "alignment.h"
#include "strutil.h"

#include "packet-smb-mailslot.h"
#include "packet-smb-pipe.h"

static int proto_smb = -1;

static int hf_smb_cmd = -1;
static int hf_smb_word_count = -1;
static int hf_smb_byte_count = -1;
static int hf_smb_buffer_format = -1;
static int hf_smb_file_name = -1;
static int hf_smb_file_name_len = -1;
static int hf_smb_old_file_name = -1;
static int hf_smb_search_attribute_read_only = -1;
static int hf_smb_search_attribute_hidden = -1;
static int hf_smb_search_attribute_system = -1;
static int hf_smb_search_attribute_volume = -1;
static int hf_smb_search_attribute_directory = -1;
static int hf_smb_search_attribute_archive = -1;
static int hf_smb_file_attr_read_only = -1;
static int hf_smb_file_attr_hidden = -1;
static int hf_smb_file_attr_system = -1;
static int hf_smb_file_attr_volume = -1;
static int hf_smb_file_attr_directory = -1;
static int hf_smb_file_attr_archive = -1;
static int hf_smb_file_attr_device = -1;
static int hf_smb_file_attr_normal = -1;
static int hf_smb_file_attr_temporary = -1;
static int hf_smb_file_attr_sparse = -1;
static int hf_smb_file_attr_reparse = -1;
static int hf_smb_file_attr_compressed = -1;
static int hf_smb_file_attr_offline = -1;
static int hf_smb_file_attr_not_content_indexed = -1;
static int hf_smb_file_attr_encrypted = -1;
static int hf_smb_file_attribute_read_only = -1;
static int hf_smb_file_attribute_hidden = -1;
static int hf_smb_file_attribute_system = -1;
static int hf_smb_file_attribute_volume = -1;
static int hf_smb_file_attribute_directory = -1;
static int hf_smb_file_attribute_archive = -1;
static int hf_smb_last_write_time = -1;
static int hf_smb_file_size = -1;
static int hf_smb_reserved = -1;
static int hf_smb_unknown = -1;
static int hf_smb_padding = -1;
static int hf_smb_dialect_name = -1;
static int hf_smb_dialect_index = -1;
static int hf_sm_mode = -1;
static int hf_sm_password = -1;
static int hf_sm_signatures = -1;
static int hf_sm_sig_required = -1;
static int hf_sm_mode16 = -1;
static int hf_sm_password16 = -1;
static int hf_smb_max_referral_level = -1;
static int hf_max_trans_buf_size = -1;
static int hf_max_buf_size = -1;
static int hf_max_mpx_count = -1;
static int hf_max_vcs_num = -1;
static int hf_vc_num = -1;
static int hf_rm_read = -1;
static int hf_rm_write = -1;
static int hf_session_key = -1;
static int hf_current_server_date = -1;
static int hf_current_server_smb_date = -1;
static int hf_current_server_smb_time = -1;
static int hf_server_timezone = -1;
static int hf_encryption_key_length = -1;
static int hf_encryption_key = -1;
static int hf_domain = -1;
static int hf_lanman = -1;
static int hf_os = -1;
static int hf_max_raw_buf_size = -1;
static int hf_server_cap_raw_mode = -1;
static int hf_server_cap_mpx_mode = -1;
static int hf_server_cap_unicode = -1;
static int hf_server_cap_large_files = -1;
static int hf_server_cap_nt_smbs = -1;
static int hf_server_cap_rpc_remote_apis = -1;
static int hf_server_cap_status32 = -1;
static int hf_server_cap_level_ii_oplocks = -1;
static int hf_server_cap_lock_and_read = -1;
static int hf_server_cap_nt_find = -1;
static int hf_server_cap_dfs = -1;
static int hf_server_cap_infolevel_passthru = -1;
static int hf_server_cap_large_readx = -1;
static int hf_server_cap_large_writex = -1;
static int hf_server_cap_unix = -1;
static int hf_server_cap_reserved = -1;
static int hf_server_cap_bulk_transfer = -1;
static int hf_server_cap_compressed_data = -1;
static int hf_server_cap_extended_security = -1;
static int hf_security_blob = -1;
static int hf_server_guid = -1;
static int hf_time_low = -1;
static int hf_time_high = -1;
static int hf_smb_dir_name = -1;
static int hf_smb_tid = -1;
static int hf_smb_fid = -1;
static int hf_smb_sid = -1;
static int hf_smb_count = -1;
static int hf_smb_maxcount = -1;
static int hf_smb_mincount = -1;
static int hf_smb_offset = -1;
static int hf_smb_resume = -1;
static int hf_smb_high_offset = -1;
static int hf_smb_remaining = -1;
static int hf_smb_change_time = -1;
static int hf_smb_access_time = -1;
static int hf_smb_access_dos_date = -1;
static int hf_smb_access_dos_time = -1;
static int hf_smb_create_time = -1;
static int hf_smb_create_dos_date = -1;
static int hf_smb_create_dos_time = -1;
static int hf_smb_last_write_date = -1;
static int hf_smb_last_write_dos_date = -1;
static int hf_smb_last_write_dos_time = -1;
static int hf_smb_file_data = -1;
static int hf_smb_data_len = -1;
static int hf_smb_data_offset = -1;
static int hf_smb_data_disp = -1;
static int hf_smb_seek_mode = -1;
static int hf_smb_timeout = -1;
static int hf_smb_dcm = -1;
static int hf_smb_units = -1;
static int hf_smb_bpu = -1;
static int hf_smb_blocksize = -1;
static int hf_smb_freeunits = -1;
static int hf_smb_andxoffset = -1;
static int hf_smb_password_len = -1;
static int hf_smb_password = -1;
static int hf_smb_account = -1;
static int hf_smb_connect_flags_dtid = -1;
static int hf_smb_path = -1;
static int hf_smb_service = -1;
static int hf_smb_setup_action_guest = -1;
static int hf_smb_connect_support_search = -1;
static int hf_smb_connect_support_in_dfs = -1;
static int hf_smb_fs = -1;
static int hf_smb_open_flags_add_info = -1;
static int hf_smb_open_flags_ex_oplock = -1;
static int hf_smb_open_flags_batch_oplock = -1;
static int hf_smb_open_flags_ealen = -1;
static int hf_smb_access_mode = -1;
static int hf_smb_access_sharing = -1;
static int hf_smb_access_locality = -1;
static int hf_smb_access_caching = -1;
static int hf_smb_access_writetru = -1;
static int hf_smb_open_function_open = -1;
static int hf_smb_open_function_create = -1;
static int hf_smb_end_of_file = -1;
static int hf_smb_end_of_search = -1;
static int hf_smb_alloc_size = -1;
static int hf_smb_file_type = -1;
static int hf_smb_open_action_open = -1;
static int hf_smb_open_action_lock = -1;
static int hf_smb_device_state = -1;
static int hf_smb_server_fid = -1;
static int hf_smb_move_flags_file = -1;
static int hf_smb_move_flags_dir = -1;
static int hf_smb_move_flags_verify = -1;
static int hf_smb_total_param_count = -1;
static int hf_smb_total_data_count = -1;
static int hf_smb_max_param_count = -1;
static int hf_smb_max_data_count = -1;
static int hf_smb_max_setup_count = -1;
static int hf_smb_transaction_flags_dtid = -1;
static int hf_smb_transaction_flags_owt = -1;
static int hf_smb_param_disp = -1;
static int hf_smb_param_count = -1;
static int hf_smb_param_offset = -1;
static int hf_smb_data_count = -1;
static int hf_smb_setup_count = -1;
static int hf_smb_search_count = -1;
static int hf_smb_trans_name = -1;
static int hf_smb_trans2_subcmd = -1;
static int hf_smb_qpi_loi = -1;
static int hf_smb_ff2_backup = -1;
static int hf_smb_ff2_continue = -1;
static int hf_smb_ff2_resume = -1;
static int hf_smb_ff2_close_eos = -1;
static int hf_smb_ff2_close = -1;
static int hf_smb_ff2_information_level = -1;
static int hf_smb_storage_type = -1;
static int hf_smb_search_pattern = -1;
static int hf_smb_qfsi_information_level = -1;
static int hf_smb_sfi_information_level = -1;
static int hf_smb_data_size = -1;
static int hf_smb_ea_size = -1;
static int hf_smb_list_length = -1;
static int hf_smb_number_of_links = -1;
static int hf_smb_delete_pending = -1;
static int hf_smb_is_directory = -1;
static int hf_smb_index_number = -1;
static int hf_smb_current_offset = -1;
static int hf_smb_t2_access_read_data = -1;
static int hf_smb_t2_access_write_data = -1;
static int hf_smb_t2_access_append_data = -1;
static int hf_smb_t2_access_read_ea = -1;
static int hf_smb_t2_access_write_ea = -1;
static int hf_smb_t2_access_execute = -1;
static int hf_smb_t2_access_read_attributes = -1;
static int hf_smb_t2_access_write_attributes = -1;
static int hf_smb_t2_access_delete = -1;
static int hf_smb_t2_access_read_control = -1;
static int hf_smb_t2_access_write_dac = -1;
static int hf_smb_t2_access_write_owner = -1;
static int hf_smb_t2_access_synchronize = -1;
static int hf_smb_t2_mode_write_through = -1;
static int hf_smb_t2_mode_sequential_only = -1;
static int hf_smb_t2_mode_sync_io_alert = -1;
static int hf_smb_t2_mode_sync_io_nonalert = -1;
static int hf_smb_t2_alignment = -1;
static int hf_smb_t2_next_entry_offset = -1;
static int hf_smb_t2_stream_name_length = -1;
static int hf_smb_t2_stream_size = -1;
static int hf_smb_t2_compressed_file_size = -1;
static int hf_smb_t2_compressed_format = -1;
static int hf_smb_t2_compressed_unit_shift = -1;
static int hf_smb_t2_compressed_chunk_shift = -1;
static int hf_smb_t2_compressed_cluster_shift = -1;
static int hf_smb_dfs_path_consumed = -1;
static int hf_smb_dfs_num_referrals = -1;
static int hf_smb_get_dfs_server_hold_storage = -1;
static int hf_smb_get_dfs_fielding = -1;
static int hf_smb_dfs_referral_version = -1;
static int hf_smb_dfs_referral_size = -1;
static int hf_smb_dfs_referral_server_type = -1;
static int hf_smb_dfs_referral_flags_strip = -1;
static int hf_smb_dfs_referral_node = -1;
static int hf_smb_dfs_referral_proximity = -1;
static int hf_smb_dfs_referral_ttl = -1;
static int hf_smb_dfs_referral_path = -1;
static int hf_smb_dfs_referral_alt_path = -1;
static int hf_smb_response_to = -1;
static int hf_smb_response_in = -1;
static int hf_smb_ea_error_offset = -1;
static int hf_smb_ea_length = -1;
static int hf_smb_last_name_offset = -1;
static int hf_smb_fs_id = -1;
static int hf_smb_sector_unit = -1;
static int hf_smb_fs_units = -1;
static int hf_smb_fs_sector = -1;
static int hf_smb_volume_serial_num = -1;
static int hf_smb_volume_label_len = -1;
static int hf_smb_volume_label = -1;
static int hf_smb_free_alloc_units = -1;
static int hf_smb_device_type = -1;
static int hf_smb_device_char_removable = -1;
static int hf_smb_device_char_read_only = -1;
static int hf_smb_device_char_floppy = -1;
static int hf_smb_device_char_write_once = -1;
static int hf_smb_device_char_remote = -1;
static int hf_smb_device_char_mounted = -1;
static int hf_smb_device_char_virtual = -1;
static int hf_smb_fs_attr_css = -1;
static int hf_smb_fs_attr_cpn = -1;
static int hf_smb_fs_attr_pacls = -1;
static int hf_smb_fs_attr_fc = -1;
static int hf_smb_fs_attr_vq = -1;
static int hf_smb_fs_attr_dim = -1;
static int hf_smb_fs_attr_vic = -1;
static int hf_smb_max_name_len = -1;
static int hf_smb_fs_name_len = -1;
static int hf_smb_fs_name = -1;
static int hf_smb_echo_count = -1;
static int hf_smb_echo_data = -1;
static int hf_smb_echo_seq_num = -1;

static gint ett_smb = -1;
static gint ett_smb_hdr = -1;
static gint ett_smb_fileattributes = -1;
static gint ett_smb_file_attributes = -1;
static gint ett_smb_capabilities = -1;
static gint ett_smb_aflags = -1;
static gint ett_smb_dialects = -1;
static gint ett_smb_dialect = -1;
static gint ett_smb_mode = -1;
static gint ett_smb_rawmode = -1;
static gint ett_smb_flags = -1;
static gint ett_smb_flags2 = -1;
static gint ett_smb_desiredaccess = -1;
static gint ett_smb_search = -1;
static gint ett_smb_file = -1;
static gint ett_smb_openfunction = -1;
static gint ett_smb_filetype = -1;
static gint ett_smb_writemode = -1;
static gint ett_smb_lock_type = -1;
static gint ett_smb_ssetupandxaction = -1;
static gint ett_smb_optionsup = -1;
static gint ett_smb_command = -1;
static gint ett_smb_time_date = -1;
static gint ett_smb_connect_flags = -1;
static gint ett_smb_setup_action = -1;
static gint ett_smb_connect_support_bits = -1;
static gint ett_smb_open_flags = -1;
static gint ett_smb_open_action = -1;
static gint ett_smb_move_flags = -1;
static gint ett_smb_transaction_flags = -1;
static gint ett_smb_transaction_params = -1;
static gint ett_smb_transaction_open_flags = -1;
static gint ett_smb_find_first2_flags = -1;
static gint ett_smb_transaction_data = -1;
static gint ett_smb_64bit_time = -1;
static gint ett_smb_get_dfs_flags = -1;
static gint ett_smb_dfs_referrals = -1;
static gint ett_smb_dfs_referral = -1;
static gint ett_smb_dfs_referral_flags = -1;
static gint ett_smb_device_characteristics = -1;
static gint ett_smb_fs_attributes = -1;

/* XXX temporary fix later */
static proto_tree *top_tree = NULL;


static const gchar *
get_unicode_or_ascii_string_tvb(tvbuff_t *tvb, int *offsetp, packet_info *pinfo, int *len, gboolean nopad, gboolean exactlen);

char *decode_smb_name(unsigned char);

static int
call_next_command(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree, guint8 cmd);

static const value_string alignment_vals[] = {
	{0,	"Byte alignment"},
	{1,	"Word (16bit) alignment"},
	{3,	"Long (32bit) alignment"},
	{7,	"8 byte boundary alignment"},
	{0x0f,	"16 byte boundary alignment"},
	{0x1f,	"32 byte boundary alignment"},
	{0x3f,	"64 byte boundary alignment"},
	{0x7f,	"128 byte boundary alignment"},
	{0xff,	"256 byte boundary alignment"},
	{0x1ff,	"512 byte boundary alignment"},
	{0, NULL}
};

static const value_string delete_pending_vals[] = {
	{0,	"Normal, no pending delete"},
	{1,	"This object has DELETE PENDING"},
};
static const value_string is_directory_vals[] = {
	{0,	"This is NOT a directory"},
	{1,	"This is a DIRECTORY"},
};
static const value_string seek_mode_vals[] = {
	{0,	"From Start Of File"},
	{1,	"From Current Position"},
	{2,	"From End Of File"},
	{0,	NULL}
};

static const value_string buffer_format_vals[] = {
	{1,     "Data Block"},
	{2,     "Dialect"},
	{3,     "Pathname"},
	{4,     "ASCII"},
	{5,     "Variable Block"},
	{0,     NULL}
};

static const value_string filetype_vals[] = {
	{ 0,		"Disk file or directory"},
	{ 1,		"Named pipe in byte mode"},
	{ 2,		"Named pipe in message mode"},
	{ 3,		"Spooled printer"},
	{0, NULL}
};


#define FILE_ATTRIBUTE_READ_ONLY       0x0001
#define FILE_ATTRIBUTE_HIDDEN          0x0002
#define FILE_ATTRIBUTE_SYSTEM          0x0004
#define FILE_ATTRIBUTE_VOLUME          0x0008
#define FILE_ATTRIBUTE_DIRECTORY       0x0010
#define FILE_ATTRIBUTE_ARCHIVE         0x0020
#define FILE_ATTRIBUTE_DEVICE          0x0040
#define FILE_ATTRIBUTE_NORMAL          0x0080
#define FILE_ATTRIBUTE_TEMPORARY       0x0100
#define FILE_ATTRIBUTE_SPARSE          0x0200
#define FILE_ATTRIBUTE_REPARSE         0x0400
#define FILE_ATTRIBUTE_COMPRESSED      0x0800
#define FILE_ATTRIBUTE_OFFLINE         0x1000
#define FILE_ATTRIBUTE_NOT_CONTENT_INDEXED         0x2000
#define FILE_ATTRIBUTE_ENCRYPTED       0x4000
static const true_false_string tfs_file_attribute_read_only = {
	"This file is READ ONLY",
	"This file is NOT read only",
};
static const true_false_string tfs_file_attribute_hidden = {
	"This is a HIDDEN file",
	"This is NOT a hidden file"
};
static const true_false_string tfs_file_attribute_system = {
	"This is a SYSTEM file",
	"This is NOT a system file"
};
static const true_false_string tfs_file_attribute_volume = {
	"This is a VOLUME",
	"This is NOT a volume"
};
static const true_false_string tfs_file_attribute_directory = {
	"This is a DIRECTORY",
	"This is NOT a directory"
};
static const true_false_string tfs_file_attribute_archive = {
	"This is an ARCHIVE file",
	"This is NOT an archive file"
};
static const true_false_string tfs_file_attribute_device = {
	"This is a DEVICE",
	"This is NOT a device"
};
static const true_false_string tfs_file_attribute_normal = {
	"This is a NORMAL file",
	"This is NOT a normal file"
};
static const true_false_string tfs_file_attribute_temporary = {
	"This is a TEMPORARY file",
	"This is NOT a temporary file"
};
static const true_false_string tfs_file_attribute_sparse = {
	"This is a SPARSE file",
	"This is NOT a sparse file"
};
static const true_false_string tfs_file_attribute_reparse = {
	"This is file has an associated REPARSE point",
	"This file does NOT have an associated reparse point"
};
static const true_false_string tfs_file_attribute_compressed = {
	"This is a COMPRESSED file",
	"This is NOT a compressed file"
};
static const true_false_string tfs_file_attribute_offline = {
	"This file is OFFLINE",
	"This file is NOT offline "
};
static const true_false_string tfs_file_attribute_not_content_indexed = {
	"This file MAY NOT be indexed by the CONTENT INDEXING service",
	"This file MAY be indexed by the content indexing service"
};
static const true_false_string tfs_file_attribute_encrypted = {
	"This is an ENCRYPTED file",
	"This is NOT an encrypted file"
};


static int
dissect_smb_UTIME(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, int hf_date)
{
	nstime_t ts;

	ts.secs = tvb_get_letohl(tvb, offset);
	ts.nsecs = tvb_get_letohl(tvb, offset);

	proto_tree_add_time(tree, hf_date, tvb, offset, 4, &ts);
	offset += 4;

	return offset;
}

static int
dissect_smb_64bit_int(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, int hf_int)
{
	guint32 i;

	/* XXX for now we just read the low 32 bits of the integer and ignore
	   the high 32 bits.
	   We need FT_UINT64.
	*/
	i = tvb_get_letohl(tvb, offset);
	proto_tree_add_uint(tree, hf_int, tvb, offset, 8, i);
	offset += 8;

	return offset;
}

static int
dissect_smb_64bit_time(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset, char *str, int hf_date)
{
	proto_item *item = NULL;
	proto_tree *tree = NULL;
	nstime_t tv;

	/* XXXX we need some way to represent this as a time 
	   properly. For now we display everything as 8 bytes*/

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 8,
			str);
		tree = proto_item_add_subtree(item, ett_smb_64bit_time);
	}

	proto_tree_add_bytes_format(tree, hf_smb_unknown, tvb, offset, 8, tvb_get_ptr(tvb, offset, 8), "%s: cant decode this yet", str);

	offset += 8;
	return offset;
}


static int
dissect_smb_timedate(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset, char *str, int hf_date, int hf_dos_date, int hf_dos_time)
{
	guint16 dos_time, dos_date;
	proto_item *item = NULL;
	proto_tree *tree = NULL;
	struct tm tm;
	time_t t;
	nstime_t tv;
	char ts[256];

	dos_time = tvb_get_letohs(tvb, offset);
	dos_date = tvb_get_letohs(tvb, offset+2);

	if((dos_time==0xffff)&&(dos_time==0xffff)){
		proto_tree_add_text(parent_tree, tvb, offset, 4, "%s : No time specified 0xffffffff", str);
		offset += 4;
		return offset;
	}

	tm.tm_sec = (dos_time&0x1f)*2;
	tm.tm_min = (dos_time>>5)&0x3f;
	tm.tm_hour = (dos_time>>11)&0x1f;
	tm.tm_mday = dos_date&0x1f;
	tm.tm_mon = (dos_date>>5)&0x0f;
	tm.tm_year = ((dos_date>>9)&0x7f)+1980;

	t = mktime(&tm);

	tv.secs = t;
	tv.nsecs = 0;
	
	if(parent_tree){
		strftime(ts, 256, "%d %d, %Y %H:%M:%S", &tm);
		item = proto_tree_add_text(parent_tree, tvb, offset, 4, "%s : %s", str, ts);
		tree = proto_item_add_subtree(item, ett_smb_time_date);
	}

	proto_tree_add_time(tree, hf_date, tvb, offset, 4, &tv);

	proto_tree_add_uint_format(tree, hf_dos_time, tvb, offset, 2, dos_time, "Dos Date: %02d:%02d:%02d (0x%04x)", tm.tm_hour, tm.tm_min, tm.tm_sec, dos_time);

	proto_tree_add_uint_format(tree, hf_dos_date, tvb, offset+2, 2, dos_date, "Dos Date: %04d-%02d-%02d (0x%04x)", tm.tm_year, tm.tm_mon, tm.tm_mday, dos_date);

	offset += 4;

	return offset;
}

static int
dissect_file_attributes(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohl(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2,
			"File Attributes: 0x%04x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_file_attributes);
	}

	proto_tree_add_boolean(tree, hf_smb_file_attr_read_only,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_file_attr_hidden,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_file_attr_system,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_file_attr_volume,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_file_attr_directory,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_file_attr_archive,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_file_attr_device,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_file_attr_normal,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_file_attr_temporary,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_file_attr_sparse,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_file_attr_reparse,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_file_attr_compressed,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_file_attr_offline,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_file_attr_not_content_indexed,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_file_attr_encrypted,
		tvb, offset, 2, mask);

	offset += 2;

	return offset;
}

static const true_false_string tfs_mf_file = {
	"MUST be a FILE",
	"Dont have to be a file"
};
static const true_false_string tfs_mf_dir = {
	"MUST be a DIRECTORY",
	"Dont have to be a directory"
};
static const true_false_string tfs_mf_verify = {
	"MUST verify all writes",
	"Dont have to verify writes"
};
static int
dissect_move_flags(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohs(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2,
			"Flags: 0x%04x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_move_flags);
	}

	proto_tree_add_boolean(tree, hf_smb_move_flags_verify,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_move_flags_dir,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_move_flags_file,
		tvb, offset, 2, mask);

	offset += 2;

	return offset;
}


static int
dissect_search_attributes(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohs(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2,
			"Search Attributes: 0x%04x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_search);
	}

	proto_tree_add_boolean(tree, hf_smb_search_attribute_read_only,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_search_attribute_hidden,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_search_attribute_system,
		tvb, offset, 2, mask);	
	proto_tree_add_boolean(tree, hf_smb_search_attribute_volume,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_search_attribute_directory,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_search_attribute_archive,
		tvb, offset, 2, mask);

	offset += 2;

	return offset;
}

static const true_false_string tfs_open_flags_add_info = {
	"ADDITIONAL INFORMATION requested",
	"Additional information NOT requested"
};
static const true_false_string tfs_open_flags_ex_oplock = {
	"EXCLUSIVE OPLOCK requested",
	"Exclusive oplock NOT requested"
};
static const true_false_string tfs_open_flags_batch_oplock = {
	"BATCH OPLOCK requested",
	"Batch oplock NOT requested"
};
static const true_false_string tfs_open_flags_ealen = {
	"Total Len of EAs requested",
	"Total length of EAs NOT requested"
};

static int
dissect_open_flags(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset, int bm)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohs(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2,
			"Flags: 0x%04x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_open_flags);
	}

	if(bm&0x0001){
		proto_tree_add_boolean(tree, hf_smb_open_flags_add_info,
			tvb, offset, 2, mask);
	}
	if(bm&0x0002){
		proto_tree_add_boolean(tree, hf_smb_open_flags_ex_oplock,
			tvb, offset, 2, mask);
	}
	if(bm&0x0004){
		proto_tree_add_boolean(tree, hf_smb_open_flags_batch_oplock,
			tvb, offset, 2, mask);
	}
	if(bm&0x0008){
		proto_tree_add_boolean(tree, hf_smb_open_flags_ealen,
			tvb, offset, 2, mask);
	}

	offset += 2;

	return offset;
}


static const value_string da_access_vals[] = {
	{ 0x00,		"Open for reading"},
	{ 0x01,		"Open for writing"},
	{ 0x02,		"Open for reading and writing"},
	{ 0x03,		"Open for execute"},
	{0, NULL}
};
static const value_string da_sharing_vals[] = {
	{ 0x00,		"Compatibility mode"},
	{ 0x01,		"Deny read/write/execute (exclusive)"},
	{ 0x02,		"Deny write"},
	{ 0x03,		"Deny read/execute"},
	{ 0x04,		"Deny none"},
	{0, NULL}
};
static const value_string da_locality_vals[] = {
	{ 0x00,		"Locality of reference unknown"},
	{ 0x01,		"Mainly sequential access"},
	{ 0x02,		"Mainly random access"},
	{ 0x03,		"Random access with some locality"},
	{0, NULL}
};
static const true_false_string tfs_da_caching = {
	"Do NOT cache this file",
	"CACHING permitted on this file"
};
static const true_false_string tfs_da_writetru = {
	"Writethrough ENABLED",
	"Writethrough DISABLED"
};
static int
dissect_access(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset, char *type)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohs(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2,
			"%s Access: 0x%04x ", type, mask);
		tree = proto_item_add_subtree(item, ett_smb_desiredaccess);
	}

	proto_tree_add_boolean(tree, hf_smb_access_writetru,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_access_caching,
		tvb, offset, 2, mask);
	proto_tree_add_uint(tree, hf_smb_access_locality,
		tvb, offset, 2, mask);
	proto_tree_add_uint(tree, hf_smb_access_sharing,
		tvb, offset, 2, mask);
	proto_tree_add_uint(tree, hf_smb_access_mode,
		tvb, offset, 2, mask);

	offset += 2;

	return offset;
}

static const value_string oa_open_vals[] = {
	{ 0,		"Open (action taken on open)"},
	{ 1,		"The file existed and was opened"},
	{ 2,		"The file did not exist but was created"},
	{ 3,		"The file existed and was truncated"},
	{0,	NULL}
};
static const true_false_string tfs_oa_lock = {
	"File is open only by this user (at this present time)",
	"File is opened by another user"
};
static int
dissect_open_action(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohs(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2,
			"Action: 0x%04x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_open_action);
	}

	proto_tree_add_boolean(tree, hf_smb_open_action_lock,
		tvb, offset, 2, mask);
	proto_tree_add_uint(tree, hf_smb_open_action_open,
		tvb, offset, 2, mask);

	offset += 2;

	return offset;
}

static const true_false_string tfs_of_create = {
	"CREATE file if it does not exist",
	"FAIL if file does not exist"
};
static const value_string of_open[] = {
	{ 0x00,		"Fail if file exists"},
	{ 0x01,		"Open file if it exists"},
	{ 0x02,		"Truncate File if it exists"},
	{0, NULL}
};
static int
dissect_open_function(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohs(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2,
			"Open Function: 0x%04x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_openfunction);
	}

	proto_tree_add_boolean(tree, hf_smb_open_function_create,
		tvb, offset, 2, mask);
	proto_tree_add_uint(tree, hf_smb_open_function_open,
		tvb, offset, 2, mask);

	offset += 2;

	return offset;
}

static const true_false_string tfs_disconnect_tid = {
	"DISCONNECT TID",
	"Do NOT disconnect TID"
};

static int
dissect_connect_flags(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohs(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2,
			"Flags: 0x%04x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_connect_flags);
	}

	proto_tree_add_boolean(tree, hf_smb_connect_flags_dtid,
		tvb, offset, 2, mask);

	offset += 2;

	return offset;
}

static const true_false_string tfs_setup_action_guest = {
	"LOGGED IN AS GUEST",
	"Normal. Not logged in as guest"
};

static int
dissect_setup_action(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohs(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2,
			"Flags: 0x%04x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_setup_action);
	}

	proto_tree_add_boolean(tree, hf_smb_setup_action_guest,
		tvb, offset, 2, mask);

	offset += 2;

	return offset;
}

static const true_false_string tfs_connect_support_search = {
	"SEARCH bits",
	"No Search"
};
static const true_false_string tfs_connect_support_in_dfs = {
	"IN DFS",
	"NOT in dfs"
};

static int
dissect_connect_support_bits(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohs(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2,
			"Flags: 0x%04x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_connect_support_bits);
	}

	proto_tree_add_boolean(tree, hf_smb_connect_support_search,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_connect_support_in_dfs,
		tvb, offset, 2, mask);

	offset += 2;

	return offset;
}


static int
dissect_query_information_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	const char *fn;
	int fn_len;

	/* Word Count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1,
		tvb_get_guint8(tvb, offset));
	offset += 1;

	/* Byte Count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2,
		tvb_get_letohs(tvb, offset));
	offset += 2;

	/* Buffer Format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1,
		tvb_get_guint8(tvb, offset));
	offset += 1;

	/* File Name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	return offset;
}

static int
dissect_query_information_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	int wc;
	nstime_t ts;

	/* Word Count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	if(wc==0){
		/* if there was an error we dont have any parameter words.
		   e.g. if the file did not exist. Seen in real capture */
		goto bytecount;
	}

	/* File Attributes */
	offset = dissect_file_attributes(tvb, pinfo, tree, offset);

	/* Last Write Time */
	ts.secs = tvb_get_letohl(tvb, offset);
	ts.nsecs = 0;
	proto_tree_add_time(tree, hf_smb_last_write_time, tvb, offset, 4, &ts);
	offset += 4;

	/* File Size */
	proto_tree_add_uint(tree, hf_smb_file_size, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* 10 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 10, tvb_get_ptr(tvb, offset, 10));
	offset += 10;

bytecount:
	/* Byte Count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2,
		tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_negprot_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	proto_item *it = NULL;
	proto_tree *tr = NULL;
	int bc;

	/* Word Count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1,
		tvb_get_guint8(tvb, offset));
	offset += 1;

	/* Byte Count */
	bc = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, bc);
	offset += 2;

	if(tree){
		it = proto_tree_add_text(tree, tvb, offset, bc,
				"Requested Dialects");
		tr = proto_item_add_subtree(it, ett_smb_dialects);
	}

	while(bc){
		int len;
		int old_offset = offset;
		const guint8 *str;
		proto_item *dit = NULL;
		proto_tree *dtr = NULL;

		len = tvb_strsize(tvb, offset+1);
		str = tvb_get_ptr(tvb, offset+1, len);

		if(tr){
			dit = proto_tree_add_text(tr, tvb, offset, len+1,
					"Dialect: %s", str);
			dtr = proto_item_add_subtree(dit, ett_smb_dialect);
		}

		/* Buffer Format */
		proto_tree_add_uint(dtr, hf_smb_buffer_format, tvb, offset, 1,
			tvb_get_guint8(tvb, offset));
		offset += 1;
		bc -= 1;

		/*Dialect Name */
		proto_tree_add_string(dtr, hf_smb_dialect_name, tvb, offset, 
			len, str);
		offset += len;
		bc -= len;
	}

	return offset;
}


#define SECURITY_MODE_MODE             0x01
#define SECURITY_MODE_PASSWORD         0x02
#define SECURITY_MODE_SIGNATURES       0x04
#define SECURITY_MODE_SIG_REQUIRED     0x08
static const true_false_string tfs_sm_mode = {
	"USER security mode",
	"SHARE security mode"
};
static const true_false_string tfs_sm_password = {
	"ENCRYPTED password. Use challenge response",
	"PLAINTEXT password"
};
static const true_false_string tfs_sm_signatures = {
	"Security signatures ENABLED",
	"Security signatures NOT enabled"
};
static const true_false_string tfs_sm_sig_required = {
	"Security signatures REQUIRED",
	"Security signatures NOT required"
};


static int
dissect_negprot_security_mode(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset, int wc)
{
	guint16 mask = 0;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	if(parent_tree){
		switch(wc){
		case 13:
			mask = tvb_get_letohs(tvb, offset);
			item = proto_tree_add_text(parent_tree, tvb, offset, 2,
					"Security Mode: 0x%04x ", mask);
			break;
		case 17:
			mask = tvb_get_guint8(tvb, offset);
			item = proto_tree_add_text(parent_tree, tvb, offset, 1,
					"Security Mode: 0x%02x ", mask);
			break;
		}
		tree = proto_item_add_subtree(item, ett_smb_mode);
	}

	if(wc==13){
		proto_tree_add_boolean(tree, hf_sm_mode16, tvb, offset, 2, mask);
		proto_tree_add_boolean(tree, hf_sm_password16, tvb, offset, 2, mask);
		offset += 2;
		return offset;
	}

	if(wc==17){
		proto_tree_add_boolean(tree, hf_sm_mode, tvb, offset, 1, mask);
		proto_tree_add_boolean(tree, hf_sm_password, tvb, offset, 1, mask);
		proto_tree_add_boolean(tree, hf_sm_signatures, tvb, offset, 1, mask);
		proto_tree_add_boolean(tree, hf_sm_sig_required, tvb, offset, 1, mask);
		offset +=1;
		return offset;
	}

	offset += 2;
	return offset;
}

#define RAWMODE_READ   0x01
#define RAWMODE_WRITE  0x02
static const true_false_string tfs_rm_read = {
	"READ raw supported",
	"Read raw NOT supported"
};
static const true_false_string tfs_rm_write = {
	"WRITE raw supported",
	"Write raw NOT supported"
};

static int
dissect_negprot_rawmode(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohs(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2, "Rawmode: 0x%04x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_rawmode);
	}

	proto_tree_add_boolean(tree, hf_rm_read, tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_rm_write, tvb, offset, 2, mask);

	offset += 2;

	return offset;
}


#define SERVER_CAP_RAW_MODE            0x00000001
#define SERVER_CAP_MPX_MODE            0x00000002
#define SERVER_CAP_UNICODE             0x00000004
#define SERVER_CAP_LARGE_FILES         0x00000008
#define SERVER_CAP_NT_SMBS             0x00000010
#define SERVER_CAP_RPC_REMOTE_APIS     0x00000020
#define SERVER_CAP_STATUS32            0x00000040
#define SERVER_CAP_LEVEL_II_OPLOCKS    0x00000080
#define SERVER_CAP_LOCK_AND_READ       0x00000100
#define SERVER_CAP_NT_FIND             0x00000200
#define SERVER_CAP_DFS                 0x00001000
#define SERVER_CAP_INFOLEVEL_PASSTHRU  0x00002000
#define SERVER_CAP_LARGE_READX         0x00004000
#define SERVER_CAP_LARGE_WRITEX        0x00008000
#define SERVER_CAP_UNIX                0x00800000
#define SERVER_CAP_RESERVED            0x02000000
#define SERVER_CAP_BULK_TRANSFER       0x20000000
#define SERVER_CAP_COMPRESSED_DATA     0x40000000
#define SERVER_CAP_EXTENDED_SECURITY   0x80000000
static const true_false_string tfs_server_cap_raw_mode = {
	"RAW MODE capability is available",
	"Raw mode capability is NOT available"
};
static const true_false_string tfs_server_cap_mpx_mode = {
	"MPX MODE capability is available",
	"Mpx mode capability is NOT available"
};
static const true_false_string tfs_server_cap_unicode = {
	"UNICODE capability is available",
	"Unicode capability is NOT available"
};
static const true_false_string tfs_server_cap_large_files = {
	"LARGE FILES capability is available",
	"Large files capability is NOT available"
};
static const true_false_string tfs_server_cap_nt_smbs = {
	"NT SMBS capability is available",
	"Nt smbs capability is NOT available"
};
static const true_false_string tfs_server_cap_rpc_remote_apis = {
	"RPC REMOTE APIS capability is available",
	"Rpc remote apis capability is NOT available"
};
static const true_false_string tfs_server_cap_status32 = {
	"STATUS32 capability is available",
	"Status32 capability is NOT available"
};
static const true_false_string tfs_server_cap_level_ii_oplocks = {
	"LEVEL II OPLOCKS capability is available",
	"Level ii oplocks capability is NOT available"
};
static const true_false_string tfs_server_cap_lock_and_read = {
	"LOCK AND READ capability is available",
	"Lock and read capability is NOT available"
};
static const true_false_string tfs_server_cap_nt_find = {
	"NT FIND capability is available",
	"Nt find capability is NOT available"
};
static const true_false_string tfs_server_cap_dfs = {
	"DFS capability is available",
	"Dfs capability is NOT available"
};
static const true_false_string tfs_server_cap_infolevel_passthru = {
	"INFOLEVEL PASSTHRU capability is available",
	"Infolevel passthru capability is NOT available"
};
static const true_false_string tfs_server_cap_large_readx = {
	"LARGE READX capability is available",
	"Large readx capability is NOT available"
};
static const true_false_string tfs_server_cap_large_writex = {
	"LARGE WRITEX capability is available",
	"Large writex capability is NOT available"
};
static const true_false_string tfs_server_cap_unix = {
	"UNIX capability is available",
	"Unix capability is NOT available"
};
static const true_false_string tfs_server_cap_reserved = {
	"Reserved",
	"Reserved"
};
static const true_false_string tfs_server_cap_bulk_transfer = {
	"BULK TRANSFER capability is available",
	"Bulk transfer capability is NOT available"
};
static const true_false_string tfs_server_cap_compressed_data = {
	"COMPRESSED DATA capability is available",
	"Compressed data capability is NOT available"
};
static const true_false_string tfs_server_cap_extended_security = {
	"EXTENDED SECURITY capability is available",
	"Extended security capability is NOT available"
};

static int
dissect_negprot_capabilities(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint32 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohl(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 4, "Capabilities: 0x%08x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_capabilities);
	}

	proto_tree_add_boolean(tree, hf_server_cap_raw_mode,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_mpx_mode,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_unicode,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_large_files,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_nt_smbs,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_rpc_remote_apis,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_status32,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_level_ii_oplocks,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_lock_and_read,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_nt_find,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_dfs,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_infolevel_passthru,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_large_readx,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_large_writex,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_unix,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_reserved,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_bulk_transfer,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_compressed_data,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_server_cap_extended_security,
		tvb, offset, 4, mask);

	return mask;
}


static int
dissect_negprot_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	int wc;
	guint16 dialect;
	const char *dn;
	int dn_len;

	/* Word Count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	/* Dialect Index */
	dialect = tvb_get_letohs(tvb, offset);
	switch(wc){
	case 1:
		if(dialect==0xffff){
			proto_tree_add_uint_format(tree, hf_smb_dialect_index,
				tvb, offset, 2, dialect,
				"Selected Index: -1, PC NETWORK PROGRAM 1.0 choosen");
		} else {
			proto_tree_add_uint(tree, hf_smb_dialect_index,
				tvb, offset, 2, dialect);
		}
		break;
	case 13:
		proto_tree_add_uint_format(tree, hf_smb_dialect_index,
			tvb, offset, 2, dialect, 
			"Dialect Index: %u, Greater than CORE PROTOCOL and up to LANMAN2.1", dialect);
		break;
	case 17:
		proto_tree_add_uint_format(tree, hf_smb_dialect_index,
			tvb, offset, 2, dialect, 
			"Dialect Index: %u,  greater than LANMAN2.1", dialect);
		break;
	}
	offset += 2;


	if(wc==1){
		/* Byte Count */
		proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2,
			tvb_get_letohs(tvb, offset));
		offset += 2;
	}


	if(wc==13){
		guint16 ekl;
		gint16 tz;

		/* Security Mode */
		offset = dissect_negprot_security_mode(tvb, pinfo, tree, offset,
				wc);

		/* Maximum Transmit Buffer Size */
		proto_tree_add_uint(tree, hf_max_trans_buf_size,
			tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* Maximum Multiplex Count */
		proto_tree_add_uint(tree, hf_max_mpx_count,
			tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* Maximum Vcs Number */
		proto_tree_add_uint(tree, hf_max_vcs_num,
			tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* raw mode */
		offset = dissect_negprot_rawmode(tvb, pinfo, tree, offset);

		/* session key */
		proto_tree_add_uint(tree, hf_session_key,
			tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* current time and date at server */
		offset = dissect_smb_timedate(tvb, pinfo, tree, offset, "Current Time and Date at server", hf_current_server_date, hf_current_server_smb_date, hf_current_server_smb_time);

		/* time zone */
		tz = tvb_get_letohs(tvb, offset);
		proto_tree_add_int_format(tree, hf_server_timezone, tvb, offset, 2, tz, "Server Time Zone: %d min from UTC", tz);
		offset += 2;

		/* encryption key length */
		ekl = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_encryption_key_length, tvb, offset, 2, ekl);
		offset += 2;

		/* 2 reserved bytes */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 2, tvb_get_ptr(tvb, offset, 2));
		offset += 2;

		/* byte count */
		proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* challange/ encryption key */
		if(ekl){
			proto_tree_add_bytes(tree, hf_encryption_key, tvb, offset, ekl, tvb_get_ptr(tvb, offset, ekl));
			offset += ekl;
		}

		/* is this string optional ? */
		if(tvb_length_remaining(tvb, offset)>0){
			/* domain */
			dn = get_unicode_or_ascii_string_tvb(tvb, &offset,
				pinfo, &dn_len, FALSE, FALSE);
			proto_tree_add_string(tree, hf_domain, tvb,
				offset, dn_len,dn);
			offset += dn_len;
		}
	}


	if(wc==17){
		guint16 ekl;
		guint32 caps;
		gint16 tz;


		/* Security Mode */
		offset = dissect_negprot_security_mode(tvb, pinfo, tree, offset, wc);

		/* Maximum Multiplex Count */
		proto_tree_add_uint(tree, hf_max_mpx_count,
			tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* Maximum Vcs Number */
		proto_tree_add_uint(tree, hf_max_vcs_num,
			tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* Maximum Transmit Buffer Size */
		proto_tree_add_uint(tree, hf_max_trans_buf_size,
			tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* maximum raw buffer size */
		proto_tree_add_uint(tree, hf_max_raw_buf_size,
			tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* session key */
		proto_tree_add_uint(tree, hf_session_key,
			tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* server capabilities */
		caps = dissect_negprot_capabilities(tvb, pinfo, tree, offset);
		offset += 4;

		/* XXX should have a routine to display this as date/time */
		/* system time low */
		proto_tree_add_uint(tree, hf_time_low,
			tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* system time high */
		proto_tree_add_uint(tree, hf_time_high,
			tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* time zone */
		tz = tvb_get_letohs(tvb, offset);
		proto_tree_add_int_format(tree, hf_server_timezone,
			tvb, offset, 2, tz,
			"Server Time Zone: %d min from UTC", tz);
		offset += 2;

		/* encryption key length */
		ekl = tvb_get_guint8(tvb, offset);
		proto_tree_add_uint(tree, hf_encryption_key_length,
			tvb, offset, 1, ekl);
		offset += 1;

		/* byte count */
		proto_tree_add_uint(tree, hf_smb_byte_count,
			tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		if(!(caps&SERVER_CAP_EXTENDED_SECURITY)){
			smb_info *si;

			/* challange/ encryption key */
			if(ekl){
				proto_tree_add_bytes(tree, hf_encryption_key,
					tvb, offset, ekl,
					tvb_get_ptr(tvb, offset, ekl));
				offset += ekl;
			}

			if(tvb_length_remaining(tvb, offset)){
				/* domain */
				/* this string is special, unicode is flagged in caps */
				/* This string is NOT padded to be 16bit aligned. (seen in actual capture) */
				si = pinfo->private;
				si->unicode = (caps&SERVER_CAP_UNICODE);
				dn = get_unicode_or_ascii_string_tvb(tvb,
					&offset, pinfo, &dn_len, TRUE, FALSE);
				proto_tree_add_string(tree, hf_domain,
					tvb, offset, dn_len,dn);
				offset += dn_len;
			}
		} else {
			int len;

			/* guid */
			proto_tree_add_bytes(tree, hf_server_guid,
				tvb, offset, 16, tvb_get_ptr(tvb, offset, 16));
			offset += 16;

			/* security blob */
			len = tvb_length_remaining(tvb, offset);
			if(len){
				proto_tree_add_bytes(tree, hf_security_blob,
					tvb, offset, len,
					tvb_get_ptr(tvb, offset, len));
				offset += len;
			}
		}

	}

	return offset;
}

static int
dissect_tree_connect_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1,
		tvb_get_guint8(tvb, offset));
	offset += 1;
 
	/* Maximum Buffer Size */
	proto_tree_add_uint(tree, hf_max_buf_size, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* tid */
	proto_tree_add_uint(tree, hf_smb_tid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2,
		tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_tree_connect_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	int an_len, pwlen;
	const char *an;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1,
		tvb_get_guint8(tvb, offset));
	offset += 1;
 
	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2,
		tvb_get_letohs(tvb, offset));
	offset += 2;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1,
		tvb_get_guint8(tvb, offset));
	offset += 1;

	/* Path */
	an = get_unicode_or_ascii_string_tvb(tvb, &offset,
		pinfo, &an_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_path, tvb,
		offset, an_len, an);
	offset += an_len;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1,
		tvb_get_guint8(tvb, offset));
	offset += 1;

	/* password, ANSI */
	pwlen = tvb_strsize(tvb, offset);
	proto_tree_add_bytes(tree, hf_smb_password, 
		tvb, offset, pwlen, tvb_get_ptr(tvb, offset, pwlen));
	offset += pwlen;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1,
		tvb_get_guint8(tvb, offset));
	offset += 1;

	/* Service */
	an = get_unicode_or_ascii_string_tvb(tvb, &offset,
		pinfo, &an_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_service, tvb,
		offset, an_len, an);
	offset += an_len;

	return offset;
}

static int
dissect_create_dir_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	int dn_len;
	const char *dn;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1,
		tvb_get_guint8(tvb, offset));
	offset += 1;
 
	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2,
		tvb_get_letohs(tvb, offset));
	offset += 2;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1,
		tvb_get_guint8(tvb, offset));
	offset += 1;

	/* dir name */
	dn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &dn_len,
		FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_dir_name, tvb, offset, dn_len,
		dn);
	offset += dn_len;

	return offset;
}

static int
dissect_lock_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* fid */
	proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* lock count */
	proto_tree_add_uint(tree, hf_smb_count, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* offset */
	proto_tree_add_uint(tree, hf_smb_offset, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_create_file_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	int fn_len;
	const char *fn;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* file attributes */
	offset = dissect_file_attributes(tvb, pinfo, tree, offset);

	/* creation time */
	offset = dissect_smb_UTIME(tvb, pinfo, tree, offset, hf_smb_create_time);

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* File Name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	return offset;
}

static int
dissect_sid_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* sid */
	proto_tree_add_uint(tree, hf_smb_sid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_fid_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* fid */
	proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_empty_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
 
	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;
 
	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_close_file_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* fid */
	proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* last write time */
	offset = dissect_smb_timedate(tvb, pinfo, tree, offset,
		"Last Write", hf_smb_last_write_date,
		hf_smb_last_write_dos_date, hf_smb_last_write_dos_time);

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}


static int
dissect_delete_file_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	int fn_len;
	const char *fn;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* file attributes */
	offset = dissect_file_attributes(tvb, pinfo, tree, offset);

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* file name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	return offset;
}

static int
dissect_rename_file_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	int fn_len;
	const char *fn;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* file attributes */
	offset = dissect_file_attributes(tvb, pinfo, tree, offset);

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* old file name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_old_file_name, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* file name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	return offset;
}

static int
dissect_move_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	int fn_len;
	const char *fn;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* read count */
	proto_tree_add_uint(tree, hf_smb_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* file name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	return offset;
}

static int
dissect_echo_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint16 ec, bc;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* echo count */
	ec = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_echo_count, tvb, offset, 2, ec);
	offset += 2;

	/* byte count */
	bc = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, bc);
	offset += 2;

	/* echo data */
	proto_tree_add_bytes(tree, hf_smb_echo_data, tvb, offset, bc, tvb_get_ptr(tvb, offset, bc));
	offset += bc;

	return offset;
}

static int
dissect_echo_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint16 bc;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* echo sequence number */
	proto_tree_add_uint(tree, hf_smb_echo_seq_num, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* byte count */
	bc = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, bc);
	offset += 2;

	/* echo data */
	proto_tree_add_bytes(tree, hf_smb_echo_data, tvb, offset, bc, tvb_get_ptr(tvb, offset, bc));
	offset += bc;

	return offset;
}

static int
dissect_move_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	int fn_len;
	guint16 tid;
	const char *fn;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* tid */
	tid = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint_format(tree, hf_smb_tid, tvb, offset, 2, tid,
		"TID (target): 0x%04x", tid);
	offset += 2;

	/* open function */
	offset = dissect_open_function(tvb, pinfo, tree, offset);

	/* move flags */
	offset = dissect_move_flags(tvb, pinfo, tree, offset);

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* file name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string_format(tree, hf_smb_file_name, tvb, offset,
		fn_len,	fn, "Old File Name: %s", fn);
	offset += fn_len;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* file name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string_format(tree, hf_smb_file_name, tvb, offset,
		fn_len,	fn, "New File Name: %s", fn);
	offset += fn_len;

	return offset;
}

static int
dissect_set_information_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	int fn_len;
	const char *fn;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* file attributes */
	offset = dissect_file_attributes(tvb, pinfo, tree, offset);

	/* last write time */
	offset = dissect_smb_timedate(tvb, pinfo, tree, offset,
		"Last Write", hf_smb_last_write_date,
		hf_smb_last_write_dos_date, hf_smb_last_write_dos_time);

	/* 10 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 10, tvb_get_ptr(tvb, offset, 10));
	offset += 10;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* file name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	return offset;
}

static int
dissect_read_file_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* fid */
	proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* read count */
	proto_tree_add_uint(tree, hf_smb_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* offset */
	proto_tree_add_uint(tree, hf_smb_offset, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* remaining */
	proto_tree_add_uint(tree, hf_smb_remaining, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_read_file_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint16 cnt, bc;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* read count */
	cnt = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_count, tvb, offset, 2, cnt);
	offset += 2;

	/* 8 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 8, tvb_get_ptr(tvb, offset, 8));
	offset += 8;

	/* byte count */
	bc = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, bc);
	offset += 2;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;
	bc -= 1;

	/* file data */
	if(bc>cnt){
		/* We have some initial padding bytes.
		   The CIFS specs does NOT mention that there can be
		   some padding bytes here. It was seen by looking at
		   some captures.
		*/
		proto_tree_add_bytes(tree, hf_smb_padding, tvb, offset, bc-cnt,
			tvb_get_ptr(tvb, offset, bc-cnt));
		offset += bc-cnt;
		bc = cnt;
	}
	if(bc>tvb_length_remaining(tvb, offset)){
		int len;
		len = tvb_length_remaining(tvb, offset);
		proto_tree_add_bytes_format(tree, hf_smb_file_data, tvb, offset, len, tvb_get_ptr(tvb, offset, len),"File Data: Incomplete. Only %d of %d bytes", len, bc);
		offset += len;
	} else {
		proto_tree_add_bytes(tree, hf_smb_file_data, tvb, offset, bc, tvb_get_ptr(tvb, offset, bc));
		offset += bc;
	}

	return offset;
}

static int
dissect_write_file_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint16 cnt;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* fid */
	proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* write count */
	cnt = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_count, tvb, offset, 2, cnt);
	offset += 2;

	/* offset */
	proto_tree_add_uint(tree, hf_smb_offset, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* remaining */
	proto_tree_add_uint(tree, hf_smb_remaining, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* data len */
	proto_tree_add_uint(tree, hf_smb_data_len, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/*XXX Do we have to do something like in dissect_read_file_response()?
	      Must check some captures. */
	/* file data */
	if(cnt>tvb_length_remaining(tvb, offset)){
		int len;
		len = tvb_length_remaining(tvb, offset);
		proto_tree_add_bytes_format(tree, hf_smb_file_data, tvb, offset, len, tvb_get_ptr(tvb, offset, len),"Incomplete data. Only %d of %d bytes", len, cnt);
		offset += len;
	} else {
		proto_tree_add_bytes(tree, hf_smb_file_data, tvb, offset, cnt, tvb_get_ptr(tvb, offset, cnt));
		offset += cnt;
	}

	return offset;
}

static int
dissect_write_and_close_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint16 cnt;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* write count */
	cnt = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_count, tvb, offset, 2, cnt);
	offset += 2;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_write_and_close_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint8 wc;
	guint16 cnt;

	/* word count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	/* fid */
	proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* write count */
	cnt = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_count, tvb, offset, 2, cnt);
	offset += 2;

	/* offset */
	proto_tree_add_uint(tree, hf_smb_offset, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* last write time */
	offset = dissect_smb_UTIME(tvb, pinfo, tree, offset, hf_smb_last_write_time);
	
	if(wc==12){
		/* 12 reserved bytes */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 12, tvb_get_ptr(tvb, offset, 12));
		offset += 12;
	}

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* 1 pad byte */
	proto_tree_add_bytes(tree, hf_smb_padding, tvb, offset, 1,
		tvb_get_ptr(tvb, offset, 1));
	offset += 1;
	
	/*XXX Do we have to do something like in dissect_read_file_response()?
	      Must check some captures. */
	/* file data */
	if(cnt>tvb_length_remaining(tvb, offset)){
		int len;
		len = tvb_length_remaining(tvb, offset);
		proto_tree_add_bytes_format(tree, hf_smb_file_data, tvb, offset, len, tvb_get_ptr(tvb, offset, len),"Incomplete data. Only %d of %d bytes", len, cnt);
		offset += len;
	} else {
		proto_tree_add_bytes(tree, hf_smb_file_data, tvb, offset, cnt, tvb_get_ptr(tvb, offset, cnt));
		offset += cnt;
	}

	return offset;
}

static int
dissect_open_file_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* fid */
	proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* File Attributes */
	offset = dissect_file_attributes(tvb, pinfo, tree, offset);

	/* last write time */
	offset = dissect_smb_UTIME(tvb, pinfo, tree, offset, hf_smb_last_write_time);
	
	/* File Size */
	proto_tree_add_uint(tree, hf_smb_file_size, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* granted access */
	offset = dissect_access(tvb, pinfo, tree, offset, "Granted");

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}



static int
dissect_open_file_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	int fn_len;
	const char *fn;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* desired access */
	offset = dissect_access(tvb, pinfo, tree, offset, "Desired");

	/* Search Attributes */
	offset = dissect_search_attributes(tvb, pinfo, tree, offset);

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* file name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	return offset;
}



static int
dissect_set_information2_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* fid */
	proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* create time */
	offset = dissect_smb_timedate(tvb, pinfo, tree, offset,
		"Creation", hf_smb_create_time,
		hf_smb_create_dos_date, hf_smb_create_dos_time);

	/* access time */
	offset = dissect_smb_timedate(tvb, pinfo, tree, offset,
		"Last Access", hf_smb_access_time,
		hf_smb_access_dos_date, hf_smb_access_dos_time);

	/* last write time */
	offset = dissect_smb_timedate(tvb, pinfo, tree, offset,
		"Last Write", hf_smb_last_write_date,
		hf_smb_last_write_dos_date, hf_smb_last_write_dos_time);

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_query_information2_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* create time */
	offset = dissect_smb_timedate(tvb, pinfo, tree, offset,
		"Creation", hf_smb_create_time,
		hf_smb_create_dos_date, hf_smb_create_dos_time);

	/* access time */
	offset = dissect_smb_timedate(tvb, pinfo, tree, offset,
		"Last Access", hf_smb_access_time,
		hf_smb_access_dos_date, hf_smb_access_dos_time);

	/* last write time */
	offset = dissect_smb_timedate(tvb, pinfo, tree, offset,
		"Last Write", hf_smb_last_write_date,
		hf_smb_last_write_dos_date, hf_smb_last_write_dos_time);

	/* data size */
	proto_tree_add_uint(tree, hf_smb_data_size, tvb, offset, 4, tvb_get_letohs(tvb, offset));
	offset += 4;

	/* allocation size */
	proto_tree_add_uint(tree, hf_smb_alloc_size, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* File Attributes */
	offset = dissect_file_attributes(tvb, pinfo, tree, offset);

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_write_file_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* write count */
	proto_tree_add_uint(tree, hf_smb_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_create_temporary_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	int fn_len;
	const char *fn;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* 2 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 2, tvb_get_ptr(tvb, offset, 2));
	offset += 2;

	/* Creation time */
	offset = dissect_smb_UTIME(tvb, pinfo, tree, offset, hf_smb_create_time);

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* directory name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_dir_name, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	return offset;
}

static int
dissect_create_temporary_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	int fn_len;
	const char *fn;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* fid */
	proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* file name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	return offset;
}

static int
dissect_seek_file_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* fid */
	proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* Seek Mode */
	proto_tree_add_uint(tree, hf_smb_seek_mode, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* offset */
	proto_tree_add_uint(tree, hf_smb_offset, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_seek_file_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* offset */
	proto_tree_add_uint(tree, hf_smb_offset, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_lock_and_read_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint16 cnt, bc;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* read count */
	cnt = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_count, tvb, offset, 2, cnt);
	offset += 2;

	/* 8 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 8, tvb_get_ptr(tvb, offset, 8));
	offset += 8;

	/* byte count */
	bc = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, bc);
	offset += 2;

	/* buffer format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;
	bc -= 1;

	/* data len */
	proto_tree_add_uint(tree, hf_smb_data_len, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}


static int
dissect_read_raw_file_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint16 wc;
	guint32 to;

	/* word count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	/* fid */
	proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* offset */
	proto_tree_add_uint(tree, hf_smb_offset, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* max count */
	proto_tree_add_uint(tree, hf_smb_maxcount, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* min count */
	proto_tree_add_uint(tree, hf_smb_mincount, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* timeout */
	to = tvb_get_letohl(tvb, offset);
	proto_tree_add_uint_format(tree, hf_smb_timeout, tvb, offset, 4, to, "Timeout: %d.%03d", to/1000, to%1000);
	offset += 4;

	/* 2 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 2, tvb_get_ptr(tvb, offset, 2));
	offset += 2;

	if(wc==10){
		/* high offset */
		proto_tree_add_uint(tree, hf_smb_high_offset, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;
	}

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_read_mpx_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* fid */
	proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* offset */
	proto_tree_add_uint(tree, hf_smb_offset, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* max count */
	proto_tree_add_uint(tree, hf_smb_maxcount, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* min count */
	proto_tree_add_uint(tree, hf_smb_mincount, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* 6 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 6, tvb_get_ptr(tvb, offset, 6));
	offset += 6;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_read_mpx_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint16 len, bc;

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* offset */
	proto_tree_add_uint(tree, hf_smb_offset, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* count */
	proto_tree_add_uint(tree, hf_smb_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* 2 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 2, tvb_get_ptr(tvb, offset, 2));
	offset += 2;

	/* data compaction mode */
	proto_tree_add_uint(tree, hf_smb_dcm, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* 2 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 2, tvb_get_ptr(tvb, offset, 2));
	offset += 2;

	/* data len */
	len = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_data_len, tvb, offset, 2, len);
	offset += 2;

	/* data offset */
	proto_tree_add_uint(tree, hf_smb_data_offset, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* byte count */
	bc = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, bc);
	offset += 2;

	/* file data */
	if(bc>len){
		/* We have some initial padding bytes. */
		proto_tree_add_bytes(tree, hf_smb_padding, tvb, offset, bc-len,
			tvb_get_ptr(tvb, offset, bc-len));
		offset += bc-len;
		bc = len;
	}
	if(bc>tvb_length_remaining(tvb, offset)){
		int len;
		len = tvb_length_remaining(tvb, offset);
		proto_tree_add_bytes_format(tree, hf_smb_file_data, tvb, offset, len, tvb_get_ptr(tvb, offset, len),"File Data: Incomplete. Only %d of %d bytes", len, bc);
		offset += len;
	} else {
		proto_tree_add_bytes(tree, hf_smb_file_data, tvb, offset, bc, tvb_get_ptr(tvb, offset, bc));
		offset += bc;
	}

	return offset;
}

static int
dissect_query_information_disk_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{

	/* word count */
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;

	/* units */
	proto_tree_add_uint(tree, hf_smb_units, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* bpu */
	proto_tree_add_uint(tree, hf_smb_bpu, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* block size */
	proto_tree_add_uint(tree, hf_smb_blocksize, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* free units */
	proto_tree_add_uint(tree, hf_smb_freeunits, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* 2 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 2, tvb_get_ptr(tvb, offset, 2));
	offset += 2;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}

static int
dissect_session_setup_andx_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint8	wc, cmd;
	guint16 andxoffset;
	int an_len;
	const char *an;

	/* word count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	/* next smb command */
	cmd = tvb_get_guint8(tvb, offset);
	if(cmd!=0xff){
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: %s (0x%02x)", decode_smb_name(cmd), cmd);
	} else {
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: No further commands");
	}
	offset += 1;

	/* reserved byte */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 1, tvb_get_ptr(tvb, offset, 1));
	offset += 1;

	/* andxoffset */
	andxoffset = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_andxoffset, tvb, offset, 2, andxoffset);
	offset += 2;

	/* Maximum Buffer Size */
	proto_tree_add_uint(tree, hf_max_buf_size, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* Maximum Multiplex Count */
	proto_tree_add_uint(tree, hf_max_mpx_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* VC Number */
	proto_tree_add_uint(tree, hf_vc_num, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* session key */
	proto_tree_add_uint(tree, hf_session_key, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	if(wc==10){
		guint16 pwlen;

		/* password length, ASCII*/
		pwlen = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint_format(tree, hf_smb_password_len,
			tvb, offset, 2, pwlen,
			"Password (ASCII) Len: %d", pwlen);
		offset += 2;

		/* 4 reserved bytes */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 4, tvb_get_ptr(tvb, offset, 4));
		offset += 4;

		/* byte count */
		proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* password, ANSI */
		proto_tree_add_bytes(tree, hf_smb_password, 
			tvb, offset, pwlen, tvb_get_ptr(tvb, offset, pwlen));
		offset += pwlen;
	}

	if(wc==13){
		guint16 apwlen, upwlen;

		/* password length, ANSI*/
		apwlen = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint_format(tree, hf_smb_password_len,
			tvb, offset, 2, apwlen,
			"Password (ANSI) Len: %d", apwlen);
		offset += 2;

		/* password length, UNICODE*/
		upwlen = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint_format(tree, hf_smb_password_len,
			tvb, offset, 2, upwlen,
			"Password (UNICODE) Len: %d", upwlen);
		offset += 2;

		/* 4 reserved bytes */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 4, tvb_get_ptr(tvb, offset, 4));
		offset += 4;

		/* capabilities */
		dissect_negprot_capabilities(tvb, pinfo, tree, offset);
		offset += 4;

		/* byte count */
		proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		if(apwlen){
			/* password, ANSI */
			proto_tree_add_bytes(tree, hf_smb_password, 
				tvb, offset, apwlen,
				tvb_get_ptr(tvb, offset, apwlen));
			offset += apwlen;
		}

		if(upwlen){
			/* password, UNICODE */
			proto_tree_add_bytes(tree, hf_smb_password, 
				tvb, offset, upwlen,
				tvb_get_ptr(tvb, offset, upwlen));
			offset += upwlen;
		}

	}

	/* Account Name */
	an = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &an_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_account, tvb, offset, an_len,
		an);
	offset += an_len;

	/* Primary domain */
	an = get_unicode_or_ascii_string_tvb(tvb, &offset,
		pinfo, &an_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_domain, tvb,
		offset, an_len, an);
	offset += an_len;

	/* OS */
	an = get_unicode_or_ascii_string_tvb(tvb, &offset,
		pinfo, &an_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_os, tvb,
		offset, an_len, an);
	offset += an_len;

	/* LANMAN */
	an = get_unicode_or_ascii_string_tvb(tvb, &offset,
		pinfo, &an_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_lanman, tvb,
		offset, an_len, an);
	offset += an_len;

	/* call AndXCommand (if there are any) */
	call_next_command(tvb, pinfo, tree, andxoffset, smb_tree, cmd);

	return offset;
}

static int
dissect_tree_connect_andx_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint8	wc, cmd;
	guint16 andxoffset, pwlen;
	int an_len;
	const char *an;

	/* word count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	/* next smb command */
	cmd = tvb_get_guint8(tvb, offset);
	if(cmd!=0xff){
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: %s (0x%02x)", decode_smb_name(cmd), cmd);
	} else {
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: No further commands");
	}
	offset += 1;

	/* reserved byte */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 1, tvb_get_ptr(tvb, offset, 1));
	offset += 1;

	/* andxoffset */
	andxoffset = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_andxoffset, tvb, offset, 2, andxoffset);
	offset += 2;

	/* flags */
	offset = dissect_connect_flags(tvb, pinfo, tree, offset);

	/* password length*/
	pwlen = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_password_len, tvb, offset, 2, pwlen);
	offset += 2;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* password */
	proto_tree_add_bytes(tree, hf_smb_password, 
		tvb, offset, pwlen,
		tvb_get_ptr(tvb, offset, pwlen));
	offset += pwlen;

	/* Path */
	an = get_unicode_or_ascii_string_tvb(tvb, &offset,
		pinfo, &an_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_path, tvb,
		offset, an_len, an);
	offset += an_len;

	/* Service */
	an = get_unicode_or_ascii_string_tvb(tvb, &offset,
		pinfo, &an_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_service, tvb,
		offset, an_len, an);
	offset += an_len;

	/* call AndXCommand (if there are any) */
	call_next_command(tvb, pinfo, tree, andxoffset, smb_tree, cmd);

	return offset;
}

static int
dissect_session_setup_andx_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint8	wc, cmd;
	guint16 andxoffset;
	int an_len;
	const char *an;

	/* word count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	/* next smb command */
	cmd = tvb_get_guint8(tvb, offset);
	if(cmd!=0xff){
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: %s (0x%02x)", decode_smb_name(cmd), cmd);
	} else {
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: No further commands");
	}
	offset += 1;

	/* reserved byte */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 1, tvb_get_ptr(tvb, offset, 1));
	offset += 1;

	/* andxoffset */
	andxoffset = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_andxoffset, tvb, offset, 2, andxoffset);
	offset += 2;

	/* flags */
	offset = dissect_setup_action(tvb, pinfo, tree, offset);

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* OS */
	an = get_unicode_or_ascii_string_tvb(tvb, &offset,
		pinfo, &an_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_os, tvb,
		offset, an_len, an);
	offset += an_len;

	/* LANMAN */
	an = get_unicode_or_ascii_string_tvb(tvb, &offset,
		pinfo, &an_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_lanman, tvb,
		offset, an_len, an);
	offset += an_len;

	/* Primary domain */
	an = get_unicode_or_ascii_string_tvb(tvb, &offset,
		pinfo, &an_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_domain, tvb,
		offset, an_len, an);
	offset += an_len;

	/* call AndXCommand (if there are any) */
	call_next_command(tvb, pinfo, tree, andxoffset, smb_tree, cmd);

	return offset;
}

static int
dissect_tree_connect_andx_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint8	wc, cmd;
	guint16 andxoffset;
	int an_len;
	const char *an;

	/* word count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	/* next smb command */
	cmd = tvb_get_guint8(tvb, offset);
	if(cmd!=0xff){
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: %s (0x%02x)", decode_smb_name(cmd), cmd);
	} else {
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: No further commands");
	}
	offset += 1;

	/* reserved byte */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 1, tvb_get_ptr(tvb, offset, 1));
	offset += 1;

	/* andxoffset */
	andxoffset = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_andxoffset, tvb, offset, 2, andxoffset);
	offset += 2;

	if(wc==3){
		/* flags */
		offset = dissect_connect_support_bits(tvb, pinfo, tree, offset);
	}

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	if(wc==3){
		/* Service */
		an = get_unicode_or_ascii_string_tvb(tvb, &offset,
			pinfo, &an_len, FALSE, FALSE);
		proto_tree_add_string(tree, hf_smb_service, tvb,
			offset, an_len, an);
		offset += an_len;

		/* Native FS */
		an = get_unicode_or_ascii_string_tvb(tvb, &offset,
			pinfo, &an_len, FALSE, FALSE);
		proto_tree_add_string(tree, hf_smb_fs, tvb,
			offset, an_len, an);
		offset += an_len;
	}

	/* call AndXCommand (if there are any) */
	call_next_command(tvb, pinfo, tree, andxoffset, smb_tree, cmd);

	return offset;
}

static int
dissect_open_andx_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint8	wc, cmd;
	guint16 andxoffset;
	int fn_len;
	const char *fn;

	/* word count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	/* next smb command */
	cmd = tvb_get_guint8(tvb, offset);
	if(cmd!=0xff){
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: %s (0x%02x)", decode_smb_name(cmd), cmd);
	} else {
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: No further commands");
	}
	offset += 1;

	/* reserved byte */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 1, tvb_get_ptr(tvb, offset, 1));
	offset += 1;

	/* andxoffset */
	andxoffset = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_andxoffset, tvb, offset, 2, andxoffset);
	offset += 2;

	/* open flags */
	offset = dissect_open_flags(tvb, pinfo, tree, offset, 0x0007);

	/* desired access */
	offset = dissect_access(tvb, pinfo, tree, offset, "Desired");

	/* Search Attributes */
	offset = dissect_search_attributes(tvb, pinfo, tree, offset);

	/* File Attributes */
	offset = dissect_file_attributes(tvb, pinfo, tree, offset);

	/* creation time */
	offset = dissect_smb_UTIME(tvb, pinfo, tree, offset, hf_smb_create_time);
	
	/* open function */
	offset = dissect_open_function(tvb, pinfo, tree, offset);

	/* allocation size */
	proto_tree_add_uint(tree, hf_smb_alloc_size, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* 8 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 8, tvb_get_ptr(tvb, offset, 8));
	offset += 8;

	/* Byte Count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2,
		tvb_get_letohs(tvb, offset));
	offset += 2;

	/* Buffer Format */
	proto_tree_add_uint(tree, hf_smb_buffer_format, tvb, offset, 1,
		tvb_get_guint8(tvb, offset));
	offset += 1;

	/* file name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	/* call AndXCommand (if there are any) */
	call_next_command(tvb, pinfo, tree, andxoffset, smb_tree, cmd);

	return offset;
}

static int
dissect_open_andx_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint8	wc, cmd;
	guint16 andxoffset;

	/* word count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	/* next smb command */
	cmd = tvb_get_guint8(tvb, offset);
	if(cmd!=0xff){
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: %s (0x%02x)", decode_smb_name(cmd), cmd);
	} else {
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: No further commands");
	}
	offset += 1;

	/* reserved byte */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 1, tvb_get_ptr(tvb, offset, 1));
	offset += 1;

	/* andxoffset */
	andxoffset = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_andxoffset, tvb, offset, 2, andxoffset);
	offset += 2;

	/* fid */
	proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* File Attributes */
	offset = dissect_file_attributes(tvb, pinfo, tree, offset);

	/* last write time */
	offset = dissect_smb_UTIME(tvb, pinfo, tree, offset, hf_smb_last_write_time);
	
	/* File Size */
	proto_tree_add_uint(tree, hf_smb_file_size, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* granted access */
	offset = dissect_access(tvb, pinfo, tree, offset, "Granted");

	/* File Type */
	proto_tree_add_uint(tree, hf_smb_file_type, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* Device State */
	proto_tree_add_uint(tree, hf_smb_device_state, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* open_action */
	offset = dissect_open_action(tvb, pinfo, tree, offset);

	/* server fid */
	proto_tree_add_uint(tree, hf_smb_server_fid, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* 2 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 2, tvb_get_ptr(tvb, offset, 2));
	offset += 2;

	/* Byte Count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2,
		tvb_get_letohs(tvb, offset));
	offset += 2;

	/* call AndXCommand (if there are any) */
	call_next_command(tvb, pinfo, tree, andxoffset, smb_tree, cmd);

	return offset;
}

static int
dissect_read_andx_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint8	wc, cmd;
	guint16 andxoffset;

	/* word count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	/* next smb command */
	cmd = tvb_get_guint8(tvb, offset);
	if(cmd!=0xff){
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: %s (0x%02x)", decode_smb_name(cmd), cmd);
	} else {
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: No further commands");
	}
	offset += 1;

	/* reserved byte */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 1, tvb_get_ptr(tvb, offset, 1));
	offset += 1;

	/* andxoffset */
	andxoffset = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_andxoffset, tvb, offset, 2, andxoffset);
	offset += 2;

	/* fid */
	proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* offset */
	proto_tree_add_uint(tree, hf_smb_offset, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* max count */
	proto_tree_add_uint(tree, hf_smb_maxcount, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* min count */
	proto_tree_add_uint(tree, hf_smb_mincount, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* 4 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 4, tvb_get_ptr(tvb, offset, 4));
	offset += 4;

	/* remaining */
	proto_tree_add_uint(tree, hf_smb_remaining, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	if(wc==12){
		/* high offset */
		proto_tree_add_uint(tree, hf_smb_high_offset, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;
	}

	/* Byte Count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2,
		tvb_get_letohs(tvb, offset));
	offset += 2;

	/* call AndXCommand (if there are any) */
	call_next_command(tvb, pinfo, tree, andxoffset, smb_tree, cmd);

	return offset;
}

static int
dissect_read_andx_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint8	wc, cmd;
	guint16 andxoffset, bc, cnt;

	/* word count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	/* next smb command */
	cmd = tvb_get_guint8(tvb, offset);
	if(cmd!=0xff){
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: %s (0x%02x)", decode_smb_name(cmd), cmd);
	} else {
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: No further commands");
	}
	offset += 1;

	/* reserved byte */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 1, tvb_get_ptr(tvb, offset, 1));
	offset += 1;

	/* andxoffset */
	andxoffset = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_andxoffset, tvb, offset, 2, andxoffset);
	offset += 2;

	/* remaining */
	proto_tree_add_uint(tree, hf_smb_remaining, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* data compaction mode */
	proto_tree_add_uint(tree, hf_smb_dcm, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* 2 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 2, tvb_get_ptr(tvb, offset, 2));
	offset += 2;

	/* data len */
	cnt = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_data_len, tvb, offset, 2, cnt);
	offset += 2;

	/* data offset */
	proto_tree_add_uint(tree, hf_smb_data_offset, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* 10 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 10, tvb_get_ptr(tvb, offset, 10));
	offset += 10;

	/* byte count */
	bc = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, bc);
	offset += 2;

	/* file data */
	if(bc>cnt){
		/* We have some initial padding bytes.
		*/
		proto_tree_add_bytes(tree, hf_smb_padding, tvb, offset, bc-cnt,
			tvb_get_ptr(tvb, offset, bc-cnt));
		offset += bc-cnt;
		bc = cnt;
	}
	if(bc>tvb_length_remaining(tvb, offset)){
		int len;
		len = tvb_length_remaining(tvb, offset);
		proto_tree_add_bytes_format(tree, hf_smb_file_data, tvb, offset, len, tvb_get_ptr(tvb, offset, len),"File Data: Incomplete. Only %d of %d bytes", len, bc);
		offset += len;
	} else {
		proto_tree_add_bytes(tree, hf_smb_file_data, tvb, offset, bc, tvb_get_ptr(tvb, offset, bc));
		offset += bc;
	}

	/* call AndXCommand (if there are any) */
	call_next_command(tvb, pinfo, tree, andxoffset, smb_tree, cmd);

	return offset;
}

static int
dissect_empty_andx(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint8	wc, cmd;
	guint16 andxoffset;

	/* word count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	/* next smb command */
	cmd = tvb_get_guint8(tvb, offset);
	if(cmd!=0xff){
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: %s (0x%02x)", decode_smb_name(cmd), cmd);
	} else {
		proto_tree_add_uint_format(tree, hf_smb_cmd, tvb, offset, 1, cmd, "AndXCommand: No further commands");
	}
	offset += 1;

	/* reserved byte */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 1, tvb_get_ptr(tvb, offset, 1));
	offset += 1;

	/* andxoffset */
	andxoffset = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_andxoffset, tvb, offset, 2, andxoffset);
	offset += 2;

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* call AndXCommand (if there are any) */
	call_next_command(tvb, pinfo, tree, andxoffset, smb_tree, cmd);

	return offset;
}




/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  Transaction2 dissector
  XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/

static const value_string device_type_vals[] = {
	{0x00000001,	"Beep"},
	{0x00000002,	"CDROM"},
	{0x00000003,	"CDROM Filesystem"},
	{0x00000004,	"Controller"},
	{0x00000005,	"Datalink"},
	{0x00000006,	"DFS"},
	{0x00000007,	"Disk"},
	{0x00000008,	"Disk Filesystem"},
	{0x00000009,	"Filesystem"},
	{0x0000000a,	"Inport Port"},
	{0x0000000b,	"Keyboard"},
	{0x0000000c,	"Mailslot"},
	{0x0000000d,	"MIDI-In"},
	{0x0000000e,	"MIDI-Out"},
	{0x0000000f,	"Mouse"},
	{0x00000010,	"Multi UNC Provider"},
	{0x00000011,	"Named Pipe"},
	{0x00000012,	"Network"},
	{0x00000013,	"Network Browser"},
	{0x00000014,	"Network Filesystem"},
	{0x00000015,	"NULL"},
	{0x00000016,	"Parallel Port"},
	{0x00000017,	"Physical card"},
	{0x00000018,	"Printer"},
	{0x00000019,	"Scanner"},
	{0x0000001a,	"Serial Mouse port"},
	{0x0000001b,	"Serial port"},
	{0x0000001c,	"Screen"},
	{0x0000001d,	"Sound"},
	{0x0000001e,	"Streams"},
	{0x0000001f,	"Tape"},
	{0x00000020,	"Tape Filesystem"},
	{0x00000021,	"Transport"},
	{0x00000022,	"Unknown"},
	{0x00000023,	"Video"},
	{0x00000024,	"Virtual Disk"},
	{0x00000025,	"WAVE-In"},
	{0x00000026,	"WAVE-Out"},
	{0x00000027,	"8042 Port"},
	{0x00000028,	"Network Redirector"},
	{0x00000029,	"Battery"},
	{0x0000002a,	"Bus Extender"},
	{0x0000002b,	"Modem"},
	{0x0000002c,	"VDM"},
	{0,	NULL}
};


static const true_false_string tfs_dfs_referral_flags_strip = {
	"STRIP off pathconsumed characters before submitting",
	"Do NOT strip off any characters"
};
static int
dissect_dfs_referral_flags(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohs(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2,
			"Flags: 0x%04x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_dfs_referral_flags);
	}

	proto_tree_add_uint(tree, hf_smb_dfs_referral_flags_strip,
		tvb, offset, 2, mask);

	offset += 2;

	return offset;
}

static const true_false_string tfs_get_dfs_server_hold_storage = {
	"Referral SERVER HOLDS STORAGE for the file",
	"Referral server does NOT hold storage for the file"
};
static const true_false_string tfs_get_dfs_fielding = {
	"The server in referral is FIELDING CAPABLE",
	"The server in referrals is NOT fielding capable"
};

static int
dissect_get_dfs_flags(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohs(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2,
			"Flags: 0x%04x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_get_dfs_flags);
	}

	proto_tree_add_boolean(tree, hf_smb_get_dfs_server_hold_storage,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_get_dfs_fielding,
		tvb, offset, 2, mask);

	offset += 2;
	return offset;
}


static const true_false_string tfs_t2_mode_write_through = {
	"WRITE THROUGH requested",
	"Normal. No write through requested"
};
static const true_false_string tfs_t2_mode_sequential_only = {
	"SEQUENTIAL access ONLY",
	"Normal. Not mainly sequential access"
};
static const true_false_string tfs_t2_mode_sync_io_alert = {
	"SYNCHRONOUS access requested",
	"Normal. NOT sequential access requested"
};
static const true_false_string tfs_t2_mode_sync_io_nonalert = {
	"SYNCHRONOUS and BLOCKING access requested",
	"NO synchronous blocking mode requested"
};
static int
dissect_trans2_mode(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint32 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohl(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 4,
			"Mode: 0x%08x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_mode);
	}

	proto_tree_add_boolean(tree, hf_smb_t2_mode_write_through,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_mode_sequential_only,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_mode_sync_io_alert,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_mode_sync_io_nonalert,
		tvb, offset, 4, mask);

	offset += 4;
	return offset;
}

static const true_false_string tfs_ff2_backup = {
	"Find WITH backup intent",
	"No backup intent"
};
static const true_false_string tfs_ff2_continue = {
	"CONTINUE search from previous position",
	"New search, do NOT continue from previous position"
};
static const true_false_string tfs_ff2_resume = {
	"Return RESUME keys",
	"Do NOT return resume keys"
};
static const true_false_string tfs_ff2_close_eos = {
	"CLOSE search if END OF SEARCH is reached",
	"Do NOT close search if end of search reached"
};
static const true_false_string tfs_ff2_close = {
	"CLOSE search after this request",
	"Do NOT close search after this request"
};
static int
dissect_ff2_flags(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohs(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2,
			"Flags: 0x%04x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_find_first2_flags);
	}

	proto_tree_add_boolean(tree, hf_smb_ff2_backup,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_ff2_continue,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_ff2_resume,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_ff2_close_eos,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_ff2_close,
		tvb, offset, 2, mask);

	offset += 2;

	return offset;
}
static const true_false_string tfs_tf_dtid = {
	"Also DISCONNECT TID",
	"Do NOT disconnect tid"
};
static const true_false_string tfs_tf_owt = {
	"One Way Transaction (NO RESPONSE)",
	"Two way transaction"
};
static guint16
dissect_transaction_flags(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint16 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohs(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 2,
			"Flags: 0x%04x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_transaction_flags);
	}

	proto_tree_add_boolean(tree, hf_smb_transaction_flags_owt,
		tvb, offset, 2, mask);
	proto_tree_add_boolean(tree, hf_smb_transaction_flags_dtid,
		tvb, offset, 2, mask);

	return mask;
}

static const value_string dfs_referral_server_type_vals[] = {
	{0,	"Don't know"},
	{1,	"SMB Server"},
	{2,	"Netware Server"},
	{3,	"Domain Server"},
	{0, NULL}
};

/* dfs inconsistency data  (4.4.2)
*/
static int
dissect_dfs_inconsistency_data(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, int cnt)
{
	int fn_len;
	const char *fn;

	/*XXX shouldn this data hold version and size? unclear from doc*/
	/* referral version */
	proto_tree_add_uint(tree, hf_smb_dfs_referral_version,
		tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* referral size */
	proto_tree_add_uint(tree, hf_smb_dfs_referral_size,
		tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* referral server type */
	proto_tree_add_uint(tree, hf_smb_dfs_referral_server_type,
		tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* referral flags */
	offset = dissect_dfs_referral_flags(tvb, pinfo, tree, offset);

	/* node name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_dfs_referral_node, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	return offset;
}

/* get dfs referral data  (4.4.1)
*/
static int
dissect_get_dfs_referral_data(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, int cnt)
{
	guint16 numref;

	/* path consumed */
	proto_tree_add_uint(tree, hf_smb_dfs_path_consumed, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* num referrals */
	numref = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_dfs_path_consumed, tvb, offset, 2, numref);
	offset += 2;

	/* get dfs flags */
	offset = dissect_get_dfs_flags(tvb, pinfo, tree, offset);

	/* if there are any referrals */
	if(numref){
		proto_item *ref_item = NULL;
		proto_tree *ref_tree = NULL;
		int old_offset=offset;

		if(tree){
			ref_item = proto_tree_add_text(tree,
				tvb, offset, 0, "Referrals");
			ref_tree = proto_item_add_subtree(ref_item,
				ett_smb_dfs_referrals);
		}

		while(numref--){
			proto_item *ri = NULL;
			proto_tree *rt = NULL;
			int old_offset=offset;
			guint16 version;

			if(tree){
				ri = proto_tree_add_text(ref_tree,
					tvb, offset, 0, "Referral");
				rt = proto_item_add_subtree(ri,
					ett_smb_dfs_referral);
			}
		
			/* referral version */
			version = tvb_get_letohs(tvb, offset);
			proto_tree_add_uint(rt, hf_smb_dfs_referral_version,
				tvb, offset, 2, version);
			offset += 2;

			/* referral size */
			proto_tree_add_uint(rt, hf_smb_dfs_referral_size,
				tvb, offset, 2, tvb_get_letohs(tvb, offset));
			offset += 2;

			/* referral server type */
			proto_tree_add_uint(rt, hf_smb_dfs_referral_server_type,
				tvb, offset, 2, tvb_get_letohs(tvb, offset));
			offset += 2;

			/* referral flags */
			offset = dissect_dfs_referral_flags(tvb, pinfo, rt, offset);

			if(version==1){
				int fn_len;
				const char *fn;

				/* node name */
				fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
				proto_tree_add_string(rt, hf_smb_dfs_referral_node, tvb, offset, fn_len,
					fn);
				offset += fn_len;
			}

			if(version==2){
				int fn_len;
				const char *fn;

				/* proximity */
				proto_tree_add_uint(rt, hf_smb_dfs_referral_proximity,
					tvb, offset, 2, tvb_get_letohs(tvb, offset));
				offset += 2;

				/* ttl */
				proto_tree_add_uint(rt, hf_smb_dfs_referral_ttl,
					tvb, offset, 2, tvb_get_letohs(tvb, offset));
				offset += 2;

				/* path */
				fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
				proto_tree_add_string(rt, hf_smb_dfs_referral_path, tvb, offset, fn_len,
					fn);
				offset += fn_len;
			
				/* alt path */
				fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
				proto_tree_add_string(rt, hf_smb_dfs_referral_alt_path, tvb, offset, fn_len,
					fn);
				offset += fn_len;
			
				/* node */
				fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
				proto_tree_add_string(rt, hf_smb_dfs_referral_node, tvb, offset, fn_len,
					fn);
				offset += fn_len;
			}

			proto_item_set_len(ri, offset-old_offset);
		}
		proto_item_set_len(ref_item, offset-old_offset);
	}

	return offset;
}
				



static const value_string trans2_cmd_vals[] = {
	{ 0x00,		"TRANS2_OPEN2" },
	{ 0x01,		"TRANS2_FIND_FIRST2" },
	{ 0x02,		"TRANS2_FIND_NEXT2" },
	{ 0x03,		"TRANS2_QUERY_FS_INFORMATION" },
	{ 0x05,		"TRANS2_QUERY_PATH_INFORMATION" },
	{ 0x06,		"TRANS2_SET_PATH_INFORMATION" },
	{ 0x07,		"TRANS2_QUERY_FILE_INFORMATION" },
	{ 0x08,		"TRANS2_SET_FILE_INFORMATION" },
	{ 0x09,		"TRANS2_FSCTL" },
	{ 0x0A,		"TRANS2_IOCTL2" },
	{ 0x0B,		"TRANS2_FIND_NOTIFY_FIRST" },
	{ 0x0C,		"TRANS2_FIND_NOTIFY_NEXT" },
	{ 0x0D,		"TRANS2_CREATE_DIRECTORY" },
	{ 0x0E,		"TRANS2_SESSION_SETUP" },
	{ 0x10,		"TRANS2_GET_DFS_REFERRAL" },
	{ 0x11,		"TRANS2_REPORT_DFS_INCONSISTENCY" },
	{ 0,    NULL }
};

static const value_string ff2_il_vals[] = {
	{ 1,		"Info Standard"},
	{ 2,		"Info Query EA Size"},
	{ 3,		"Info Query EAs From List"},
	{ 0x0101,	"Find File Directory Info"},
	{ 0x0102,	"Find File Full Directory Info"},
	{ 0x0103,	"Find File Names Info"},
	{ 0x0104,	"Find File Both Directory Info"},
	{0, NULL}
};
static const value_string sfi_il_vals[] = {
	{ 1,		"Info Standard"},
	{ 2,		"Info Query EA Size"},
	{ 0x0101,	"File Basic Info"},
	{ 0x0102,	"File Disposition Info"},
	{ 0x0103,	"File Allocation Info"},
	{ 0x0104,	"File End-Of-File Info"},
	{0, NULL}
};




static const true_false_string tfs_t2_access_read_data = {
	"This file can be READ",
	"This file is NOT readable"
};
static const true_false_string tfs_t2_access_write_data = {
	"This file can be WRITTEN",
	"This file is not writable"
};
static const true_false_string tfs_t2_access_append_data = {
	"This file can be APPENDED",
	"This file does NOT support append"
};
static const true_false_string tfs_t2_access_read_ea = {
	"EA can be READ from this file",
	"Ea can NOT be read"
};
static const true_false_string tfs_t2_access_write_ea = {
	"EA can be WRITTEN for this file",
	"Ea can NOT be written for this file"
};
static const true_false_string tfs_t2_access_execute = {
	"This file is EXECUTABLE",
	"This file is NOT executable"
};
static const true_false_string tfs_t2_access_read_attributes = {
	"ATTRIBUTES can be READ",
	"Can NOT read attributes for this file"
};
static const true_false_string tfs_t2_access_write_attributes = {
	"ATTRIBUTES can be WRITTEN",
	"Can NOT write attributes for this file"
};
static const true_false_string tfs_t2_access_delete = {
	"This file can be DELETED",
	"This file can NOT be deleted"
};
static const true_false_string tfs_t2_access_read_control = {
	"Can READ ACL and ATTRIBUTES",
	"Can NOT read acl and attributes"
};
static const true_false_string tfs_t2_access_write_dac = {
	"Can WRITE ACL and ATTRIBUTES",
	"Can NOT write acl or attributes"
};
static const true_false_string tfs_t2_access_write_owner = {
	"Can WRITE OWNER (take ownership)",
	"Can NOT write owner (take ownership)"
};
static const true_false_string tfs_t2_access_synchronize = {
	"SYNCHRONOUS updates supported",
	"Synchronous updates NOT supported"
};

static int
dissect_trans2_access(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint32 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohl(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 4,
			"Access: 0x%08x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_desiredaccess);
	}

	proto_tree_add_boolean(tree, hf_smb_t2_access_read_data,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_access_write_data,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_access_append_data,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_access_read_ea,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_access_write_ea,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_access_execute,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_access_read_attributes,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_access_write_attributes,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_access_delete,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_access_read_control,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_access_write_dac,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_access_write_owner,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_t2_access_synchronize,
		tvb, offset, 4, mask);

	offset += 4;
	return offset;
}




/* this dissects the SMB_INFO_STANDARD and SMB_INFO_QUERY_EA_SIZE
   as described in 4.2.14.1
*/
static int
dissect_4_2_14_1(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, int cnt)
{
	int old_offset = offset;

	/* create time */
	offset = dissect_smb_timedate(tvb, pinfo, tree, offset,
		"Creation", hf_smb_create_time,
		hf_smb_create_dos_date, hf_smb_create_dos_time);

	/* access time */
	offset = dissect_smb_timedate(tvb, pinfo, tree, offset,
		"Last Access", hf_smb_access_time,
		hf_smb_access_dos_date, hf_smb_access_dos_time);

	/* last write time */
	offset = dissect_smb_timedate(tvb, pinfo, tree, offset,
		"Last Write", hf_smb_last_write_date,
		hf_smb_last_write_dos_date, hf_smb_last_write_dos_time);

	/* data size */
	proto_tree_add_uint(tree, hf_smb_data_size, tvb, offset, 4, tvb_get_letohs(tvb, offset));
	offset += 4;

	/* allocation size */
	proto_tree_add_uint(tree, hf_smb_alloc_size, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* File Attributes */
	offset = dissect_file_attributes(tvb, pinfo, tree, offset);

	if((offset-old_offset)<cnt){
		/* ea size */
		proto_tree_add_uint(tree, hf_smb_ea_size, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;
	}

	return offset;
}

/* this dissects the SMB_INFO_QUERY_EAS_FROM_LIST and SMB_INFO_QUERY_ALL_EAS
   as described in 4.2.14.2
*/
static int
dissect_4_2_14_2(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, int cnt)
{
	int old_offset = offset;

	/* list length */
	proto_tree_add_uint(tree, hf_smb_list_length, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* XXX dont know how to dissect EALists */

	return old_offset+cnt;
}

/* this dissects the SMB_INFO_IS_NAME_VALID
   as described in 4.2.14.3
*/
static int
dissect_4_2_14_3(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, int cnt)
{
	int fn_len;
	const char *fn;

	/* file name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	return offset;
}

/* this dissects the SMB_QUERY_FILE_BASIC_INFO
   as described in 4.2.14.4
*/
static int
dissect_4_2_14_4(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, int cnt)
{
	/* create time */
	offset = dissect_smb_64bit_time(tvb, pinfo, tree, offset,
		"Create Time", hf_smb_create_time);
	
	/* access time */
	offset = dissect_smb_64bit_time(tvb, pinfo, tree, offset,
		"Access Time", hf_smb_access_time);
	
	/* last write time */
	offset = dissect_smb_64bit_time(tvb, pinfo, tree, offset,
		"Write Time", hf_smb_last_write_date);
	
	/* last change time */
	offset = dissect_smb_64bit_time(tvb, pinfo, tree, offset,
		"Change Time", hf_smb_change_time);
	
	/* File Attributes */
	offset = dissect_file_attributes(tvb, pinfo, tree, offset);

	return offset;
}

/* this dissects the SMB_QUERY_FILE_STANDARD_INFO
   as described in 4.2.14.5
*/
static int
dissect_4_2_14_5(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, int cnt)
{
	/* allocation size */
	offset = dissect_smb_64bit_int(tvb, pinfo, tree, offset,
			hf_smb_alloc_size);

	/* end of file */
	offset = dissect_smb_64bit_int(tvb, pinfo, tree, offset,
			hf_smb_end_of_file);

	/* number of links */
	proto_tree_add_uint(tree, hf_smb_number_of_links, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* delete pending */
	proto_tree_add_uint(tree, hf_smb_delete_pending, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* is directory */
	proto_tree_add_uint(tree, hf_smb_is_directory, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	return offset;
}


/* this dissects the SMB_QUERY_FILE_EA_INFO
   as described in 4.2.14.6
*/
static int
dissect_4_2_14_6(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, int cnt)
{
	/* ea size */
	proto_tree_add_uint(tree, hf_smb_ea_size, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	return offset;
}

/* this dissects the SMB_QUERY_FILE_NAME_INFO
   as described in 4.2.14.7
   this is the same as SMB_QUERY_FILE_ALT_NAME_INFO
   as described in 4.2.14.9
*/
static int
dissect_4_2_14_7(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, int cnt)
{
	int fn_len;
	const char *fn;

	/* file name len */
	proto_tree_add_uint(tree, hf_smb_file_name_len, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;

	/* file name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	return offset;
}

/* this dissects the SMB_QUERY_FILE_ALL_INFO
   as described in 4.2.14.8
*/
static int
dissect_4_2_14_8(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, int cnt)
{

	offset = dissect_4_2_14_4(tvb, pinfo, tree, offset, cnt);
	offset = dissect_4_2_14_5(tvb, pinfo, tree, offset, cnt);

	/* index number */
	offset = dissect_smb_64bit_int(tvb, pinfo, tree, offset,
			hf_smb_index_number);

	offset = dissect_4_2_14_6(tvb, pinfo, tree, offset, cnt);

	/* access flags */
	offset = dissect_trans2_access(tvb, pinfo, tree, offset);

	/* index number */
	offset = dissect_smb_64bit_int(tvb, pinfo, tree, offset,
			hf_smb_index_number);

	/* current offset */
	offset = dissect_smb_64bit_int(tvb, pinfo, tree, offset,
			hf_smb_current_offset);

	/* mode */
	offset = dissect_trans2_mode(tvb, pinfo, tree, offset);

	/* alignment */
	proto_tree_add_uint(tree, hf_smb_t2_alignment, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;
	
	offset = dissect_4_2_14_6(tvb, pinfo, tree, offset, cnt);

	return offset;
}

/* this dissects the SMB_QUERY_FILE_STREAM_INFO
   as described in 4.2.14.10
*/
static int
dissect_4_2_14_10(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, int cnt)
{
	int fn_len;
	const char *fn;

	/* next entry offset */
	proto_tree_add_uint(tree, hf_smb_t2_next_entry_offset, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;
	
	/* stream name len */
	proto_tree_add_uint(tree, hf_smb_t2_stream_name_length, tvb, offset, 4, tvb_get_letohl(tvb, offset));
	offset += 4;
	
	/* stream size */
	offset = dissect_smb_64bit_int(tvb, pinfo, tree, offset,
			hf_smb_t2_stream_size);

	/* allocation size */
	offset = dissect_smb_64bit_int(tvb, pinfo, tree, offset,
			hf_smb_alloc_size);

	/* file name */
	fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
	proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
		fn);
	offset += fn_len;

	return offset;
}

/* this dissects the SMB_QUERY_FILE_COMPRESSION_INFO
   as described in 4.2.14.11
*/
static int
dissect_4_2_14_11(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, int cnt)
{
	/* compressed file size */
	offset = dissect_smb_64bit_int(tvb, pinfo, tree, offset,
			hf_smb_t2_compressed_file_size);

	/* compression format */
	proto_tree_add_uint(tree, hf_smb_t2_compressed_format, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;
	
	/* compression unit shift */
	proto_tree_add_uint(tree, hf_smb_t2_compressed_unit_shift, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;
	
	/* compression chunk shift */
	proto_tree_add_uint(tree, hf_smb_t2_compressed_chunk_shift, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;
	
	/* compression cluster shift */
	proto_tree_add_uint(tree, hf_smb_t2_compressed_cluster_shift, tvb, offset, 1, tvb_get_guint8(tvb, offset));
	offset += 1;
	
	/* 3 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 3, tvb_get_ptr(tvb, offset, 3));
	offset += 3;

	return offset;
}

/* values used by :
	TRANS2_QUERY_PATH_INFORMATION
	TRANS2_SET_PATH_INFORMATION
*/
static const value_string qpi_loi_vals[] = {
	{ 1,		"Info Standard  (4.2.14.1)"},
	{ 2,		"Info Query EA Size  (4.2.14.1)"},
	{ 3,		"Info Query EAs From List  (4.2.14.2)"},
	{ 4,		"Info Query All EAs  (4.2.14.2)"},
	{ 6,		"Info Is Name Valid  (4.2.14.3)"},
	{ 0x0101,	"Query File Basic Info  (4.2.14.4)"},
	{ 0x0102,	"Query File Standard Info  (4.2.14.5)"},
	{ 0x0103,	"Query File EA Info  (4.2.14.6)"},
	{ 0x0104,	"Query File Name Info  (4.2.14.7)"},
	{ 0x0107,	"Query File All Info  (4.2.14.8)"},
	{ 0x0108,	"Query File Alt File Info  (4.2.14.7)"},
	{ 0x0109,	"Query File Stream Info  (4.2.14.10)"},
	{ 0x010b,	"Query File Compression Info  (4.2.14.11)"},
	{ 0x0200,	"Set File Unix Basic"},
	{ 0x0201,	"Set File Unix Link"},
	{ 0x0202,	"Set File Unix HardLink"},
	{0, NULL}
};

/*dissect the data block for TRANS2_QUERY_PATH_INFORMATION*/
static int
dissect_qpi_loi_vals(tvbuff_t * tvb, packet_info * pinfo, proto_tree * tree, int offset, int cnt)
{
	smb_info *si;

	if(!cnt){
		return offset;
	}
	
	si = (smb_info *)pinfo->private;
	switch(si->info_level){
	case 1:		/*Info Standard*/
	case 2:		/*Info Query EA Size*/
		offset = dissect_4_2_14_1(tvb, pinfo, tree, offset, cnt);
		break;
	case 3:		/*Info Query EAs From List*/
	case 4:		/*Info Query All EAs*/
		offset = dissect_4_2_14_2(tvb, pinfo, tree, offset, cnt);
		break;
	case 6:		/*Info Is Name Valid*/
		offset = dissect_4_2_14_3(tvb, pinfo, tree, offset, cnt);
		break;
	case 0x0101:	/*Query File Basic Info*/
		offset = dissect_4_2_14_4(tvb, pinfo, tree, offset, cnt);
		break;
	case 0x0102:	/*Query File Standard Info*/
		offset = dissect_4_2_14_5(tvb, pinfo, tree, offset, cnt);
		break;
	case 0x0103:	/*Query File EA Info*/
		offset = dissect_4_2_14_6(tvb, pinfo, tree, offset, cnt);
		break;
	case 0x0104:	/*Query File Name Info*/
		offset = dissect_4_2_14_7(tvb, pinfo, tree, offset, cnt);
		break;
	case 0x0107:	/*Query File All Info*/
		offset = dissect_4_2_14_8(tvb, pinfo, tree, offset, cnt);
		break;
	case 0x0108:	/*Query File Alt File Info*/
		offset = dissect_4_2_14_7(tvb, pinfo, tree, offset, cnt);
		break;
	case 0x0109:	/*Query File Stream Info*/
		offset = dissect_4_2_14_10(tvb, pinfo, tree, offset, cnt);
		break;
	case 0x010b:	/*Query File Compression Info*/
		offset = dissect_4_2_14_11(tvb, pinfo, tree, offset, cnt);
		break;
	case 0x0200:	/*Set File Unix Basic*/
		/* XXX add this from the SNIA doc */
		break;
	case 0x0201:	/*Set File Unix Link*/
		/* XXX add this from the SNIA doc */
		break;
	case 0x0202:	/*Set File Unix HardLink*/
		/* XXX add this from the SNIA doc */
		break;
	}
	
	return offset;
}


static const true_false_string tfs_fs_attr_css = {
	"This FS supports CASE SENSITIVE SEARCHes",
	"This FS does NOT support case sensitive searches"
};
static const true_false_string tfs_fs_attr_cpn = {
	"This FS supports CASE PRESERVED NAMES",
	"This FS does NOT support case preserved names"
};
static const true_false_string tfs_fs_attr_pacls = {
	"This FS supports PERSISTENT ACLs",
	"This FS does NOT support persistent acls"
};
static const true_false_string tfs_fs_attr_fc = {
	"This FS supports COMPRESSED FILES",
	"This FS does NOT support compressed files"
};
static const true_false_string tfs_fs_attr_vq = {
	"This FS supports VOLUME QUOTAS",
	"This FS does NOT support volume quotas"
};
static const true_false_string tfs_fs_attr_dim = {
	"This FS is on a MOUNTED DEVICE",
	"This FS is NOT on a mounted device"
};
static const true_false_string tfs_fs_attr_vic = {
	"This FS is on a COMPRESSED VOLUME",
	"This FS is NOT on a compressed volume"
};
static int
dissect_fs_attributes(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint32 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohl(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 4,
			"FS Attributes: 0x%08x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_fs_attributes);
	}

	proto_tree_add_boolean(tree, hf_smb_fs_attr_css,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_fs_attr_cpn,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_fs_attr_pacls,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_fs_attr_fc,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_fs_attr_vq,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_fs_attr_dim,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_fs_attr_vic,
		tvb, offset, 4, mask);

	offset += 4;
	return offset;
}

static const true_false_string tfs_device_char_removable = {
	"This is a REMOVABLE device",
	"This is NOT a removable device"
};
static const true_false_string tfs_device_char_read_only = {
	"This is a READ-ONLY device",
	"This is NOT a read-only device"
};
static const true_false_string tfs_device_char_floppy = {
	"This is a FLOPPY DISK device",
	"This is NOT a floppy disk device"
};
static const true_false_string tfs_device_char_write_once = {
	"This is a WRITE-ONCE device",
	"This is NOT a write-once device"
};
static const true_false_string tfs_device_char_remote = {
	"This is a REMOTE device",
	"This is NOT a remote device"
};
static const true_false_string tfs_device_char_mounted = {
	"This device is MOUNTED",
	"This device is NOT mounted"
};
static const true_false_string tfs_device_char_virtual = {
	"This is a VIRTUAL device",
	"This is NOT a virtual device"
};
static int
dissect_device_characteristics(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset)
{
	guint32 mask;
	proto_item *item = NULL;
	proto_tree *tree = NULL;

	mask = tvb_get_letohl(tvb, offset);

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, 4,
			"Device Characteristics: 0x%08x ", mask);
		tree = proto_item_add_subtree(item, ett_smb_device_characteristics);
	}

	proto_tree_add_boolean(tree, hf_smb_device_char_removable,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_device_char_read_only,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_device_char_floppy,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_device_char_write_once,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_device_char_remote,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_device_char_mounted,
		tvb, offset, 4, mask);
	proto_tree_add_boolean(tree, hf_smb_device_char_virtual,
		tvb, offset, 4, mask);

	offset += 4;
	return offset;
}


static const value_string qfsi_vals[] = {
	{ 1,		"Info Allocation"},
	{ 2,		"Info Volume"},
	{ 0x0102,	"Query FS Volume Info"},
	{ 0x0103,	"Query FS Size Info"},
	{ 0x0104,	"Query FS Device Info"},
	{ 0x0105,	"Query FS Attribute Info"},
	{0, NULL}
};
/*dissect the data block for TRANS2_QUERY_FS_INFORMATION*/
static int
dissect_qfsi_vals(tvbuff_t * tvb, packet_info * pinfo, proto_tree * tree, int offset, int cnt)
{
	smb_info *si;
	int fn_len, vll, fnl;
	const char *fn;

	if(!cnt){
		return offset;
	}
	
	si = (smb_info *)pinfo->private;
	switch(si->info_level){
	case 1:		/* SMB_INFO_ALLOCATION */
		/* filesystem id */
		proto_tree_add_uint(tree, hf_smb_fs_id, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* sectors per unit */
		proto_tree_add_uint(tree, hf_smb_sector_unit, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* units */
		proto_tree_add_uint(tree, hf_smb_fs_units, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* bytes per sector */
		proto_tree_add_uint(tree, hf_smb_fs_sector, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		break;
	case 2:		/* SMB_INFO_VOLUME */
		/* volume serial number */
		proto_tree_add_uint(tree, hf_smb_volume_serial_num, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* volume label length */
		proto_tree_add_uint(tree, hf_smb_volume_label_len, tvb, offset, 1, tvb_get_guint8(tvb, offset));
		offset += 1;

		/* label */
		fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
		proto_tree_add_string(tree, hf_smb_volume_label, tvb, offset, fn_len,
			fn);
		offset += fn_len;

		break;
	case 0x0102:	/* SMB_QUERY_FS_VOLUME_INFO */
		/* create time */
		offset = dissect_smb_64bit_time(tvb, pinfo, tree, offset,
			"Create Time", hf_smb_create_time);
	
		/* volume serial number */
		proto_tree_add_uint(tree, hf_smb_volume_serial_num, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* volume label length */
		vll = tvb_get_letohl(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_volume_label_len, tvb, offset, 4, vll);
		offset += 4;

		/* 2 reserved bytes */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 2, tvb_get_ptr(tvb, offset, 2));
		offset += 2;

		/* label */
		fn_len = vll;
		fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, TRUE);
		proto_tree_add_string(tree, hf_smb_volume_label, tvb, offset, fn_len,
			fn);
		offset += fn_len;

		break;
	case 0x0103:	/* SMB_QUERY_FS_SIZE_INFO */
		/* allocation size */
		offset = dissect_smb_64bit_int(tvb, pinfo, tree, offset,
				hf_smb_alloc_size);

		/* free allocation units */
		offset = dissect_smb_64bit_int(tvb, pinfo, tree, offset,
				hf_smb_free_alloc_units);

		/* sectors per unit */
		proto_tree_add_uint(tree, hf_smb_sector_unit, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* bytes per sector */
		proto_tree_add_uint(tree, hf_smb_fs_sector, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		break;
	case 0x0104:	/* SMB_QUERY_FS_DEVICE_INFO */
		/* device type */
		proto_tree_add_uint(tree, hf_smb_device_type, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* device characteristics */
		offset = dissect_device_characteristics(tvb, pinfo, tree, offset);
	
		break;
	case 0x0105:	/* SMB_QUERY_FS_ATTRIBUTE_INFO */
		/* device characteristics */
		offset = dissect_fs_attributes(tvb, pinfo, tree, offset);
	
		/* max name len */
		proto_tree_add_uint(tree, hf_smb_max_name_len, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* fs name length */
		fnl = tvb_get_letohl(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_fs_name_len, tvb, offset, 4, fnl);
		offset += 4;

		/* label */
		fn_len = fnl;
		fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, TRUE);
		proto_tree_add_string(tree, hf_smb_fs_name, tvb, offset, fn_len,
			fn);
		offset += fn_len;

		break;
	}

	return offset;
}


static int
dissect_transaction2_request_data(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset, int dc)
{
	proto_item *item = NULL;
	proto_tree *tree = NULL;
	smb_info *si;
	int fn_len;
	const char *fn;
	int old_offset = offset;

	si = (smb_info *)pinfo->private;

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, dc, "Data");
		tree = proto_item_add_subtree(item, ett_smb_transaction_data);
	}

	switch(si->subcmd){
	case 0x00:	/*TRANS2_OPEN2*/
		/* XXX FAEList here?*/
		break;
	case 0x01:	/*TRANS2_FIND_FIRST2*/
		/* XXX FAEList here?*/
		break;
	case 0x02:	/*TRANS2_FIND_NEXT2*/
		/* no data field in this request */
		break;
	case 0x03:	/*TRANS2_QUERY_FS_INFORMATION*/
		/* no data field in this request */
		break;
	case 0x05:	/*TRANS2_QUERY_PATH_INFORMATION*/
		offset = dissect_qpi_loi_vals(tvb, pinfo, tree, offset, dc);
		break;
	case 0x06:	/*TRANS2_SET_PATH_INFORMATION*/
		offset = dissect_qpi_loi_vals(tvb, pinfo, tree, offset, dc);
		break;
	case 0x07:	/*TRANS2_QUERY_FILE_INFORMATION*/
		/* no data field in this request */
		break;
	case 0x08:	/*TRANS2_SET_FILE_INFORMATION*/
		offset = dissect_qpi_loi_vals(tvb, pinfo, tree, offset, dc);
		break;
	case 0x09:	/*TRANS2_FSCTL*/
		/*XXX dont know how to decode this yet */
		break;
	case 0x0a:	/*TRANS2_IOCTL2*/
		/*XXX dont know how to decode this yet */
		break;
	case 0x0b:	/*TRANS2_FIND_NOTIFY_FIRST*/
		/*XXX dont know how to decode this yet */
		break;
	case 0x0c:	/*TRANS2_FIND_NOTIFY_NEXT*/
		/*XXX dont know how to decode this yet */
		break;
	case 0x0d:	/*TRANS2_CREATE_DIRECTORY*/
		/* no data block for this one */
		break;
	case 0x0e:	/*TRANS2_SESSION_SETUP*/
		/*XXX dont know how to decode this yet */
		break;
	case 0x10:	/*TRANS2_GET_DFS_REFERRAL*/
		offset = dissect_get_dfs_referral_data(tvb, pinfo, tree, offset, dc);
		break;
	case 0x11:	/*TRANS2_REPORT_DFS_INCONSISTENCY*/
		offset = dissect_dfs_inconsistency_data(tvb, pinfo, tree, offset, dc);
		break;
	}

	/* ooops there were data we didnt know how to process */
	if((offset-old_offset)<dc){
		proto_tree_add_bytes(tree, hf_smb_unknown, tvb, offset, dc-(offset-old_offset), tvb_get_ptr(tvb, offset, 1));
		offset += dc-(offset-old_offset);
	}

	return offset;
}

static int
dissect_transaction2_response_data(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset, int dc)
{
	proto_item *item = NULL;
	proto_tree *tree = NULL;
	smb_info *si;
	int fn_len;
	const char *fn;
	int old_offset = offset;

	si = (smb_info *)pinfo->private;

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, dc, "Data");
		tree = proto_item_add_subtree(item, ett_smb_transaction_data);
	}

	switch(si->subcmd){
	case 0x00:	/*TRANS2_OPEN2*/
		/* XXX not implemented yet. See SNIA doc */
		break;
	case 0x01:	/*TRANS2_FIND_FIRST2*/
		/* XXX not implemented yet. See SNIA doc */
		break;
	case 0x02:	/*TRANS2_FIND_NEXT2*/
		/* XXX not implemented yet. See SNIA doc */
		break;
	case 0x03:	/*TRANS2_QUERY_FS_INFORMATION*/
		offset = dissect_qfsi_vals(tvb, pinfo, tree, offset, dc);
		break;
	case 0x05:	/*TRANS2_QUERY_PATH_INFORMATION*/
		offset = dissect_qpi_loi_vals(tvb, pinfo, tree, offset, dc);
		break;
	case 0x06:	/*TRANS2_SET_PATH_INFORMATION*/
		/* no data in this response */
		break;
	case 0x07:	/*TRANS2_QUERY_FILE_INFORMATION*/
		/* identical to QUERY_PATH_INFO */
		offset = dissect_qpi_loi_vals(tvb, pinfo, tree, offset, dc);
		break;
	case 0x08:	/*TRANS2_SET_FILE_INFORMATION*/
		/* no data in this response */
		break;
	case 0x09:	/*TRANS2_FSCTL*/
		/* XXX dont know how to dissect this one (yet)*/
		break;
	case 0x0a:	/*TRANS2_IOCTL2*/
		/* XXX dont know how to dissect this one (yet)*/
		break;
	case 0x0b:	/*TRANS2_FIND_NOTIFY_FIRST*/
		/* XXX dont know how to dissect this one (yet)*/
		break;
	case 0x0c:	/*TRANS2_FIND_NOTIFY_NEXT*/
		/* XXX dont know how to dissect this one (yet)*/
		break;
	case 0x0d:	/*TRANS2_CREATE_DIRECTORY*/
		/* no data in this response */
		break;
	case 0x0e:	/*TRANS2_SESSION_SETUP*/
		/* XXX dont know how to dissect this one (yet)*/
		break;
	case 0x10:	/*TRANS2_GET_DFS_REFERRAL*/
		/* XXX dont know how to dissect this one (yet) see SNIA doc*/
		break;
	case 0x11:	/*TRANS2_REPORT_DFS_INCONSISTENCY*/
		/* XXX dont know how to dissect this one (yet) see SNIA doc*/
		break;
	}

	/* ooops there were data we didnt know how to process */
	if((offset-old_offset)<dc){
		proto_tree_add_bytes(tree, hf_smb_unknown, tvb, offset, dc-(offset-old_offset), tvb_get_ptr(tvb, offset, 1));
		offset += dc-(offset-old_offset);
	}

	return offset;
}

static int
dissect_transaction2_response_parameters(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset, int pc, int od)
{
	proto_item *item = NULL;
	proto_tree *tree = NULL;
	smb_info *si;
	int fn_len, lno;
	const char *fn;
	int old_offset = offset;

	si = (smb_info *)pinfo->private;

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, pc, "Parameters");
		tree = proto_item_add_subtree(item, ett_smb_transaction_params);
	}

	switch(si->subcmd){
	case 0x00:	/*TRANS2_OPEN2*/
		/* fid */
		proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* File Attributes */
		offset = dissect_file_attributes(tvb, pinfo, tree, offset);

		/* create time */
		offset = dissect_smb_timedate(tvb, pinfo, tree, offset,
			"Creation", hf_smb_create_time,
			hf_smb_create_dos_date, hf_smb_create_dos_time);

		/* data size */
		proto_tree_add_uint(tree, hf_smb_data_size, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* granted access */
		offset = dissect_access(tvb, pinfo, tree, offset, "Granted");

		/* File Type */
		proto_tree_add_uint(tree, hf_smb_file_type, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* Device State */
		proto_tree_add_uint(tree, hf_smb_device_state, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* open_action */
		offset = dissect_open_action(tvb, pinfo, tree, offset);

		/* 4 reserved bytes */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 4, tvb_get_ptr(tvb, offset, 4));
		offset += 4;

		/* ea error offset */
		proto_tree_add_uint(tree, hf_smb_ea_error_offset, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* ea length */
		proto_tree_add_uint(tree, hf_smb_ea_length, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		break;
	case 0x01:	/*TRANS2_FIND_FIRST2*/
		/* sid */
		proto_tree_add_uint(tree, hf_smb_sid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* search count */
		proto_tree_add_uint(tree, hf_smb_search_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* end of search */
		proto_tree_add_uint(tree, hf_smb_end_of_search, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* ea error offset */
		proto_tree_add_uint(tree, hf_smb_ea_error_offset, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* last name offset */
		lno = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_last_name_offset, tvb, offset, 2, lno);
		offset += 2;

#ifdef BROKEN
		/* XXX this is broken */
		if(lno){
			/* file name */
			fn = get_unicode_or_ascii_string_tvb(tvb, &lno, pinfo, &fn_len, FALSE, FALSE);
			proto_tree_add_string(tree, hf_smb_file_name, tvb, lno, fn_len,
				fn);
			offset += fn_len;
		}
#endif

		break;
	case 0x02:	/*TRANS2_FIND_NEXT2*/
		/* search count */
		proto_tree_add_uint(tree, hf_smb_search_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* end of search */
		proto_tree_add_uint(tree, hf_smb_end_of_search, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* ea error offset */
		proto_tree_add_uint(tree, hf_smb_ea_error_offset, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* last name offset */
		lno = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_last_name_offset, tvb, offset, 2, lno);
		offset += 2;

#ifdef BROKEN
		/* XXX this is broken */
		if(lno){
			/* file name */
			fn = get_unicode_or_ascii_string_tvb(tvb, &lno, pinfo, &fn_len, FALSE, FALSE);
			proto_tree_add_string(tree, hf_smb_file_name, tvb, lno, fn_len,
				fn);
			offset += fn_len;
		}
#endif

		break;
	case 0x03:	/*TRANS2_QUERY_FS_INFORMATION*/
		/* no parameter block here */
		break;
	case 0x05:	/*TRANS2_QUERY_PATH_INFORMATION*/
		/* no parameter block here */
		break;
	case 0x06:	/*TRANS2_SET_PATH_INFORMATION*/
		/* no parameter block here */
		break;
	case 0x07:	/*TRANS2_QUERY_FILE_INFORMATION*/
		/* no parameter block here */
		break;
	case 0x08:	/*TRANS2_SET_FILE_INFORMATION*/
		/* no parameter block here */
		break;
	case 0x09:	/*TRANS2_FSCTL*/
		/* XXX dont know how to dissect this one (yet)*/
		break;
	case 0x0a:	/*TRANS2_IOCTL2*/
		/* XXX dont know how to dissect this one (yet)*/
		break;
	case 0x0b:	/*TRANS2_FIND_NOTIFY_FIRST*/
		/* XXX dont know how to dissect this one (yet)*/
		break;
	case 0x0c:	/*TRANS2_FIND_NOTIFY_NEXT*/
		/* XXX dont know how to dissect this one (yet)*/
		break;
	case 0x0d:	/*TRANS2_CREATE_DIRECTORY*/
		/* ea error offset */
		proto_tree_add_uint(tree, hf_smb_ea_error_offset, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		break;
	case 0x0e:	/*TRANS2_SESSION_SETUP*/
		/* XXX dont know how to dissect this one (yet)*/
		break;
	case 0x10:	/*TRANS2_GET_DFS_REFERRAL*/
		/* XXX dont know how to dissect this one (yet) see SNIA doc*/
		break;
	case 0x11:	/*TRANS2_REPORT_DFS_INCONSISTENCY*/
		/* XXX dont know how to dissect this one (yet) see SNIA doc*/
		break;
		break;
	}

	/* ooops there were data we didnt know how to process */
	if((offset-old_offset)<pc){
		proto_tree_add_bytes(tree, hf_smb_unknown, tvb, offset, pc-(offset-old_offset), tvb_get_ptr(tvb, offset, 1));
		offset += pc-(offset-old_offset);
	}

	return offset;
}

static int
dissect_transaction2_request_parameters(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree, int offset, int pc)
{
	proto_item *item = NULL;
	proto_tree *tree = NULL;
	smb_info *si;
	int fn_len;
	const char *fn;
	int old_offset = offset;

	si = (smb_info *)pinfo->private;

	if(parent_tree){
		item = proto_tree_add_text(parent_tree, tvb, offset, pc, "Parameters");
		tree = proto_item_add_subtree(item, ett_smb_transaction_params);
	}

	switch(si->subcmd){
	case 0x00:	/*TRANS2_OPEN2*/
		/* open flags */
		offset = dissect_open_flags(tvb, pinfo, tree, offset, 0x000f);

		/* desired access */
		offset = dissect_access(tvb, pinfo, tree, offset, "Desired");

		/* 2 reserved bytes */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 2, tvb_get_ptr(tvb, offset, 2));
		offset += 2;

		/* File Attributes */
		offset = dissect_file_attributes(tvb, pinfo, tree, offset);

		/* create time */
		offset = dissect_smb_timedate(tvb, pinfo, tree, offset,
			"Creation", hf_smb_create_time,
			hf_smb_create_dos_date, hf_smb_create_dos_time);

		/* open function */
		offset = dissect_open_function(tvb, pinfo, tree, offset);

		/* allocation size */
		proto_tree_add_uint(tree, hf_smb_alloc_size, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* 10 reserved bytes */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 10, tvb_get_ptr(tvb, offset, 10));
		offset += 10;

		/* file name */
		fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
		proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
			fn);
		offset += fn_len;

		/* XXX dont know how to decode FEAList */
		break;
	case 0x01:	/*TRANS2_FIND_FIRST2*/
		/* Search Attributes */
		offset = dissect_search_attributes(tvb, pinfo, tree, offset);

		/* search count */
		proto_tree_add_uint(tree, hf_smb_search_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* Find First2 flags */
		offset = dissect_ff2_flags(tvb, pinfo, tree, offset);

		/* Find First2 information level */
		proto_tree_add_uint(tree, hf_smb_ff2_information_level, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* storage type */
		proto_tree_add_uint(tree, hf_smb_storage_type, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* search pattern */
		fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
		proto_tree_add_string(tree, hf_smb_search_pattern, tvb, offset, fn_len,
			fn);
		offset += fn_len;

		/* XXX dont know how to decode FEAList */

		break;
	case 0x02:	/*TRANS2_FIND_NEXT2*/
		/* sid */
		proto_tree_add_uint(tree, hf_smb_sid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* search count */
		proto_tree_add_uint(tree, hf_smb_search_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* Find First2 information level */
		proto_tree_add_uint(tree, hf_smb_ff2_information_level, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* resume key */
		proto_tree_add_uint(tree, hf_smb_resume, tvb, offset, 4, tvb_get_letohl(tvb, offset));
		offset += 4;

		/* Find First2 flags */
		offset = dissect_ff2_flags(tvb, pinfo, tree, offset);

		/* file name */
		fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
		proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
			fn);
		offset += fn_len;

		break;
	case 0x03:	/*TRANS2_QUERY_FS_INFORMATION*/
		/* level of interest */
		si->info_level = tvb_get_letohs(tvb, offset);

		/* trans2 query fs information : information level */
		proto_tree_add_uint(tree, hf_smb_qfsi_information_level, tvb, offset, 2, si->info_level);
		offset += 2;

		break;
	case 0x05:	/*TRANS2_QUERY_PATH_INFORMATION*/
		/* level of interest */
		si->info_level = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_qpi_loi, tvb, offset, 2, si->info_level);
		offset += 2;
		
		/* 4 reserved bytes */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 4, tvb_get_ptr(tvb, offset, 4));
		offset += 4;

		/* file name */
		fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
		proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
			fn);
		offset += fn_len;
		break;
	case 0x06:	/*TRANS2_SET_PATH_INFORMATION*/
		/* level of interest */
		si->info_level = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_qpi_loi, tvb, offset, 2, si->info_level);
		offset += 2;
		
		/* 4 reserved bytes */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 4, tvb_get_ptr(tvb, offset, 4));
		offset += 4;

		/* file name */
		fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
		proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
			fn);
		offset += fn_len;

		break;
	case 0x07:	/*TRANS2_QUERY_FILE_INFORMATION*/
		/* fid */
		proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* level of interest */
		si->info_level = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_qpi_loi, tvb, offset, 2, si->info_level);
		offset += 2;
		
		break;
	case 0x08:	/*TRANS2_SET_FILE_INFORMATION*/
		/* fid */
		proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* level of interest */
		si->info_level = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_qpi_loi, tvb, offset, 2, si->info_level);
		offset += 2;
		
		/* 2 reserved bytes */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 2, tvb_get_ptr(tvb, offset, 2));
		offset += 2;

		break;
	case 0x09:	/*TRANS2_FSCTL*/
	case 0x0a:	/*TRANS2_IOCTL2*/
		/* these calls have no parameter block in the request */
		break;
	case 0x0b:	/*TRANS2_FIND_NOTIFY_FIRST*/
	case 0x0c:	/*TRANS2_FIND_NOTIFY_NEXT*/
		/* XXX unknown structure*/
		break;
	case 0x0d:	/*TRANS2_CREATE_DIRECTORY*/
		/* 4 reserved bytes */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 4, tvb_get_ptr(tvb, offset, 4));
		offset += 4;

		/* dir name */
		fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len,
			FALSE, FALSE);
		proto_tree_add_string(tree, hf_smb_dir_name, tvb, offset, fn_len,
			fn);
		offset += fn_len;

		/* XXX optional FEAList, unknown what FEAList looks like*/
		break;
	case 0x0e:	/*TRANS2_SESSION_SETUP*/
		/* XXX unknown structure*/
		break;
	case 0x10:	/*TRANS2_GET_DFS_REFERRAL*/
		/* referral level */
		proto_tree_add_uint(tree, hf_smb_max_referral_level, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;
		
		/* file name */
		fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
		proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
			fn);
		offset += fn_len;

		break;
	case 0x11:	/*TRANS2_REPORT_DFS_INCONSISTENCY*/
		/* file name */
		fn = get_unicode_or_ascii_string_tvb(tvb, &offset, pinfo, &fn_len, FALSE, FALSE);
		proto_tree_add_string(tree, hf_smb_file_name, tvb, offset, fn_len,
			fn);
		offset += fn_len;
		break;
	}

	/* ooops there were data we didnt know how to process */
	if((offset-old_offset)<pc){
		proto_tree_add_bytes(tree, hf_smb_unknown, tvb, offset, pc-(offset-old_offset), tvb_get_ptr(tvb, offset, 1));
		offset += pc-(offset-old_offset);
	}

	return offset;
}


static int
dissect_transaction_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint8 sc=0;
	int so=0;
	guint16 wc, od, tf, po, pc, dc, pd, dd;
	guint32 to;
	int an_len;
	const char *an = NULL;
	smb_info *si;

	si = (smb_info *)pinfo->private;

	/* word count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	if(wc==8){
	/*secondary client request*/
		/* total param count */
		proto_tree_add_uint(tree, hf_smb_total_param_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;
	
		/* total data count */
		proto_tree_add_uint(tree, hf_smb_total_data_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* param count */
		pc = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_param_count, tvb, offset, 2, pc);
		offset += 2;

		/* param offset */
		po = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_param_offset, tvb, offset, 2, po);
		offset += 2;

		/* param disp */
		pd = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_param_disp, tvb, offset, 2, pd);
		offset += 2;
	
		/* data count */
		dc = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_data_count, tvb, offset, 2, dc);
		offset += 2;

		/* data offset */
		od = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_data_offset, tvb, offset, 2, od);
		offset += 2;
	
		/* data disp */
		dd = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_data_disp, tvb, offset, 2, dd);
		offset += 2;

		if(si->cmd==0x32){
			/* fid */
			proto_tree_add_uint(tree, hf_smb_fid, tvb, offset, 2, tvb_get_letohs(tvb, offset));
			offset += 2;
		}

		/* byte count */
		proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

	} else {
	/* it is not a secondary request */
		/* total param count */
		proto_tree_add_uint(tree, hf_smb_total_param_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* total data count */
		proto_tree_add_uint(tree, hf_smb_total_data_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* max param count */
		proto_tree_add_uint(tree, hf_smb_max_param_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* max data count */
		proto_tree_add_uint(tree, hf_smb_max_data_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		/* max setup count */
		proto_tree_add_uint(tree, hf_smb_max_setup_count, tvb, offset, 1, tvb_get_guint8(tvb, offset));
		offset += 1;

		/* reserved byte */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 1, tvb_get_ptr(tvb, offset, 1));
		offset += 1;

		/* transaction flags */
		tf = dissect_transaction_flags(tvb, pinfo, tree, offset);
		offset += 2;

		/* timeout */
		to = tvb_get_letohl(tvb, offset);
		proto_tree_add_uint_format(tree, hf_smb_timeout, tvb, offset, 4, to, "Timeout: %d.%03d", to/1000, to%1000);
		offset += 4;

		/* 2 reserved bytes */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 2, tvb_get_ptr(tvb, offset, 2));
		offset += 2;

		/* param count */
		pc = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_param_count, tvb, offset, 2, pc);
		offset += 2;
	
		/* param offset */
		po = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_param_offset, tvb, offset, 2, po);
		offset += 2;

		/* data count */
		dc = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_data_count, tvb, offset, 2, dc);
		offset += 2;

		/* data offset */
		od = tvb_get_letohs(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_data_offset, tvb, offset, 2, od);
		offset += 2;

		/* data displacement is zero here */
		dd = 0;

		/* setup count */
		sc = tvb_get_guint8(tvb, offset);
		proto_tree_add_uint(tree, hf_smb_setup_count, tvb, offset, 1, sc);
		offset += 1;

		/* reserved byte */
		proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 1, tvb_get_ptr(tvb, offset, 1));
		offset += 1;
		
		/* this is where the setup bytes start */	
		so = offset;

		/* if there were any setup bytes, decode them */
		if(sc){
			/* TRANSACTION2 only has one setup word and that is the
			   subcommand code */
			if(si->cmd==0x32){
				/* subcommand */
				si->subcmd = tvb_get_letohs(tvb, offset);
				proto_tree_add_uint(tree, hf_smb_trans2_subcmd, tvb, offset, 2, si->subcmd);

				if (check_col(pinfo->fd, COL_INFO)) {
					col_add_fstr(pinfo->fd, COL_INFO, "%s %s Request",
					decode_smb_name(si->cmd), 
					val_to_str(si->subcmd, trans2_cmd_vals, 
						"Unknown (0x%02X)"));
				}
			}

			if(si->cmd==0x25){
				/*XXX decode the TRANSACTION setup words */
			}

			offset += sc*2;
		}

		/* byte count */
		proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;
	
		/* name is NULL if transaction2 */
		if(si->cmd!=0x32){
			/* Transaction Name */
			an = get_unicode_or_ascii_string_tvb(tvb, &offset,
				pinfo, &an_len, FALSE, FALSE);
			proto_tree_add_string(tree, hf_smb_trans_name, tvb,
				offset, an_len, an);
			offset += an_len;
		}
	}

	/* parameters */
	if(po>offset){
		/* We have some initial padding bytes.
		*/
		proto_tree_add_bytes(tree, hf_smb_padding, tvb, offset, po-offset,
			tvb_get_ptr(tvb, offset, po-offset));
		offset += po-offset;
	}
	if(pc){
		/* TRANSACTION2 parameters*/
		if(si->cmd==0x32){
			offset = dissect_transaction2_request_parameters(tvb, pinfo, tree, offset, pc);
		}
		if(si->cmd==0x25){
			/*XXXX process data*/
			offset += pc;
		}
	}

	/* data */
	if(od>offset){
		/* We have some initial padding bytes.
		*/
		proto_tree_add_bytes(tree, hf_smb_padding, tvb, offset, od-offset,
			tvb_get_ptr(tvb, offset, od-offset));
		offset += od-offset;
	}
	if(dc){
		/* TRANSACTION2 data*/
		if(si->cmd==0x32){
			offset = dissect_transaction2_request_data(tvb, pinfo, tree, offset, dc);
		}
		if(si->cmd==0x25){
			/*XXXX process data*/
			offset += dc;
		}
	}


	/*TRANSACTION request parameters */
/*qqq*/
	if(si->cmd==0x25){
		/*XXX replace this block with a function and use that one 
		     for both requests/responses*/
		if(dd==0){
			tvbuff_t *p_tvb, *d_tvb, *s_tvb;

			if(pc>0){
				if(pc>tvb_length_remaining(tvb, po)){
    					p_tvb = tvb_new_subset(tvb, po, tvb_length_remaining(tvb, po), pc);
				} else {
    					p_tvb = tvb_new_subset(tvb, po, pc, pc);
				}
			} else {
				p_tvb = NULL;
			}
			if(dc>0){
				if(dc>tvb_length_remaining(tvb, od)){
    					d_tvb = tvb_new_subset(tvb, od, tvb_length_remaining(tvb, od), dc);
				} else {
    					d_tvb = tvb_new_subset(tvb, od, dc, dc);
				}
			} else {
				d_tvb = NULL;
			}
			if(sc){
    				s_tvb = tvb_new_subset(tvb, so, tvb_length_remaining(tvb, so), tvb_length_remaining(tvb, so));
			} else {
				s_tvb = NULL;
			}

			if(!strncmp("\\PIPE\\LANMAN", an, 12)){
				si->subcmd=TRANSACTION_PIPE_LANMAN;
       				dissect_pipe_smb(p_tvb, d_tvb, pinfo, top_tree);
			}
/*qqq*/
			if(!strncmp("\\MAILSLOT\\", an, 10)){
				si->subcmd=TRANSACTION_MAILSLOT;
       				dissect_mailslot_smb(s_tvb, d_tvb, an+10, pinfo, top_tree);
			}
		} else {
			if(check_col(pinfo->fd, COL_INFO)){
				col_add_str(pinfo->fd, COL_INFO,
					"[transact continuation]");
			}
		}
	}

	return offset;
}

static int
dissect_transaction_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree)
{
	guint8 sc;
	guint16 wc, od, tf, po, pc, pd, dc, dd;
	int so;
	guint32 to;
	smb_info *si;

	si = (smb_info *)pinfo->private;

	if(si->cmd==0x32){
		if (check_col(pinfo->fd, COL_INFO)) {
			col_add_fstr(pinfo->fd, COL_INFO, "%s %s Response",
				decode_smb_name(si->cmd), 
				val_to_str(si->subcmd, trans2_cmd_vals, 
					"Unknown (0x%02X)"));
		}
	}

	/* word count */
	wc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_word_count, tvb, offset, 1, wc);
	offset += 1;

	/* is this an interim response ?*/
	if(wc==0){
		proto_tree_add_text(tree, tvb, offset, 0,
			"Interim response");
		
		/* byte count */
		proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
		offset += 2;

		return offset;
	}

	/* total param count */
	proto_tree_add_uint(tree, hf_smb_total_param_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* total data count */
	proto_tree_add_uint(tree, hf_smb_total_data_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;

	/* 2 reserved bytes */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 2, tvb_get_ptr(tvb, offset, 2));
	offset += 2;

	/* param count */
	pc = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_param_count, tvb, offset, 2, pc);
	offset += 2;

	/* param offset */
	po = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_param_offset, tvb, offset, 2, po);
	offset += 2;

	/* param disp */
	pd = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_param_disp, tvb, offset, 2, pd);
	offset += 2;
	
	/* data count */
	dc = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_data_count, tvb, offset, 2, dc);
	offset += 2;

	/* data offset */
	od = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_data_offset, tvb, offset, 2, od);
	offset += 2;

	/* data disp */
	dd = tvb_get_letohs(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_data_disp, tvb, offset, 2, dd);
	offset += 2;

	/* setup count */
	sc = tvb_get_guint8(tvb, offset);
	proto_tree_add_uint(tree, hf_smb_setup_count, tvb, offset, 1, sc);
	offset += 1;

	/* reserved byte */
	proto_tree_add_bytes(tree, hf_smb_reserved, tvb, offset, 1, tvb_get_ptr(tvb, offset, 1));
	offset += 1;

	/* save setup offset */	
	so=offset;

	/* if there were any setup bytes, decode them */
	if(sc){
		/* XXXX dissect setup words */
		offset += sc*2;
	}

	/* byte count */
	proto_tree_add_uint(tree, hf_smb_byte_count, tvb, offset, 2, tvb_get_letohs(tvb, offset));
	offset += 2;
	
	/* parameters */
	if(po>offset){
		/* We have some initial padding bytes.
		*/
		proto_tree_add_bytes(tree, hf_smb_padding, tvb, offset, po-offset,
			tvb_get_ptr(tvb, offset, po-offset));
		offset += po-offset;
	}
	if(pc){
		/* TRANSACTION2 parameters*/
		if(si->cmd==0x32){
			offset = dissect_transaction2_response_parameters(tvb, pinfo, tree, offset, pc, od);
		}
		if(si->cmd==0x25){
			/*XXXX process parameters*/
			offset += pc;
		}
	}

	/* data */
	if(od>offset){
		/* We have some initial padding bytes.
		*/
		proto_tree_add_bytes(tree, hf_smb_padding, tvb, offset, od-offset,
			tvb_get_ptr(tvb, offset, od-offset));
		offset += od-offset;
	}
	if(dc){
		/* TRANSACTION2 data*/
		if(si->cmd==0x32){
			offset = dissect_transaction2_response_data(tvb, pinfo, tree, offset, dc);
		}
		if(si->cmd==0x25){
			offset += dc;
			/*XXX*/
		}
	}

	/* TRANSACTION response parameters */
	if(si->cmd==0x25){
/*qqq*/
		/* only call subdissector for the first packet */
		if(dd==0){
			tvbuff_t *p_tvb, *d_tvb, *s_tvb;

			if(pc>0){
				if(pc>tvb_length_remaining(tvb, po)){
    					p_tvb = tvb_new_subset(tvb, po, tvb_length_remaining(tvb, po), pc);
				} else {
    					p_tvb = tvb_new_subset(tvb, po, pc, pc);
				}
			} else {
				p_tvb = NULL;
			}
			if(dc>0){
				if(dc>tvb_length_remaining(tvb, od)){
    					d_tvb = tvb_new_subset(tvb, od, tvb_length_remaining(tvb, od), dc);
				} else {
    					d_tvb = tvb_new_subset(tvb, od, dc, dc);
				}
			} else {
				d_tvb = NULL;
			}
			if(sc){
    				s_tvb = tvb_new_subset(tvb, so, tvb_length_remaining(tvb, so), tvb_length_remaining(tvb, so));
			} else {
				s_tvb = NULL;
			}

			
			if(si->subcmd==TRANSACTION_PIPE_LANMAN){
       				dissect_pipe_smb(p_tvb, d_tvb, pinfo, top_tree);
			}
			if(si->subcmd==TRANSACTION_MAILSLOT){
       				dissect_mailslot_smb(s_tvb, d_tvb, NULL, pinfo, top_tree);
			}
		} else {
			if(check_col(pinfo->fd, COL_INFO)){
				col_add_str(pinfo->fd, COL_INFO,
					"[transact continuation]");
			}
		}
	}

	return offset;
}

/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  End of Transaction2 dissector
  XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/


typedef struct _smb_function {
       int (*request)(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree);
       int (*response)(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree);
} smb_function;

smb_function smb_dissector[256] = {
/* 0x00 Create Dir*/		{dissect_create_dir_request, dissect_empty_response},
/* 0x01 Delete Dir*/		{dissect_create_dir_request, dissect_empty_response},
/* 0x02 Open File*/		{dissect_open_file_request, dissect_open_file_response},
/* 0x03 Create File*/		{dissect_create_file_request, dissect_fid_response},
/* 0x04 Close File*/		{dissect_close_file_request, dissect_empty_response},
/* 0x05 Flush File*/		{dissect_fid_response, dissect_empty_response},
/* 0x06 Delete File*/		{dissect_delete_file_request, dissect_empty_response},
/* 0x07 Rename File*/		{dissect_rename_file_request, dissect_empty_response},
/* 0x08 Query Information*/	{dissect_query_information_request, dissect_query_information_response},
/* 0x09 Set Information*/	{dissect_set_information_request, dissect_empty_response},
/* 0x0a Read File*/		{dissect_read_file_request, dissect_read_file_response},
/* 0x0b Write File*/		{dissect_write_file_request, dissect_write_file_response},
/* 0x0c Lock Byte Range*/	{dissect_lock_request, dissect_empty_response},
/* 0x0d Unlock Byte Range*/	{dissect_lock_request, dissect_empty_response},
/* 0x0e Create Temporary*/	{dissect_create_temporary_request, dissect_create_temporary_response},
/* 0x0f Create New*/		{dissect_create_file_request, dissect_fid_response},
/* 0x10 Check Dir*/		{dissect_create_dir_request, dissect_empty_response},
/* 0x11 Process Exit*/		{dissect_empty_response, dissect_empty_response},
/* 0x12 Seek File*/		{dissect_seek_file_request, dissect_seek_file_response},
/* 0x13 Lock And Read*/		{dissect_read_file_request, dissect_lock_and_read_response},
/* 0x14 Write And Unlock*/	{dissect_write_file_request, dissect_write_file_response},
/* 0x15 */  {NULL, NULL},
/* 0x16 */  {NULL, NULL},
/* 0x17 */  {NULL, NULL},
/* 0x18 */  {NULL, NULL},
/* 0x19 */  {NULL, NULL},
/* 0x1a Read Raw*/		{dissect_read_raw_file_request, NULL},
/* 0x1b Read Mpx*/		{dissect_read_mpx_request, dissect_read_mpx_response},
/* 0x1c */  {NULL, NULL},
/* 0x1d */  {NULL, NULL},
/* 0x1e */  {NULL, NULL},
/* 0x1f */  {NULL, NULL},
/* 0x20 */  {NULL, NULL},
/* 0x21 */  {NULL, NULL},
/* 0x22 SetInformation2*/	{dissect_set_information2_request, dissect_empty_response},
/* 0x23 QueryInformation2*/	{dissect_fid_response, dissect_query_information2_response},
/* 0x24 */  {NULL, NULL},
/* 0x25 Transaction*/		{dissect_transaction_request, dissect_transaction_response},
/* 0x26 */  {NULL, NULL},
/* 0x27 */  {NULL, NULL},
/* 0x28 */  {NULL, NULL},
/* 0x29 */  {NULL, NULL},
/* 0x2a Move File*/		{dissect_move_request, dissect_move_response},
/* 0x2b Echo*/			{dissect_echo_request, dissect_echo_response},
/* 0x2c Write And Close*/	{dissect_write_and_close_request, dissect_write_and_close_response},
/* 0x2d Open AndX*/		{dissect_open_andx_request, dissect_open_andx_response},
/* 0x2e Read AndX*/		{dissect_read_andx_request, dissect_read_andx_response},
/* 0x2f */  {NULL, NULL},
/* 0x30 */  {NULL, NULL},
/* 0x31 */  {NULL, NULL},
/* 0x32 Transaction2*/		{dissect_transaction_request, dissect_transaction_response},
/* 0x33 */  {NULL, NULL},
/* 0x34 Find Close2*/		{dissect_sid_request, dissect_empty_response},
/* 0x35 */  {NULL, NULL},
/* 0x36 */  {NULL, NULL},
/* 0x37 */  {NULL, NULL},
/* 0x38 */  {NULL, NULL},
/* 0x39 */  {NULL, NULL},
/* 0x3a */  {NULL, NULL},
/* 0x3b */  {NULL, NULL},
/* 0x3c */  {NULL, NULL},
/* 0x3d */  {NULL, NULL},
/* 0x3e */  {NULL, NULL},
/* 0x3f */  {NULL, NULL},
/* 0x40 */  {NULL, NULL},
/* 0x41 */  {NULL, NULL},
/* 0x42 */  {NULL, NULL},
/* 0x43 */  {NULL, NULL},
/* 0x44 */  {NULL, NULL},
/* 0x45 */  {NULL, NULL},
/* 0x46 */  {NULL, NULL},
/* 0x47 */  {NULL, NULL},
/* 0x48 */  {NULL, NULL},
/* 0x49 */  {NULL, NULL},
/* 0x4a */  {NULL, NULL},
/* 0x4b */  {NULL, NULL},
/* 0x4c */  {NULL, NULL},
/* 0x4d */  {NULL, NULL},
/* 0x4e */  {NULL, NULL},
/* 0x4f */  {NULL, NULL},
/* 0x50 */  {NULL, NULL},
/* 0x51 */  {NULL, NULL},
/* 0x52 */  {NULL, NULL},
/* 0x53 */  {NULL, NULL},
/* 0x54 */  {NULL, NULL},
/* 0x55 */  {NULL, NULL},
/* 0x56 */  {NULL, NULL},
/* 0x57 */  {NULL, NULL},
/* 0x58 */  {NULL, NULL},
/* 0x59 */  {NULL, NULL},
/* 0x5a */  {NULL, NULL},
/* 0x5b */  {NULL, NULL},
/* 0x5c */  {NULL, NULL},
/* 0x5d */  {NULL, NULL},
/* 0x5e */  {NULL, NULL},
/* 0x5f */  {NULL, NULL},
/* 0x60 */  {NULL, NULL},
/* 0x61 */  {NULL, NULL},
/* 0x62 */  {NULL, NULL},
/* 0x63 */  {NULL, NULL},
/* 0x64 */  {NULL, NULL},
/* 0x65 */  {NULL, NULL},
/* 0x66 */  {NULL, NULL},
/* 0x67 */  {NULL, NULL},
/* 0x68 */  {NULL, NULL},
/* 0x69 */  {NULL, NULL},
/* 0x6a */  {NULL, NULL},
/* 0x6b */  {NULL, NULL},
/* 0x6c */  {NULL, NULL},
/* 0x6d */  {NULL, NULL},
/* 0x6e */  {NULL, NULL},
/* 0x6f */  {NULL, NULL},
/* 0x70 Tree Connect*/		{dissect_tree_connect_request, dissect_tree_connect_response},
/* 0x71 Tree Disconnect*/	{dissect_empty_response, dissect_empty_response},
/* 0x72 Negotiate Protocol*/	{dissect_negprot_request, dissect_negprot_response},
/* 0x73 Session Setup AndX*/	{dissect_session_setup_andx_request, dissect_session_setup_andx_response},
/* 0x74 Logoff AndX*/		{dissect_empty_andx, dissect_empty_andx},
/* 0x75 Tree Connect AndX*/	{dissect_tree_connect_andx_request, dissect_tree_connect_andx_response},
/* 0x76 */  {NULL, NULL},
/* 0x77 */  {NULL, NULL},
/* 0x78 */  {NULL, NULL},
/* 0x79 */  {NULL, NULL},
/* 0x7a */  {NULL, NULL},
/* 0x7b */  {NULL, NULL},
/* 0x7c */  {NULL, NULL},
/* 0x7d */  {NULL, NULL},
/* 0x7e */  {NULL, NULL},
/* 0x7f */  {NULL, NULL},
/* 0x80 Query Info Disk*/	{dissect_empty_response, dissect_query_information_disk_response},
/* 0x81 */  {NULL, NULL},
/* 0x82 */  {NULL, NULL},
/* 0x83 */  {NULL, NULL},
/* 0x84 */  {NULL, NULL},
/* 0x85 */  {NULL, NULL},
/* 0x86 */  {NULL, NULL},
/* 0x87 */  {NULL, NULL},
/* 0x88 */  {NULL, NULL},
/* 0x89 */  {NULL, NULL},
/* 0x8a */  {NULL, NULL},
/* 0x8b */  {NULL, NULL},
/* 0x8c */  {NULL, NULL},
/* 0x8d */  {NULL, NULL},
/* 0x8e */  {NULL, NULL},
/* 0x8f */  {NULL, NULL},
/* 0x90 */  {NULL, NULL},
/* 0x91 */  {NULL, NULL},
/* 0x92 */  {NULL, NULL},
/* 0x93 */  {NULL, NULL},
/* 0x94 */  {NULL, NULL},
/* 0x95 */  {NULL, NULL},
/* 0x96 */  {NULL, NULL},
/* 0x97 */  {NULL, NULL},
/* 0x98 */  {NULL, NULL},
/* 0x99 */  {NULL, NULL},
/* 0x9a */  {NULL, NULL},
/* 0x9b */  {NULL, NULL},
/* 0x9c */  {NULL, NULL},
/* 0x9d */  {NULL, NULL},
/* 0x9e */  {NULL, NULL},
/* 0x9f */  {NULL, NULL},
/* 0xa0 */  {NULL, NULL},
/* 0xa1 */  {NULL, NULL},
/* 0xa2 */  {NULL, NULL},
/* 0xa3 */  {NULL, NULL},
/* 0xa4 */  {NULL, NULL},
/* 0xa5 */  {NULL, NULL},
/* 0xa6 */  {NULL, NULL},
/* 0xa7 */  {NULL, NULL},
/* 0xa8 */  {NULL, NULL},
/* 0xa9 */  {NULL, NULL},
/* 0xaa */  {NULL, NULL},
/* 0xab */  {NULL, NULL},
/* 0xac */  {NULL, NULL},
/* 0xad */  {NULL, NULL},
/* 0xae */  {NULL, NULL},
/* 0xaf */  {NULL, NULL},
/* 0xb0 */  {NULL, NULL},
/* 0xb1 */  {NULL, NULL},
/* 0xb2 */  {NULL, NULL},
/* 0xb3 */  {NULL, NULL},
/* 0xb4 */  {NULL, NULL},
/* 0xb5 */  {NULL, NULL},
/* 0xb6 */  {NULL, NULL},
/* 0xb7 */  {NULL, NULL},
/* 0xb8 */  {NULL, NULL},
/* 0xb9 */  {NULL, NULL},
/* 0xba */  {NULL, NULL},
/* 0xbb */  {NULL, NULL},
/* 0xbc */  {NULL, NULL},
/* 0xbd */  {NULL, NULL},
/* 0xbe */  {NULL, NULL},
/* 0xbf */  {NULL, NULL},
/* 0xc0 */  {NULL, NULL},
/* 0xc1 */  {NULL, NULL},
/* 0xc2 */  {NULL, NULL},
/* 0xc3 */  {NULL, NULL},
/* 0xc4 */  {NULL, NULL},
/* 0xc5 */  {NULL, NULL},
/* 0xc6 */  {NULL, NULL},
/* 0xc7 */  {NULL, NULL},
/* 0xc8 */  {NULL, NULL},
/* 0xc9 */  {NULL, NULL},
/* 0xca */  {NULL, NULL},
/* 0xcb */  {NULL, NULL},
/* 0xcc */  {NULL, NULL},
/* 0xcd */  {NULL, NULL},
/* 0xce */  {NULL, NULL},
/* 0xcf */  {NULL, NULL},
/* 0xd0 */  {NULL, NULL},
/* 0xd1 */  {NULL, NULL},
/* 0xd2 */  {NULL, NULL},
/* 0xd3 */  {NULL, NULL},
/* 0xd4 */  {NULL, NULL},
/* 0xd5 */  {NULL, NULL},
/* 0xd6 */  {NULL, NULL},
/* 0xd7 */  {NULL, NULL},
/* 0xd8 */  {NULL, NULL},
/* 0xd9 */  {NULL, NULL},
/* 0xda */  {NULL, NULL},
/* 0xdb */  {NULL, NULL},
/* 0xdc */  {NULL, NULL},
/* 0xdd */  {NULL, NULL},
/* 0xde */  {NULL, NULL},
/* 0xdf */  {NULL, NULL},
/* 0xe0 */  {NULL, NULL},
/* 0xe1 */  {NULL, NULL},
/* 0xe2 */  {NULL, NULL},
/* 0xe3 */  {NULL, NULL},
/* 0xe4 */  {NULL, NULL},
/* 0xe5 */  {NULL, NULL},
/* 0xe6 */  {NULL, NULL},
/* 0xe7 */  {NULL, NULL},
/* 0xe8 */  {NULL, NULL},
/* 0xe9 */  {NULL, NULL},
/* 0xea */  {NULL, NULL},
/* 0xeb */  {NULL, NULL},
/* 0xec */  {NULL, NULL},
/* 0xed */  {NULL, NULL},
/* 0xee */  {NULL, NULL},
/* 0xef */  {NULL, NULL},
/* 0xf0 */  {NULL, NULL},
/* 0xf1 */  {NULL, NULL},
/* 0xf2 */  {NULL, NULL},
/* 0xf3 */  {NULL, NULL},
/* 0xf4 */  {NULL, NULL},
/* 0xf5 */  {NULL, NULL},
/* 0xf6 */  {NULL, NULL},
/* 0xf7 */  {NULL, NULL},
/* 0xf8 */  {NULL, NULL},
/* 0xf9 */  {NULL, NULL},
/* 0xfa */  {NULL, NULL},
/* 0xfb */  {NULL, NULL},
/* 0xfc */  {NULL, NULL},
/* 0xfd */  {NULL, NULL},
/* 0xfe */  {NULL, NULL},
/* 0xff */  {NULL, NULL}
};

static int
call_next_command(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree, guint8 cmd)
{
	if(cmd!=0xff){
		proto_item *cmd_item;
		proto_tree *cmd_tree;
		int (*new_dissector)(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree);

		cmd_item = proto_tree_add_text(smb_tree, tvb, offset,
			END_OF_FRAME, "%s %s (0x%02x)",
			decode_smb_name(cmd), 
			(pinfo->match_port==pinfo->destport)?"Request":"Response",
			cmd);

		cmd_tree = proto_item_add_subtree(cmd_item, ett_smb_command);

		new_dissector = (pinfo->match_port==pinfo->destport)?
			smb_dissector[cmd].request:smb_dissector[cmd].response;

		if(new_dissector){
			offset = (*new_dissector)(tvb, pinfo, cmd_tree, offset, smb_tree);
		}
	}
	return offset;
}


/* XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
 * Everything tvbuffified above this line
 * XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
 */


static void (*dissect[256])(const u_char *, int, frame_data *, proto_tree *, proto_tree *, smb_info si, int, int, int);

/* NOTE: this value_string array will also be used to access data directly by
 * index instead of val_to_str() since 
 * 1, the array will always span every value from 0x00 to 0xff and
 * 2, smb_cmd_vals[i].strptr  is much cheaper than  val_to_str(i, smb_cmd_vals,)
 * This means that this value_string array MUST always
 * 1, contain all entries 0x00 to 0xff
 * 2, all entries must be in order.
 */
static const value_string smb_cmd_vals[] = {
  { 0x00, "SMBcreatedirectory" },
  { 0x01, "SMBdeletedirectory" },
  { 0x02, "SMBopen" },
  { 0x03, "SMBcreate" },
  { 0x04, "SMBclose" },
  { 0x05, "SMBflush" },
  { 0x06, "SMBunlink" },
  { 0x07, "SMBmv" },
  { 0x08, "SMBQueryInformation" },
  { 0x09, "SMBsetatr" },
  { 0x0A, "SMBread" },
  { 0x0B, "SMBwrite" },
  { 0x0C, "SMBlock" },
  { 0x0D, "SMBunlock" },
  { 0x0E, "SMBctemp" },
  { 0x0F, "SMBmknew" },
  { 0x10, "SMBchkpth" },
  { 0x11, "SMBexit" },
  { 0x12, "SMBlseek" },
  { 0x13, "SMBlockread" },
  { 0x14, "SMBwriteunlock" },
  { 0x15, "unknown-0x15" },
  { 0x16, "unknown-0x16" },
  { 0x17, "unknown-0x17" },
  { 0x18, "unknown-0x18" },
  { 0x19, "unknown-0x19" },
  { 0x1A, "SMBreadBraw" },
  { 0x1B, "SMBreadBmpx" },
  { 0x1C, "SMBreadBs" },
  { 0x1D, "SMBwriteBraw" },
  { 0x1E, "SMBwriteBmpx" },
  { 0x1F, "SMBwriteBs" },
  { 0x20, "SMBwriteC" },
  { 0x21, "unknown-0x21" },
  { 0x22, "SMBSetInformation2" },
  { 0x23, "SMBQueryInformation2" },
  { 0x24, "SMBlockingX" },
  { 0x25, "SMBtrans" },
  { 0x26, "SMBtranss" },
  { 0x27, "SMBioctl" },
  { 0x28, "SMBioctls" },
  { 0x29, "SMBcopy" },
  { 0x2A, "SMBmove" },
  { 0x2B, "SMBecho" },
  { 0x2C, "SMBwriteclose" },
  { 0x2D, "SMBopenX" },
  { 0x2E, "SMBreadX" },
  { 0x2F, "SMBwriteX" },
  { 0x30, "unknown-0x30" },
  { 0x31, "SMBcloseandtreedisc" },
  { 0x32, "SMBtrans2" },
  { 0x33, "SMBtrans2secondary" },
  { 0x34, "SMBfindclose2" },
  { 0x35, "SMBfindnotifyclose" },
  { 0x36, "unknown-0x36" },
  { 0x37, "unknown-0x37" },
  { 0x38, "unknown-0x38" },
  { 0x39, "unknown-0x39" },
  { 0x3A, "unknown-0x3A" },
  { 0x3B, "unknown-0x3B" },
  { 0x3C, "unknown-0x3C" },
  { 0x3D, "unknown-0x3D" },
  { 0x3E, "unknown-0x3E" },
  { 0x3F, "unknown-0x3F" },
  { 0x40, "unknown-0x40" },
  { 0x41, "unknown-0x41" },
  { 0x42, "unknown-0x42" },
  { 0x43, "unknown-0x43" },
  { 0x44, "unknown-0x44" },
  { 0x45, "unknown-0x45" },
  { 0x46, "unknown-0x46" },
  { 0x47, "unknown-0x47" },
  { 0x48, "unknown-0x48" },
  { 0x49, "unknown-0x49" },
  { 0x4A, "unknown-0x4A" },
  { 0x4B, "unknown-0x4B" },
  { 0x4C, "unknown-0x4C" },
  { 0x4D, "unknown-0x4D" },
  { 0x4E, "unknown-0x4E" },
  { 0x4F, "unknown-0x4F" },
  { 0x50, "unknown-0x50" },
  { 0x51, "unknown-0x51" },
  { 0x52, "unknown-0x52" },
  { 0x53, "unknown-0x53" },
  { 0x54, "unknown-0x54" },
  { 0x55, "unknown-0x55" },
  { 0x56, "unknown-0x56" },
  { 0x57, "unknown-0x57" },
  { 0x58, "unknown-0x58" },
  { 0x59, "unknown-0x59" },
  { 0x5A, "unknown-0x5A" },
  { 0x5B, "unknown-0x5B" },
  { 0x5C, "unknown-0x5C" },
  { 0x5D, "unknown-0x5D" },
  { 0x5E, "unknown-0x5E" },
  { 0x5F, "unknown-0x5F" },
  { 0x60, "unknown-0x60" },
  { 0x61, "unknown-0x61" },
  { 0x62, "unknown-0x62" },
  { 0x63, "unknown-0x63" },
  { 0x64, "unknown-0x64" },
  { 0x65, "unknown-0x65" },
  { 0x66, "unknown-0x66" },
  { 0x67, "unknown-0x67" },
  { 0x68, "unknown-0x68" },
  { 0x69, "unknown-0x69" },
  { 0x6A, "unknown-0x6A" },
  { 0x6B, "unknown-0x6B" },
  { 0x6C, "unknown-0x6C" },
  { 0x6D, "unknown-0x6D" },
  { 0x6E, "unknown-0x6E" },
  { 0x6F, "unknown-0x6F" },
  { 0x70, "SMBtcon" },
  { 0x71, "SMBtdis" },
  { 0x72, "SMBnegprot" },
  { 0x73, "SMBsesssetupX" },
  { 0x74, "SMBlogoffX" },
  { 0x75, "SMBtconX" },
  { 0x76, "unknown-0x76" },
  { 0x77, "unknown-0x77" },
  { 0x78, "unknown-0x78" },
  { 0x79, "unknown-0x79" },
  { 0x7A, "unknown-0x7A" },
  { 0x7B, "unknown-0x7B" },
  { 0x7C, "unknown-0x7C" },
  { 0x7D, "unknown-0x7D" },
  { 0x7E, "unknown-0x7E" },
  { 0x7F, "unknown-0x7F" },
  { 0x80, "SMBdskattr" },
  { 0x81, "SMBsearch" },
  { 0x82, "SMBffirst" },
  { 0x83, "SMBfunique" },
  { 0x84, "SMBfclose" },
  { 0x85, "unknown-0x85" },
  { 0x86, "unknown-0x86" },
  { 0x87, "unknown-0x87" },
  { 0x88, "unknown-0x88" },
  { 0x89, "unknown-0x89" },
  { 0x8A, "unknown-0x8A" },
  { 0x8B, "unknown-0x8B" },
  { 0x8C, "unknown-0x8C" },
  { 0x8D, "unknown-0x8D" },
  { 0x8E, "unknown-0x8E" },
  { 0x8F, "unknown-0x8F" },
  { 0x90, "unknown-0x90" },
  { 0x91, "unknown-0x91" },
  { 0x92, "unknown-0x92" },
  { 0x93, "unknown-0x93" },
  { 0x94, "unknown-0x94" },
  { 0x95, "unknown-0x95" },
  { 0x96, "unknown-0x96" },
  { 0x97, "unknown-0x97" },
  { 0x98, "unknown-0x98" },
  { 0x99, "unknown-0x99" },
  { 0x9A, "unknown-0x9A" },
  { 0x9B, "unknown-0x9B" },
  { 0x9C, "unknown-0x9C" },
  { 0x9D, "unknown-0x9D" },
  { 0x9E, "unknown-0x9E" },
  { 0x9F, "unknown-0x9F" },
  { 0xA0, "SMBnttransact" },
  { 0xA1, "SMBnttransactsecondary" },
  { 0xA2, "SMBntcreateX" },
  { 0xA3, "unknown-0xA3" },
  { 0xA4, "SMBntcancel" },
  { 0xA5, "unknown-0xA5" },
  { 0xA6, "unknown-0xA6" },
  { 0xA7, "unknown-0xA7" },
  { 0xA8, "unknown-0xA8" },
  { 0xA9, "unknown-0xA9" },
  { 0xAA, "unknown-0xAA" },
  { 0xAB, "unknown-0xAB" },
  { 0xAC, "unknown-0xAC" },
  { 0xAD, "unknown-0xAD" },
  { 0xAE, "unknown-0xAE" },
  { 0xAF, "unknown-0xAF" },
  { 0xB0, "unknown-0xB0" },
  { 0xB1, "unknown-0xB1" },
  { 0xB2, "unknown-0xB2" },
  { 0xB3, "unknown-0xB3" },
  { 0xB4, "unknown-0xB4" },
  { 0xB5, "unknown-0xB5" },
  { 0xB6, "unknown-0xB6" },
  { 0xB7, "unknown-0xB7" },
  { 0xB8, "unknown-0xB8" },
  { 0xB9, "unknown-0xB9" },
  { 0xBA, "unknown-0xBA" },
  { 0xBB, "unknown-0xBB" },
  { 0xBC, "unknown-0xBC" },
  { 0xBD, "unknown-0xBD" },
  { 0xBE, "unknown-0xBE" },
  { 0xBF, "unknown-0xBF" },
  { 0xC0, "SMBsplopen" },
  { 0xC1, "SMBsplwr" },
  { 0xC2, "SMBsplclose" },
  { 0xC3, "SMBsplretq" },
  { 0xC4, "unknown-0xC4" },
  { 0xC5, "unknown-0xC5" },
  { 0xC6, "unknown-0xC6" },
  { 0xC7, "unknown-0xC7" },
  { 0xC8, "unknown-0xC8" },
  { 0xC9, "unknown-0xC9" },
  { 0xCA, "unknown-0xCA" },
  { 0xCB, "unknown-0xCB" },
  { 0xCC, "unknown-0xCC" },
  { 0xCD, "unknown-0xCD" },
  { 0xCE, "unknown-0xCE" },
  { 0xCF, "unknown-0xCF" },
  { 0xD0, "SMBsends" },
  { 0xD1, "SMBsendb" },
  { 0xD2, "SMBfwdname" },
  { 0xD3, "SMBcancelf" },
  { 0xD4, "SMBgetmac" },
  { 0xD5, "SMBsendstrt" },
  { 0xD6, "SMBsendend" },
  { 0xD7, "SMBsendtxt" },
  { 0xD8, "SMBreadbulk" },
  { 0xD9, "SMBwritebulk" },
  { 0xDA, "SMBwritebulkdata" },
  { 0xDB, "unknown-0xDB" },
  { 0xDC, "unknown-0xDC" },
  { 0xDD, "unknown-0xDD" },
  { 0xDE, "unknown-0xDE" },
  { 0xDF, "unknown-0xDF" },
  { 0xE0, "unknown-0xE0" },
  { 0xE1, "unknown-0xE1" },
  { 0xE2, "unknown-0xE2" },
  { 0xE3, "unknown-0xE3" },
  { 0xE4, "unknown-0xE4" },
  { 0xE5, "unknown-0xE5" },
  { 0xE6, "unknown-0xE6" },
  { 0xE7, "unknown-0xE7" },
  { 0xE8, "unknown-0xE8" },
  { 0xE9, "unknown-0xE9" },
  { 0xEA, "unknown-0xEA" },
  { 0xEB, "unknown-0xEB" },
  { 0xEC, "unknown-0xEC" },
  { 0xED, "unknown-0xED" },
  { 0xEE, "unknown-0xEE" },
  { 0xEF, "unknown-0xEF" },
  { 0xF0, "unknown-0xF0" },
  { 0xF1, "unknown-0xF1" },
  { 0xF2, "unknown-0xF2" },
  { 0xF3, "unknown-0xF3" },
  { 0xF4, "unknown-0xF4" },
  { 0xF5, "unknown-0xF5" },
  { 0xF6, "unknown-0xF6" },
  { 0xF7, "unknown-0xF7" },
  { 0xF8, "unknown-0xF8" },
  { 0xF9, "unknown-0xF9" },
  { 0xFA, "unknown-0xFA" },
  { 0xFB, "unknown-0xFB" },
  { 0xFC, "unknown-0xFC" },
  { 0xFD, "unknown-0xFD" },
  { 0xFE, "SMBinvalid" },
  { 0xFF, "unknown-0xFF" },
  { 0x00, NULL },
};

void 
dissect_unknown_smb(const u_char *pd, int offset, frame_data *fd, proto_tree *parent, proto_tree *tree, smb_info si, int max_data, int SMB_offset, int errcode)
{

  if (tree) {

    proto_tree_add_text(tree, NullTVB, offset, END_OF_FRAME, "Data (%u bytes)", 
			END_OF_FRAME); 

  }

}

/* 
 * Dissect a UNIX like date ...
 */

struct tm *_gtime; /* Add leading underscore ("_") to prevent symbol
                      conflict with /usr/include/time.h on some NetBSD
                      systems */

static char *
dissect_smbu_date(guint16 date, guint16 time)

{
  static char         datebuf[4+2+2+2+1+10];
  time_t              ltime = (date << 16) + time;

  _gtime = gmtime(&ltime);

  if (_gtime)
    sprintf(datebuf, "%04d-%02d-%02d",
	    1900 + (_gtime -> tm_year), 1 + (_gtime -> tm_mon), _gtime -> tm_mday);
  else 
    sprintf(datebuf, "Bad date format");

  return datebuf;

}

/*
 * Relies on time
 */
static char *
dissect_smbu_time(guint16 date, guint16 time)

{
  static char timebuf[2+2+2+2+1+10];

  if (_gtime)
    sprintf(timebuf, "%02d:%02d:%02d",
	    _gtime -> tm_hour, _gtime -> tm_min, _gtime -> tm_sec);
  else
    sprintf(timebuf, "Bad time format");

  return timebuf;

}

/*
 * Dissect a DOS-format date.
 */
static char *
dissect_dos_date(guint16 date)
{
	static char datebuf[4+2+2+1];

	sprintf(datebuf, "%04d-%02d-%02d",
	    ((date>>9)&0x7F) + 1980, (date>>5)&0x0F, date&0x1F);
	return datebuf;
}

/*
 * Dissect a DOS-format time.
 */
static char *
dissect_dos_time(guint16 time)
{
	static char timebuf[2+2+2+1];

	sprintf(timebuf, "%02d:%02d:%02d",
	    (time>>11)&0x1F, (time>>5)&0x3F, (time&0x1F)*2);
	return timebuf;
}

/* Max string length for displaying Unicode strings.  */
#define	MAX_UNICODE_STR_LEN	256

/* Turn a little-endian Unicode '\0'-terminated string into a string we
   can display.
   XXX - for now, we just handle the ISO 8859-1 characters.

   If exactlen==TRUE then us_lenp contains the exact len of the string in
   bytes. It might not be null terminated !
*/
static gchar *
unicode_to_str(const guint8 *us, int *us_lenp, gboolean exactlen) {
  static gchar  str[3][MAX_UNICODE_STR_LEN+3+1];
  static gchar *cur;
  gchar        *p;
  int           len;
  int           us_len;
  int           overflow = 0;

  if (cur == &str[0][0]) {
    cur = &str[1][0];
  } else if (cur == &str[1][0]) {  
    cur = &str[2][0];
  } else {  
    cur = &str[0][0];
  }
  p = cur;
  len = MAX_UNICODE_STR_LEN;
  us_len = 0;
  while (*us != 0 || *(us + 1) != 0) {
    if (len > 0) {
      *p++ = *us;
      len--;
    } else
      overflow = 1;
    us += 2;
    us_len += 2;
    if(exactlen){
      if(us_len>= *us_lenp){
        break;
      }
    }
  }
  if (overflow) {
    /* Note that we're not showing the full string.  */
    *p++ = '.';
    *p++ = '.';
    *p++ = '.';
  }
  *p = '\0';
  *us_lenp = us_len;
  return cur;
}

static const gchar *
get_unicode_or_ascii_string(const u_char *pd, int *offsetp, gboolean is_unicode,
    int *len)
{
  int offset = *offsetp;
  const gchar *string;
  int string_len;

  if (is_unicode) {
    if (offset % 2) {
      /*
       * XXX - this should be an offset relative to the beginning of the SMB,
       * not an offset relative to the beginning of the frame; if the stuff
       * before the SMB has an odd number of bytes, an offset relative to
       * the beginning of the frame will give the wrong answer.
       */
      offset++;   /* Looks like a pad byte there sometimes */
      *offsetp = offset;
    }
    string = unicode_to_str(pd + offset, &string_len, FALSE);
    string_len += 2;
  } else {
    string = pd + offset;
    string_len = strlen(string) + 1;
  }
  *len = string_len;
  return string;
}

/* nopad == TRUE : Do not add any padding before this string
 * exactlen == TRUE : len contains the exact len of the string in bytes.
 */
static const gchar *
get_unicode_or_ascii_string_tvb(tvbuff_t *tvb, int *offsetp, packet_info *pinfo, int *len, gboolean nopad, gboolean exactlen)
{
  int offset = *offsetp;
  const gchar *string;
  int string_len;
  smb_info *si;

  si = pinfo->private;
  if (si->unicode) {
    if ((!nopad) && (*offsetp % 2)) {
      /*
       * XXX - this should be an offset relative to the beginning of the SMB,
       * not an offset relative to the beginning of the frame; if the stuff
       * before the SMB has an odd number of bytes, an offset relative to
       * the beginning of the frame will give the wrong answer.
       */
      (*offsetp)++;   /* Looks like a pad byte there sometimes */
    }
    if(exactlen){
      string_len = *len;
      string = unicode_to_str(tvb_get_ptr(tvb, *offsetp, -1), &string_len, exactlen);
    } else {
      string = unicode_to_str(tvb_get_ptr(tvb, *offsetp, -1), &string_len, exactlen);
      string_len += 2;
    }
  } else {
    if(exactlen){
      string = tvb_get_ptr(tvb, *offsetp, *len);
      string_len = *len;
    } else {
      string = tvb_get_ptr(tvb, *offsetp, -1);
      string_len = strlen(string) + 1;
    }
  }
  *len = string_len;
  return string;
}

/*
 * Each dissect routine is passed an offset to wct and works from there 
 */

void
dissect_write_raw_smb(const u_char *pd, int offset, frame_data *fd, proto_tree *parent, proto_tree *tree, smb_info si, int max_data, int SMB_offset, int errcode)

{
  proto_tree    *WriteMode_tree;
  proto_item    *ti;
  guint8        WordCount;
  guint8        Pad;
  guint32       Timeout;
  guint32       Reserved2;
  guint32       Offset;
  guint16       WriteMode;
  guint16       Reserved1;
  guint16       Remaining;
  guint16       FID;
  guint16       DataOffset;
  guint16       DataLength;
  guint16       Count;
  guint16       ByteCount;

  if (si.request) {
    /* Request(s) dissect code */

    WordCount = GBYTE(pd, offset);

    switch (WordCount) {

    case 12:

      /* Build display for: Word Count (WCT) */

      WordCount = GBYTE(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

      }

      offset += 1; /* Skip Word Count (WCT) */

      /* Build display for: FID */

      FID = GSHORT(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 2, "FID: %u", FID);

      }

      offset += 2; /* Skip FID */

      /* Build display for: Count */

      Count = GSHORT(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 2, "Count: %u", Count);

      }

      offset += 2; /* Skip Count */

      /* Build display for: Reserved 1 */

      Reserved1 = GSHORT(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 2, "Reserved 1: %u", Reserved1);

      }

      offset += 2; /* Skip Reserved 1 */

      /* Build display for: Offset */

      Offset = GWORD(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 4, "Offset: %u", Offset);

      }

      offset += 4; /* Skip Offset */

      /* Build display for: Timeout */

      Timeout = GWORD(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 4, "Timeout: %u", Timeout);

      }

      offset += 4; /* Skip Timeout */

      /* Build display for: WriteMode */

      WriteMode = GSHORT(pd, offset);

      if (tree) {

        ti = proto_tree_add_text(tree, NullTVB, offset, 2, "WriteMode: 0x%02x", WriteMode);
        WriteMode_tree = proto_item_add_subtree(ti, ett_smb_writemode);
        proto_tree_add_text(WriteMode_tree, NullTVB, offset, 2, "%s",
                            decode_boolean_bitfield(WriteMode, 0x01, 16, "Write through requested", "Write through not requested"));
        proto_tree_add_text(WriteMode_tree, NullTVB, offset, 2, "%s",
                            decode_boolean_bitfield(WriteMode, 0x02, 16, "Return Remaining (pipe/dev)", "Dont return Remaining (pipe/dev)"));
      
}

      offset += 2; /* Skip WriteMode */

      /* Build display for: Reserved 2 */

      Reserved2 = GWORD(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 4, "Reserved 2: %u", Reserved2);

      }

      offset += 4; /* Skip Reserved 2 */

      /* Build display for: Data Length */

      DataLength = GSHORT(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 2, "Data Length: %u", DataLength);

      }

      offset += 2; /* Skip Data Length */

      /* Build display for: Data Offset */

      DataOffset = GSHORT(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 2, "Data Offset: %u", DataOffset);

      }

      offset += 2; /* Skip Data Offset */

      /* Build display for: Byte Count (BCC) */

      ByteCount = GSHORT(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

      }

      offset += 2; /* Skip Byte Count (BCC) */

      /* Build display for: Pad */

      Pad = GBYTE(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 1, "Pad: %u", Pad);

      }

      offset += 1; /* Skip Pad */

    break;

    case 14:

      /* Build display for: Word Count (WCT) */

      WordCount = GBYTE(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

      }

      offset += 1; /* Skip Word Count (WCT) */

      /* Build display for: FID */

      FID = GSHORT(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 2, "FID: %u", FID);

      }

      offset += 2; /* Skip FID */

      /* Build display for: Count */

      Count = GSHORT(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 2, "Count: %u", Count);

      }

      offset += 2; /* Skip Count */

      /* Build display for: Reserved 1 */

      Reserved1 = GSHORT(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 2, "Reserved 1: %u", Reserved1);

      }

      offset += 2; /* Skip Reserved 1 */

      /* Build display for: Timeout */

      Timeout = GWORD(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 4, "Timeout: %u", Timeout);

      }

      offset += 4; /* Skip Timeout */

      /* Build display for: WriteMode */

      WriteMode = GSHORT(pd, offset);

      if (tree) {

        ti = proto_tree_add_text(tree, NullTVB, offset, 2, "WriteMode: 0x%02x", WriteMode);
        WriteMode_tree = proto_item_add_subtree(ti, ett_smb_writemode);
        proto_tree_add_text(WriteMode_tree, NullTVB, offset, 2, "%s",
                            decode_boolean_bitfield(WriteMode, 0x01, 16, "Write through requested", "Write through not requested"));
        proto_tree_add_text(WriteMode_tree, NullTVB, offset, 2, "%s",
                            decode_boolean_bitfield(WriteMode, 0x02, 16, "Return Remaining (pipe/dev)", "Dont return Remaining (pipe/dev)"));
      
}

      offset += 2; /* Skip WriteMode */

      /* Build display for: Reserved 2 */

      Reserved2 = GWORD(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 4, "Reserved 2: %u", Reserved2);

      }

      offset += 4; /* Skip Reserved 2 */

      /* Build display for: Data Length */

      DataLength = GSHORT(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 2, "Data Length: %u", DataLength);

      }

      offset += 2; /* Skip Data Length */

      /* Build display for: Data Offset */

      DataOffset = GSHORT(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 2, "Data Offset: %u", DataOffset);

      }

      offset += 2; /* Skip Data Offset */

      /* Build display for: Byte Count (BCC) */

      ByteCount = GSHORT(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

      }

      offset += 2; /* Skip Byte Count (BCC) */

      /* Build display for: Pad */

      Pad = GBYTE(pd, offset);

      if (tree) {

        proto_tree_add_text(tree, NullTVB, offset, 1, "Pad: %u", Pad);

      }

      offset += 1; /* Skip Pad */

    break;

    }

  } else {
    /* Response(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    if (WordCount > 0) {

      /* Build display for: Remaining */

      Remaining = GSHORT(pd, offset);

      if (tree) {

	proto_tree_add_text(tree, NullTVB, offset, 2, "Remaining: %u", Remaining);

      }

      offset += 2; /* Skip Remaining */

    }

    /* Build display for: Byte Count */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count: %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count */

  }

}

void
dissect_open_print_file_smb(const u_char *pd, int offset, frame_data *fd, proto_tree *parent, proto_tree *tree, smb_info si, int max_data, int SMB_offset, int errcode)

{
  static const value_string Mode_0x03[] = {
	{ 0, "Text mode (DOS expands TABs)"},
	{ 1, "Graphics mode"},
	{ 0, NULL}
  };
  proto_tree    *Mode_tree;
  proto_item    *ti;
  guint8        WordCount;
  guint8        BufferFormat;
  guint16       SetupLength;
  guint16       Mode;
  guint16       FID;
  guint16       ByteCount;
  const char    *IdentifierString;
  int           string_len;

  if (si.request) {
    /* Request(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    /* Build display for: Setup Length */

    SetupLength = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Setup Length: %u", SetupLength);

    }

    offset += 2; /* Skip Setup Length */

    /* Build display for: Mode */

    Mode = GSHORT(pd, offset);

    if (tree) {

      ti = proto_tree_add_text(tree, NullTVB, offset, 2, "Mode: 0x%02x", Mode);
      Mode_tree = proto_item_add_subtree(ti, ett_smb_mode);
      proto_tree_add_text(Mode_tree, NullTVB, offset, 2, "%s",
                          decode_enumerated_bitfield(Mode, 0x03, 16, Mode_0x03, "%s"));
    
}

    offset += 2; /* Skip Mode */

    /* Build display for: Byte Count (BCC) */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count (BCC) */

    /* Build display for: Buffer Format */

    BufferFormat = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Buffer Format: %s (%u)",
			  val_to_str(BufferFormat, buffer_format_vals, "Unknown"),
			  BufferFormat);

    }

    offset += 1; /* Skip Buffer Format */

    /* Build display for: Identifier String */

    IdentifierString = get_unicode_or_ascii_string(pd, &offset, si.unicode, &string_len);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, string_len, "Identifier String: %s", IdentifierString);

    }

    offset += string_len; /* Skip Identifier String */

  } else {
    /* Response(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    /* Build display for: FID */

    FID = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "FID: %u", FID);

    }

    offset += 2; /* Skip FID */

    /* Build display for: Byte Count (BCC) */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count (BCC) */

  }

}

void
dissect_close_print_file_smb(const u_char *pd, int offset, frame_data *fd, proto_tree *parent, proto_tree *tree, smb_info si, int max_data, int SMB_offset, int errcode)

{
  guint8        WordCount;
  guint16       FID;
  guint16       ByteCount;

  if (si.request) {
    /* Request(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    /* Build display for: FID */

    FID = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "FID: %u", FID);

    }

    offset += 2; /* Skip FID */

    /* Build display for: Byte Count (BCC) */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count (BCC) */

  } else {
    /* Response(s) dissect code */

    /* Build display for: Word Count */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count: %u", WordCount);

    }

    offset += 1; /* Skip Word Count */

    /* Build display for: Byte Count (BCC) */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count (BCC) */

  }

}


void
dissect_seek_file_smb(const u_char *pd, int offset, frame_data *fd, proto_tree *parent, proto_tree *tree, smb_info si, int max_data, int SMB_offset, int errcode)

{
  static const value_string Mode_0x03[] = {
	{ 0, "Seek from start of file"},
	{ 1, "Seek from current position"},
	{ 2, "Seek from end of file"},
	{ 0, NULL}
  };
  proto_tree    *Mode_tree;
  proto_item    *ti;
  guint8        WordCount;
  guint32       Offset;
  guint16       Mode;
  guint16       FID;
  guint16       ByteCount;

  if (si.request) {
    /* Request(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    /* Build display for: FID */

    FID = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "FID: %u", FID);

    }

    offset += 2; /* Skip FID */

    /* Build display for: Mode */

    Mode = GSHORT(pd, offset);

    if (tree) {

      ti = proto_tree_add_text(tree, NullTVB, offset, 2, "Mode: 0x%02x", Mode);
      Mode_tree = proto_item_add_subtree(ti, ett_smb_mode);
      proto_tree_add_text(Mode_tree, NullTVB, offset, 2, "%s",
                          decode_enumerated_bitfield(Mode, 0x03, 16, Mode_0x03, "%s"));
    
}

    offset += 2; /* Skip Mode */

    /* Build display for: Offset */

    Offset = GWORD(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 4, "Offset: %u", Offset);

    }

    offset += 4; /* Skip Offset */

    /* Build display for: Byte Count (BCC) */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count (BCC) */

  } else {
    /* Response(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    /* Build display for: Offset */

    Offset = GWORD(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 4, "Offset: %u", Offset);

    }

    offset += 4; /* Skip Offset */

    /* Build display for: Byte Count (BCC) */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count (BCC) */

  }

}


void
dissect_get_print_queue_smb(const u_char *pd, int offset, frame_data *fd, proto_tree *parent, proto_tree *tree, smb_info si, int max_data, int SMB_offset, int errcode)

{
  guint8        WordCount;
  guint8        BufferFormat;
  guint16       StartIndex;
  guint16       RestartIndex;
  guint16       MaxCount;
  guint16       DataLength;
  guint16       Count;
  guint16       ByteCount;

  if (si.request) {
    /* Request(s) dissect code */

    /* Build display for: Word Count */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count: %u", WordCount);

    }

    offset += 1; /* Skip Word Count */

    /* Build display for: Max Count */

    MaxCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Max Count: %u", MaxCount);

    }

    offset += 2; /* Skip Max Count */

    /* Build display for: Start Index */

    StartIndex = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Start Index: %u", StartIndex);

    }

    offset += 2; /* Skip Start Index */

    /* Build display for: Byte Count (BCC) */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count (BCC) */

  } else {
    /* Response(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    if (WordCount > 0) {

      /* Build display for: Count */

      Count = GSHORT(pd, offset);

      if (tree) {

	proto_tree_add_text(tree, NullTVB, offset, 2, "Count: %u", Count);

      }

      offset += 2; /* Skip Count */

      /* Build display for: Restart Index */

      RestartIndex = GSHORT(pd, offset);

      if (tree) {

	proto_tree_add_text(tree, NullTVB, offset, 2, "Restart Index: %u", RestartIndex);

      }

      offset += 2; /* Skip Restart Index */

      /* Build display for: Byte Count (BCC) */

    }

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count (BCC) */

    /* Build display for: Buffer Format */

    BufferFormat = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Buffer Format: %s (%u)",
			  val_to_str(BufferFormat, buffer_format_vals, "Unknown"),
			  BufferFormat);

    }

    offset += 1; /* Skip Buffer Format */

    /* Build display for: Data Length */

    DataLength = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Data Length: %u", DataLength);

    }

    offset += 2; /* Skip Data Length */

  }

}

void
dissect_locking_andx_smb(const u_char *pd, int offset, frame_data *fd, proto_tree *parent, proto_tree *tree, smb_info si, int max_data, int SMB_offset, int errcode)

{
  proto_tree    *LockType_tree;
  proto_item    *ti;
  guint8        LockType;
  guint8        WordCount;
  guint8        OplockLevel;
  guint8        AndXReserved;
  guint8        AndXCommand = 0xFF;
  guint32       Timeout;
  guint16       NumberofLocks;
  guint16       NumberOfUnlocks;
  guint16       FID;
  guint16       ByteCount;
  guint16       AndXoffset;
  guint16       AndXOffset = 0;

  if (si.request) {
    /* Request(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    /* Build display for: AndXCommand */

    AndXCommand = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "AndXCommand: %u", AndXCommand);

    }

    offset += 1; /* Skip AndXCommand */

    /* Build display for: AndXReserved */

    AndXReserved = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "AndXReserved: %u", AndXReserved);

    }

    offset += 1; /* Skip AndXReserved */

    /* Build display for: AndXOffset */

    AndXOffset = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "AndXOffset: %u", AndXOffset);

    }

    offset += 2; /* Skip AndXOffset */

    /* Build display for: FID */

    FID = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "FID: %u", FID);

    }

    offset += 2; /* Skip FID */

    /* Build display for: Lock Type */

    LockType = GBYTE(pd, offset);

    if (tree) {

      ti = proto_tree_add_text(tree, NullTVB, offset, 1, "Lock Type: 0x%01x", LockType);
      LockType_tree = proto_item_add_subtree(ti, ett_smb_lock_type);
      proto_tree_add_text(LockType_tree, NullTVB, offset, 1, "%s",
                          decode_boolean_bitfield(LockType, 0x01, 16, "Read-only lock", "Not a Read-only lock"));
      proto_tree_add_text(LockType_tree, NullTVB, offset, 1, "%s",
                          decode_boolean_bitfield(LockType, 0x02, 16, "Oplock break notification", "Not an Oplock break notification"));
      proto_tree_add_text(LockType_tree, NullTVB, offset, 1, "%s",
                          decode_boolean_bitfield(LockType, 0x04, 16, "Change lock type", "Not a lock type change"));
      proto_tree_add_text(LockType_tree, NullTVB, offset, 1, "%s",
                          decode_boolean_bitfield(LockType, 0x08, 16, "Cancel outstanding request", "Dont cancel outstanding request"));
      proto_tree_add_text(LockType_tree, NullTVB, offset, 1, "%s",
                          decode_boolean_bitfield(LockType, 0x10, 16, "Large file locking format", "Not a large file locking format"));
    
}

    offset += 1; /* Skip Lock Type */

    /* Build display for: OplockLevel */

    OplockLevel = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "OplockLevel: %u", OplockLevel);

    }

    offset += 1; /* Skip OplockLevel */

    /* Build display for: Timeout */

    Timeout = GWORD(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 4, "Timeout: %u", Timeout);

    }

    offset += 4; /* Skip Timeout */

    /* Build display for: Number Of Unlocks */

    NumberOfUnlocks = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Number Of Unlocks: %u", NumberOfUnlocks);

    }

    offset += 2; /* Skip Number Of Unlocks */

    /* Build display for: Number of Locks */

    NumberofLocks = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Number of Locks: %u", NumberofLocks);

    }

    offset += 2; /* Skip Number of Locks */

    /* Build display for: Byte Count (BCC) */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count (BCC) */


    if (AndXCommand != 0xFF) {

      (dissect[AndXCommand])(pd, SMB_offset + AndXOffset, fd, parent, tree, si, max_data, SMB_offset, errcode);

    }

  } else {
    /* Response(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    if (WordCount > 0) {

      /* Build display for: AndXCommand */

      AndXCommand = GBYTE(pd, offset);

      if (tree) {

	proto_tree_add_text(tree, NullTVB, offset, 1, "AndXCommand: %s", 
			    (AndXCommand == 0xFF ? "No further commands" : decode_smb_name(AndXCommand)));

      }

      offset += 1; /* Skip AndXCommand */

      /* Build display for: AndXReserved */

      AndXReserved = GBYTE(pd, offset);

      if (tree) {

	proto_tree_add_text(tree, NullTVB, offset, 1, "AndXReserved: %u", AndXReserved);

      }

      offset += 1; /* Skip AndXReserved */

      /* Build display for: AndXoffset */

      AndXoffset = GSHORT(pd, offset);

      if (tree) {

	proto_tree_add_text(tree, NullTVB, offset, 2, "AndXoffset: %u", AndXoffset);

      }

      offset += 2; /* Skip AndXoffset */

    }

    /* Build display for: Byte Count */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count: %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count */


    if (AndXCommand != 0xFF) {

      (dissect[AndXCommand])(pd, SMB_offset + AndXOffset, fd, parent, tree, si, max_data, SMB_offset, errcode);

    }

  }

}

void
dissect_search_dir_smb(const u_char *pd, int offset, frame_data *fd, proto_tree *parent, proto_tree *tree, smb_info si, int max_data, int SMB_offset, int errcode)

{
  guint8        WordCount;
  guint8        BufferFormat2;
  guint8        BufferFormat1;
  guint8        BufferFormat;
  guint16       SearchAttributes;
  guint16       ResumeKeyLength;
  guint16       MaxCount;
  guint16       DataLength;
  guint16       Count;
  guint16       ByteCount;
  const char    *FileName;
  int           string_len;

  if (si.request) {
    /* Request(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    /* Build display for: Max Count */

    MaxCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Max Count: %u", MaxCount);

    }

    offset += 2; /* Skip Max Count */

    /* Build display for: Search Attributes */

    SearchAttributes = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Search Attributes: %u", SearchAttributes);

    }

    offset += 2; /* Skip Search Attributes */

    /* Build display for: Byte Count (BCC) */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count (BCC) */

    /* Build display for: Buffer Format 1 */

    BufferFormat1 = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Buffer Format 1: %s (%u)",
			  val_to_str(BufferFormat1, buffer_format_vals, "Unknown"),
			  BufferFormat1);

    }

    offset += 1; /* Skip Buffer Format 1 */

    /* Build display for: File Name */

    FileName = get_unicode_or_ascii_string(pd, &offset, si.unicode, &string_len);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, string_len, "File Name: %s", FileName);

    }

    offset += string_len; /* Skip File Name */

    /* Build display for: Buffer Format 2 */

    BufferFormat2 = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Buffer Format 2: %s (%u)",
			  val_to_str(BufferFormat2, buffer_format_vals, "Unknown"),
			  BufferFormat2);

    }

    offset += 1; /* Skip Buffer Format 2 */

    /* Build display for: Resume Key Length */

    ResumeKeyLength = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Resume Key Length: %u", ResumeKeyLength);

    }

    offset += 2; /* Skip Resume Key Length */

  } else {
    /* Response(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    if (WordCount > 0) {

      /* Build display for: Count */

      Count = GSHORT(pd, offset);

      if (tree) {

	proto_tree_add_text(tree, NullTVB, offset, 2, "Count: %u", Count);

      }

      offset += 2; /* Skip Count */

    }

    /* Build display for: Byte Count (BCC) */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count (BCC) */

    /* Build display for: Buffer Format */

    BufferFormat = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Buffer Format: %s (%u)",
			  val_to_str(BufferFormat, buffer_format_vals, "Unknown"),
			  BufferFormat);

    }

    offset += 1; /* Skip Buffer Format */

    /* Build display for: Data Length */

    DataLength = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Data Length: %u", DataLength);

    }

    offset += 2; /* Skip Data Length */

  }

}

void
dissect_write_print_file_smb(const u_char *pd, int offset, frame_data *fd, proto_tree *parent, proto_tree *tree, smb_info si, int max_data, int SMB_offset, int errcode)

{
  guint8        WordCount;
  guint8        BufferFormat;
  guint16       FID;
  guint16       DataLength;
  guint16       ByteCount;

  if (si.request) {
    /* Request(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    /* Build display for: FID */

    FID = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "FID: %u", FID);

    }

    offset += 2; /* Skip FID */

    /* Build display for: Byte Count (BCC) */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count (BCC) */

    /* Build display for: Buffer Format */

    BufferFormat = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Buffer Format: %s (%u)",
			  val_to_str(BufferFormat, buffer_format_vals, "Unknown"),
			  BufferFormat);

    }

    offset += 1; /* Skip Buffer Format */

    /* Build display for: Data Length */

    DataLength = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Data Length: %u", DataLength);

    }

    offset += 2; /* Skip Data Length */

  } else {
    /* Response(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    /* Build display for: Byte Count (BCC) */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count (BCC) */

  }

}


void
dissect_write_mpx_smb(const u_char *pd, int offset, frame_data *fd, proto_tree *parent, proto_tree *tree, smb_info si, int max_data, int SMB_offset, int errcode)

{
  proto_tree    *WriteMode_tree;
  proto_item    *ti;
  guint8        WordCount;
  guint8        Pad;
  guint32       Timeout;
  guint32       ResponseMask;
  guint32       RequestMask;
  guint16       WriteMode;
  guint16       Reserved1;
  guint16       FID;
  guint16       DataOffset;
  guint16       DataLength;
  guint16       Count;
  guint16       ByteCount;

  if (si.request) {
    /* Request(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    /* Build display for: FID */

    FID = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "FID: %u", FID);

    }

    offset += 2; /* Skip FID */

    /* Build display for: Count */

    Count = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Count: %u", Count);

    }

    offset += 2; /* Skip Count */

    /* Build display for: Reserved 1 */

    Reserved1 = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Reserved 1: %u", Reserved1);

    }

    offset += 2; /* Skip Reserved 1 */

    /* Build display for: Timeout */

    Timeout = GWORD(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 4, "Timeout: %u", Timeout);

    }

    offset += 4; /* Skip Timeout */

    /* Build display for: WriteMode */

    WriteMode = GSHORT(pd, offset);

    if (tree) {

      ti = proto_tree_add_text(tree, NullTVB, offset, 2, "WriteMode: 0x%02x", WriteMode);
      WriteMode_tree = proto_item_add_subtree(ti, ett_smb_writemode);
      proto_tree_add_text(WriteMode_tree, NullTVB, offset, 2, "%s",
                          decode_boolean_bitfield(WriteMode, 0x01, 16, "Write through requested", "Write through not requested"));
      proto_tree_add_text(WriteMode_tree, NullTVB, offset, 2, "%s",
                          decode_boolean_bitfield(WriteMode, 0x02, 16, "Return Remaining", "Dont return Remaining"));
      proto_tree_add_text(WriteMode_tree, NullTVB, offset, 2, "%s",
                          decode_boolean_bitfield(WriteMode, 0x40, 16, "Connectionless mode requested", "Connectionless mode not requested"));
    
}

    offset += 2; /* Skip WriteMode */

    /* Build display for: Request Mask */

    RequestMask = GWORD(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 4, "Request Mask: %u", RequestMask);

    }

    offset += 4; /* Skip Request Mask */

    /* Build display for: Data Length */

    DataLength = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Data Length: %u", DataLength);

    }

    offset += 2; /* Skip Data Length */

    /* Build display for: Data Offset */

    DataOffset = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Data Offset: %u", DataOffset);

    }

    offset += 2; /* Skip Data Offset */

    /* Build display for: Byte Count (BCC) */

    ByteCount = GSHORT(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

    }

    offset += 2; /* Skip Byte Count (BCC) */

    /* Build display for: Pad */

    Pad = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Pad: %u", Pad);

    }

    offset += 1; /* Skip Pad */

  } else {
    /* Response(s) dissect code */

    /* Build display for: Word Count (WCT) */

    WordCount = GBYTE(pd, offset);

    if (tree) {

      proto_tree_add_text(tree, NullTVB, offset, 1, "Word Count (WCT): %u", WordCount);

    }

    offset += 1; /* Skip Word Count (WCT) */

    if (WordCount > 0) {

      /* Build display for: Response Mask */

      ResponseMask = GWORD(pd, offset);

      if (tree) {

	proto_tree_add_text(tree, NullTVB, offset, 4, "Response Mask: %u", ResponseMask);

      }

      offset += 4; /* Skip Response Mask */

      /* Build display for: Byte Count (BCC) */

      ByteCount = GSHORT(pd, offset);

      if (tree) {

	proto_tree_add_text(tree, NullTVB, offset, 2, "Byte Count (BCC): %u", ByteCount);

      }

    }

    offset += 2; /* Skip Byte Count (BCC) */

  }

}



static void (*dissect[256])(const u_char *, int, frame_data *, proto_tree *, proto_tree *, smb_info, int, int, int) = {

  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb, 
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,      /*0x0f */
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,      /* unknown SMB 0x15 */
  dissect_unknown_smb,      /* unknown SMB 0x16 */
  dissect_unknown_smb,      /* unknown SMB 0x17 */
  dissect_unknown_smb,      /* unknown SMB 0x18 */
  dissect_unknown_smb,      /* unknown SMB 0x19 */
  dissect_unknown_smb,      /* 0x1A */
  dissect_unknown_smb,
  dissect_unknown_smb,      /* SMBreadBs read block (secondary response) */
  dissect_write_raw_smb,    /* SMBwriteBraw write block raw */
  dissect_write_mpx_smb,    /* SMBwriteBmpx write block multiplexed */
  dissect_unknown_smb,      /* SMBwriteBs write block (secondary request) */
  dissect_unknown_smb,      /* SMBwriteC write complete response */
  dissect_unknown_smb,      /* unknown SMB 0x21 */
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_locking_andx_smb, /* SMBlockingX lock/unlock byte ranges and X */
  dissect_unknown_smb,
  dissect_unknown_smb,      /* SMBtranss transaction (secondary request/response) */
  dissect_unknown_smb,      /* SMBioctl IOCTL */
  dissect_unknown_smb,      /* SMBioctls IOCTL (secondary request/response) */
  dissect_unknown_smb,      /* SMBcopy copy */
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,      /* SMBwriteX write and X */
  dissect_unknown_smb,      /* unknown SMB 0x30 */
  dissect_unknown_smb,      /* unknown SMB 0x31 */
  dissect_unknown_smb,
  dissect_unknown_smb,      /* unknown SMB 0x33 */
  dissect_unknown_smb,
  dissect_unknown_smb,      /* unknown SMB 0x35 */
  dissect_unknown_smb,      /* unknown SMB 0x36 */
  dissect_unknown_smb,      /* unknown SMB 0x37 */
  dissect_unknown_smb,      /* unknown SMB 0x38 */
  dissect_unknown_smb,      /* unknown SMB 0x39 */
  dissect_unknown_smb,      /* unknown SMB 0x3a */
  dissect_unknown_smb,      /* unknown SMB 0x3b */
  dissect_unknown_smb,      /* unknown SMB 0x3c */
  dissect_unknown_smb,      /* unknown SMB 0x3d */
  dissect_unknown_smb,      /* unknown SMB 0x3e */
  dissect_unknown_smb,      /* unknown SMB 0x3f */
  dissect_unknown_smb,      /* unknown SMB 0x40 */
  dissect_unknown_smb,      /* unknown SMB 0x41 */
  dissect_unknown_smb,      /* unknown SMB 0x42 */
  dissect_unknown_smb,      /* unknown SMB 0x43 */
  dissect_unknown_smb,      /* unknown SMB 0x44 */
  dissect_unknown_smb,      /* unknown SMB 0x45 */
  dissect_unknown_smb,      /* unknown SMB 0x46 */
  dissect_unknown_smb,      /* unknown SMB 0x47 */
  dissect_unknown_smb,      /* unknown SMB 0x48 */
  dissect_unknown_smb,      /* unknown SMB 0x49 */
  dissect_unknown_smb,      /* unknown SMB 0x4a */
  dissect_unknown_smb,      /* unknown SMB 0x4b */
  dissect_unknown_smb,      /* unknown SMB 0x4c */
  dissect_unknown_smb,      /* unknown SMB 0x4d */
  dissect_unknown_smb,      /* unknown SMB 0x4e */
  dissect_unknown_smb,      /* unknown SMB 0x4f */
  dissect_unknown_smb,      /* unknown SMB 0x50 */
  dissect_unknown_smb,      /* unknown SMB 0x51 */
  dissect_unknown_smb,      /* unknown SMB 0x52 */
  dissect_unknown_smb,      /* unknown SMB 0x53 */
  dissect_unknown_smb,      /* unknown SMB 0x54 */
  dissect_unknown_smb,      /* unknown SMB 0x55 */
  dissect_unknown_smb,      /* unknown SMB 0x56 */
  dissect_unknown_smb,      /* unknown SMB 0x57 */
  dissect_unknown_smb,      /* unknown SMB 0x58 */
  dissect_unknown_smb,      /* unknown SMB 0x59 */
  dissect_unknown_smb,      /* unknown SMB 0x5a */
  dissect_unknown_smb,      /* unknown SMB 0x5b */
  dissect_unknown_smb,      /* unknown SMB 0x5c */
  dissect_unknown_smb,      /* unknown SMB 0x5d */
  dissect_unknown_smb,      /* unknown SMB 0x5e */
  dissect_unknown_smb,      /* unknown SMB 0x5f */
  dissect_unknown_smb,      /* unknown SMB 0x60 */
  dissect_unknown_smb,      /* unknown SMB 0x61 */
  dissect_unknown_smb,      /* unknown SMB 0x62 */
  dissect_unknown_smb,      /* unknown SMB 0x63 */
  dissect_unknown_smb,      /* unknown SMB 0x64 */
  dissect_unknown_smb,      /* unknown SMB 0x65 */
  dissect_unknown_smb,      /* unknown SMB 0x66 */
  dissect_unknown_smb,      /* unknown SMB 0x67 */
  dissect_unknown_smb,      /* unknown SMB 0x68 */
  dissect_unknown_smb,      /* unknown SMB 0x69 */
  dissect_unknown_smb,      /* unknown SMB 0x6a */
  dissect_unknown_smb,      /* unknown SMB 0x6b */
  dissect_unknown_smb,      /* unknown SMB 0x6c */
  dissect_unknown_smb,      /* unknown SMB 0x6d */
  dissect_unknown_smb,      /* unknown SMB 0x6e */
  dissect_unknown_smb,      /* unknown SMB 0x6f */
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,
  dissect_unknown_smb,      /* unknown SMB 0x76 */
  dissect_unknown_smb,      /* unknown SMB 0x77 */
  dissect_unknown_smb,      /* unknown SMB 0x78 */
  dissect_unknown_smb,      /* unknown SMB 0x79 */
  dissect_unknown_smb,      /* unknown SMB 0x7a */
  dissect_unknown_smb,      /* unknown SMB 0x7b */
  dissect_unknown_smb,      /* unknown SMB 0x7c */
  dissect_unknown_smb,      /* unknown SMB 0x7d */
  dissect_unknown_smb,      /* unknown SMB 0x7e */
  dissect_unknown_smb,      /* unknown SMB 0x7f */
  dissect_unknown_smb,
  dissect_search_dir_smb,   /* SMBsearch search a directory */
  dissect_unknown_smb,      /* SMBffirst find first */
  dissect_unknown_smb,      /* SMBfunique find unique */
  dissect_unknown_smb,      /* SMBfclose find close */
  dissect_unknown_smb,      /* unknown SMB 0x85 */
  dissect_unknown_smb,      /* unknown SMB 0x86 */
  dissect_unknown_smb,      /* unknown SMB 0x87 */
  dissect_unknown_smb,      /* unknown SMB 0x88 */
  dissect_unknown_smb,      /* unknown SMB 0x89 */
  dissect_unknown_smb,      /* unknown SMB 0x8a */
  dissect_unknown_smb,      /* unknown SMB 0x8b */
  dissect_unknown_smb,      /* unknown SMB 0x8c */
  dissect_unknown_smb,      /* unknown SMB 0x8d */
  dissect_unknown_smb,      /* unknown SMB 0x8e */
  dissect_unknown_smb,      /* unknown SMB 0x8f */
  dissect_unknown_smb,      /* unknown SMB 0x90 */
  dissect_unknown_smb,      /* unknown SMB 0x91 */
  dissect_unknown_smb,      /* unknown SMB 0x92 */
  dissect_unknown_smb,      /* unknown SMB 0x93 */
  dissect_unknown_smb,      /* unknown SMB 0x94 */
  dissect_unknown_smb,      /* unknown SMB 0x95 */
  dissect_unknown_smb,      /* unknown SMB 0x96 */
  dissect_unknown_smb,      /* unknown SMB 0x97 */
  dissect_unknown_smb,      /* unknown SMB 0x98 */
  dissect_unknown_smb,      /* unknown SMB 0x99 */
  dissect_unknown_smb,      /* unknown SMB 0x9a */
  dissect_unknown_smb,      /* unknown SMB 0x9b */
  dissect_unknown_smb,      /* unknown SMB 0x9c */
  dissect_unknown_smb,      /* unknown SMB 0x9d */
  dissect_unknown_smb,      /* unknown SMB 0x9e */
  dissect_unknown_smb,      /* unknown SMB 0x9f */
  dissect_unknown_smb,      /* unknown SMB 0xa0 */
  dissect_unknown_smb,      /* unknown SMB 0xa1 */
  dissect_unknown_smb,      /* unknown SMB 0xa2 */
  dissect_unknown_smb,      /* unknown SMB 0xa3 */
  dissect_unknown_smb,      /* unknown SMB 0xa4 */
  dissect_unknown_smb,      /* unknown SMB 0xa5 */
  dissect_unknown_smb,      /* unknown SMB 0xa6 */
  dissect_unknown_smb,      /* unknown SMB 0xa7 */
  dissect_unknown_smb,      /* unknown SMB 0xa8 */
  dissect_unknown_smb,      /* unknown SMB 0xa9 */
  dissect_unknown_smb,      /* unknown SMB 0xaa */
  dissect_unknown_smb,      /* unknown SMB 0xab */
  dissect_unknown_smb,      /* unknown SMB 0xac */
  dissect_unknown_smb,      /* unknown SMB 0xad */
  dissect_unknown_smb,      /* unknown SMB 0xae */
  dissect_unknown_smb,      /* unknown SMB 0xaf */
  dissect_unknown_smb,      /* unknown SMB 0xb0 */
  dissect_unknown_smb,      /* unknown SMB 0xb1 */
  dissect_unknown_smb,      /* unknown SMB 0xb2 */
  dissect_unknown_smb,      /* unknown SMB 0xb3 */
  dissect_unknown_smb,      /* unknown SMB 0xb4 */
  dissect_unknown_smb,      /* unknown SMB 0xb5 */
  dissect_unknown_smb,      /* unknown SMB 0xb6 */
  dissect_unknown_smb,      /* unknown SMB 0xb7 */
  dissect_unknown_smb,      /* unknown SMB 0xb8 */
  dissect_unknown_smb,      /* unknown SMB 0xb9 */
  dissect_unknown_smb,      /* unknown SMB 0xba */
  dissect_unknown_smb,      /* unknown SMB 0xbb */
  dissect_unknown_smb,      /* unknown SMB 0xbc */
  dissect_unknown_smb,      /* unknown SMB 0xbd */
  dissect_unknown_smb,      /* unknown SMB 0xbe */
  dissect_unknown_smb,      /* unknown SMB 0xbf */
  dissect_unknown_smb,      /* SMBsplopen open a print spool file */
  dissect_write_print_file_smb,/* SMBsplwr write to a print spool file */
  dissect_close_print_file_smb,/* SMBsplclose close a print spool file */
  dissect_get_print_queue_smb, /* SMBsplretq return print queue */
  dissect_unknown_smb,      /* unknown SMB 0xc4 */
  dissect_unknown_smb,      /* unknown SMB 0xc5 */
  dissect_unknown_smb,      /* unknown SMB 0xc6 */
  dissect_unknown_smb,      /* unknown SMB 0xc7 */
  dissect_unknown_smb,      /* unknown SMB 0xc8 */
  dissect_unknown_smb,      /* unknown SMB 0xc9 */
  dissect_unknown_smb,      /* unknown SMB 0xca */
  dissect_unknown_smb,      /* unknown SMB 0xcb */
  dissect_unknown_smb,      /* unknown SMB 0xcc */
  dissect_unknown_smb,      /* unknown SMB 0xcd */
  dissect_unknown_smb,      /* unknown SMB 0xce */
  dissect_unknown_smb,      /* unknown SMB 0xcf */
  dissect_unknown_smb,      /* SMBsends send a single block message */
  dissect_unknown_smb,      /* SMBsendb send a broadcast message */
  dissect_unknown_smb,      /* SMBfwdname forward user name */
  dissect_unknown_smb,      /* SMBcancelf cancel forward */
  dissect_unknown_smb,      /* SMBgetmac get a machine name */
  dissect_unknown_smb,      /* SMBsendstrt send start of multi-block message */
  dissect_unknown_smb,      /* SMBsendend send end of multi-block message */
  dissect_unknown_smb,      /* SMBsendtxt send text of multi-block message */
  dissect_unknown_smb,      /* unknown SMB 0xd8 */
  dissect_unknown_smb,      /* unknown SMB 0xd9 */
  dissect_unknown_smb,      /* unknown SMB 0xda */
  dissect_unknown_smb,      /* unknown SMB 0xdb */
  dissect_unknown_smb,      /* unknown SMB 0xdc */
  dissect_unknown_smb,      /* unknown SMB 0xdd */
  dissect_unknown_smb,      /* unknown SMB 0xde */
  dissect_unknown_smb,      /* unknown SMB 0xdf */
  dissect_unknown_smb,      /* unknown SMB 0xe0 */
  dissect_unknown_smb,      /* unknown SMB 0xe1 */
  dissect_unknown_smb,      /* unknown SMB 0xe2 */
  dissect_unknown_smb,      /* unknown SMB 0xe3 */
  dissect_unknown_smb,      /* unknown SMB 0xe4 */
  dissect_unknown_smb,      /* unknown SMB 0xe5 */
  dissect_unknown_smb,      /* unknown SMB 0xe6 */
  dissect_unknown_smb,      /* unknown SMB 0xe7 */
  dissect_unknown_smb,      /* unknown SMB 0xe8 */
  dissect_unknown_smb,      /* unknown SMB 0xe9 */
  dissect_unknown_smb,      /* unknown SMB 0xea */
  dissect_unknown_smb,      /* unknown SMB 0xeb */
  dissect_unknown_smb,      /* unknown SMB 0xec */
  dissect_unknown_smb,      /* unknown SMB 0xed */
  dissect_unknown_smb,      /* unknown SMB 0xee */
  dissect_unknown_smb,      /* unknown SMB 0xef */
  dissect_unknown_smb,      /* unknown SMB 0xf0 */
  dissect_unknown_smb,      /* unknown SMB 0xf1 */
  dissect_unknown_smb,      /* unknown SMB 0xf2 */
  dissect_unknown_smb,      /* unknown SMB 0xf3 */
  dissect_unknown_smb,      /* unknown SMB 0xf4 */
  dissect_unknown_smb,      /* unknown SMB 0xf5 */
  dissect_unknown_smb,      /* unknown SMB 0xf6 */
  dissect_unknown_smb,      /* unknown SMB 0xf7 */
  dissect_unknown_smb,      /* unknown SMB 0xf8 */
  dissect_unknown_smb,      /* unknown SMB 0xf9 */
  dissect_unknown_smb,      /* unknown SMB 0xfa */
  dissect_unknown_smb,      /* unknown SMB 0xfb */
  dissect_unknown_smb,      /* unknown SMB 0xfc */
  dissect_unknown_smb,      /* unknown SMB 0xfd */
  dissect_unknown_smb,      /* SMBinvalid invalid command */
  dissect_unknown_smb       /* unknown SMB 0xff */

};

static const value_string errcls_types[] = {
  { SMB_SUCCESS, "Success"},
  { SMB_ERRDOS, "DOS Error"},
  { SMB_ERRSRV, "Server Error"},
  { SMB_ERRHRD, "Hardware Error"},
  { SMB_ERRCMD, "Command Error - Not an SMB format command"},
  { 0, NULL }
};

char *decode_smb_name(unsigned char cmd)
{

  return(smb_cmd_vals[cmd].strptr);

}

static const value_string DOS_errors[] = {
  {SMBE_badfunc, "Invalid function (or system call)"},
  {SMBE_badfile, "File not found (pathname error)"},
  {SMBE_badpath, "Directory not found"},
  {SMBE_nofids, "Too many open files"},
  {SMBE_noaccess, "Access denied"},
  {SMBE_badfid, "Invalid fid"},
  {SMBE_nomem,  "Out of memory"},
  {SMBE_badmem, "Invalid memory block address"},
  {SMBE_badenv, "Invalid environment"},
  {SMBE_badaccess, "Invalid open mode"},
  {SMBE_baddata, "Invalid data (only from ioctl call)"},
  {SMBE_res, "Reserved error code?"}, 
  {SMBE_baddrive, "Invalid drive"},
  {SMBE_remcd, "Attempt to delete current directory"},
  {SMBE_diffdevice, "Rename/move across different filesystems"},
  {SMBE_nofiles, "no more files found in file search"},
  {SMBE_badshare, "Share mode on file conflict with open mode"},
  {SMBE_lock, "Lock request conflicts with existing lock"},
  {SMBE_unsup, "Request unsupported, returned by Win 95"},
  {SMBE_nosuchshare, "Requested share does not exist"},
  {SMBE_filexists, "File in operation already exists"},
  {SMBE_cannotopen, "Cannot open the file specified"},
  {SMBE_unknownlevel, "Unknown level??"},
  {SMBE_badpipe, "Named pipe invalid"},
  {SMBE_pipebusy, "All instances of pipe are busy"},
  {SMBE_pipeclosing, "Named pipe close in progress"},
  {SMBE_notconnected, "No process on other end of named pipe"},
  {SMBE_moredata, "More data to be returned"},
  {SMBE_baddirectory,  "Invalid directory name in a path."},
  {SMBE_eas_didnt_fit, "Extended attributes didn't fit"},
  {SMBE_eas_nsup, "Extended attributes not supported"},
  {SMBE_notify_buf_small, "Buffer too small to return change notify."},
  {SMBE_unknownipc, "Unknown IPC Operation"},
  {SMBE_noipc, "Don't support ipc"},
  {0, NULL}
  };

/* Error codes for the ERRSRV class */

static const value_string SRV_errors[] = {
  {SMBE_error, "Non specific error code"},
  {SMBE_badpw, "Bad password"},
  {SMBE_badtype, "Reserved"},
  {SMBE_access, "No permissions to perform the requested operation"},
  {SMBE_invnid, "TID invalid"},
  {SMBE_invnetname, "Invalid network name. Service not found"},
  {SMBE_invdevice, "Invalid device"},
  {SMBE_unknownsmb, "Unknown SMB, from NT 3.5 response"},
  {SMBE_qfull, "Print queue full"},
  {SMBE_qtoobig, "Queued item too big"},
  {SMBE_qeof, "EOF on print queue dump"},
  {SMBE_invpfid, "Invalid print file in smb_fid"},
  {SMBE_smbcmd, "Unrecognised command"},
  {SMBE_srverror, "SMB server internal error"},
  {SMBE_filespecs, "Fid and pathname invalid combination"},
  {SMBE_badlink, "Bad link in request ???"},
  {SMBE_badpermits, "Access specified for a file is not valid"},
  {SMBE_badpid, "Bad process id in request"},
  {SMBE_setattrmode, "Attribute mode invalid"},
  {SMBE_paused, "Message server paused"},
  {SMBE_msgoff, "Not receiving messages"},
  {SMBE_noroom, "No room for message"},
  {SMBE_rmuns, "Too many remote usernames"},
  {SMBE_timeout, "Operation timed out"},
  {SMBE_noresource, "No resources currently available for request."},
  {SMBE_toomanyuids, "Too many userids"},
  {SMBE_baduid, "Bad userid"},
  {SMBE_useMPX, "Temporarily unable to use raw mode, use MPX mode"},
  {SMBE_useSTD, "Temporarily unable to use raw mode, use standard mode"},
  {SMBE_contMPX, "Resume MPX mode"},
  {SMBE_badPW, "Bad Password???"},
  {SMBE_nosupport, "Operation not supported"},
  { 0, NULL}
};

/* Error codes for the ERRHRD class */

static const value_string HRD_errors[] = {
  {SMBE_nowrite, "read only media"},
  {SMBE_badunit, "Unknown device"},
  {SMBE_notready, "Drive not ready"},
  {SMBE_badcmd, "Unknown command"},
  {SMBE_data, "Data (CRC) error"},
  {SMBE_badreq, "Bad request structure length"},
  {SMBE_seek, "Seek error???"},
  {SMBE_badmedia, "Bad media???"},
  {SMBE_badsector, "Bad sector???"},
  {SMBE_nopaper, "No paper in printer???"},
  {SMBE_write, "Write error???"},
  {SMBE_read, "Read error???"},
  {SMBE_general, "General error???"},
  {SMBE_badshare, "A open conflicts with an existing open"},
  {SMBE_lock, "Lock/unlock error"},
  {SMBE_wrongdisk,  "Wrong disk???"},
  {SMBE_FCBunavail, "FCB unavailable???"},
  {SMBE_sharebufexc, "Share buffer excluded???"},
  {SMBE_diskfull, "Disk full???"},
  {0, NULL}
};

char *decode_smb_error(guint8 errcls, guint16 errcode)
{

  switch (errcls) {

  case SMB_SUCCESS:

    return("No Error");   /* No error ??? */
    break;

  case SMB_ERRDOS:

    return(val_to_str(errcode, DOS_errors, "Unknown DOS error (%x)"));
    break;

  case SMB_ERRSRV:

    return(val_to_str(errcode, SRV_errors, "Unknown SRV error (%x)"));
    break;

  case SMB_ERRHRD:

    return(val_to_str(errcode, HRD_errors, "Unknown HRD error (%x)"));
    break;

  default:

    return("Unknown error class!");

  }

}

/* XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
   These variables and functions are used to match
   responses with calls
   XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX */
static GMemChunk *smb_info_chunk = NULL;
static int smb_info_init_count = 200;
static GHashTable *smb_info_table = NULL;

static gint
smb_info_equal(gconstpointer k1, gconstpointer k2)
{
	smb_info *key1 = (smb_info *)k1;
	smb_info *key2 = (smb_info *)k2;
	gint res;

	/* make sure to always compare mid first since this is most
	   likely to differ ==> shortcircuiting the expression */
	res= (	(key1->mid==key2->mid)
		&&	(key1->pid==key2->pid)
		&&	(key1->uid==key2->uid)
		&&	(key1->cmd==key2->cmd)
		&&	(key1->tid==key2->tid)
		&&	(ADDRESSES_EQUAL(key1->src, key2->src))
		&&	(ADDRESSES_EQUAL(key1->dst, key2->dst)) );

	return res;
}

static guint
smb_info_hash(gconstpointer k)
{
	smb_info *key = (smb_info *)k;

	/* multiplex id is very likely to differ between calls
	   it should be sufficient for a good distribution of hash
	   values.
	*/
	return key->mid;
}

static gboolean
free_all_smb_info(gpointer key_arg, gpointer value, gpointer user_data)
{
	smb_info *key = (smb_info *)key_arg;

	if((key->src)&&(key->src->data)){
		g_free((gpointer)key->src->data);
		key->src->data = NULL;
		g_free((gpointer)key->src);
		key->src = NULL;
	}

	if((key->dst)&&(key->dst->data)){
		g_free((gpointer)key->dst->data);
		key->dst->data = NULL;
		g_free((gpointer)key->dst);
		key->dst = NULL;
	}
 
	return TRUE;
}

void
smb_info_init(void)
{
	if(smb_info_table){
		g_hash_table_foreach_remove(smb_info_table,
			free_all_smb_info, NULL);
	} else {
		smb_info_table = g_hash_table_new(smb_info_hash,
			smb_info_equal);
	}

	if(smb_info_chunk){
		g_mem_chunk_destroy(smb_info_chunk);
	}
	smb_info_chunk = g_mem_chunk_new("smb_info_chunk",
		sizeof(smb_info),
		smb_info_init_count*sizeof(smb_info),
		G_ALLOC_ONLY);
}




#define SMB_FLAGS_DIRN 0x80

void
dissect_smb(const u_char *pd, int offset, frame_data *fd, proto_tree *tree, int max_data)
{
	proto_tree      *smb_tree = NULL, *smb_hdr_tree = NULL, *flags_tree, *flags2_tree;
	proto_item      *ti=NULL, *tf=NULL, *th=NULL;
	guint8          errcls, errcode1, flags;
	guint16         flags2, errcode;
	guint32         status;
	int             SMB_offset = offset;
	smb_info 	si;
	smb_info	*sip;
	int old_offset = offset;
	proto_item *cmd_item = NULL;
	proto_tree *cmd_tree = NULL;
	int (*new_dissector)(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, int offset, proto_tree *smb_tree);

	/* XXX need this for lanman and mailslot until we are completely
	   tvbuffified */
	top_tree = tree;

	/* start off using the local variable, we will allocate a new one if we
	   need to*/
	sip = &si;
	sip->mid = GSHORT(pd, 30+SMB_offset);
	sip->uid = GSHORT(pd, 28+SMB_offset);
	sip->pid = GSHORT(pd, 26+SMB_offset);
	sip->tid = GSHORT(pd, 24+SMB_offset);
	flags2 = GSHORT(pd,   10+SMB_offset);
	if(flags2 & 0x8000){
		sip->unicode = TRUE; /* Mark them as Unicode */
	} else {
		sip->unicode = FALSE;
	}
	flags = pd[9+SMB_offset];
	sip->request = !(flags&SMB_FLAGS_DIRN);
	sip->cmd = pd[4+SMB_offset];
	sip->subcmd = 0;
	sip->extra_type = 0;
	sip->extra = NULL;

	OLD_CHECK_DISPLAY_AS_DATA(proto_smb, pd, offset, fd, tree);

/*    	sip->ddisp = 0;*/

	if (check_col(fd, COL_PROTOCOL))
		col_set_str(fd, COL_PROTOCOL, "SMB");

	/* Hmmm, poor coding here ... Also, should check the type */

	if (check_col(fd, COL_INFO)) {

	  col_add_fstr(fd, COL_INFO, "%s %s", decode_smb_name(sip->cmd), (pi.match_port == pi.destport)? "Request" : "Response");

	}

	if (tree) {

	  ti = proto_tree_add_item(tree, proto_smb, NullTVB, offset, END_OF_FRAME, FALSE);
	  smb_tree = proto_item_add_subtree(ti, ett_smb);

	  th = proto_tree_add_text(smb_tree, NullTVB, offset, 32, "SMB Header");

	  smb_hdr_tree = proto_item_add_subtree(th, ett_smb_hdr);

	  /* 0xFFSMB is actually a 1 byte msg type and 3 byte server
	   * component ... SMB is only one used
	   */

	  proto_tree_add_text(smb_hdr_tree, NullTVB, offset, 1, "Message Type: 0xFF");
	  proto_tree_add_text(smb_hdr_tree, NullTVB, offset+1, 3, "Server Component: SMB");

	}

	offset += 4;  /* Skip the marker */


	/* store smb_info structure so we can retreive it from the reply */
	if(sip->request){
		sip->src = &pi.src;
		sip->dst = &pi.dst;
		if(!pi.fd->flags.visited){
			if( (sip->mid==0)
			&&  (sip->uid==0)
			&&  (sip->pid==0)
			&&  (sip->tid==0) ){
				/* this is a broadcast SMB packet,
				   there will not be a reply.
				   We dont need to do anything */
				sip->unidir=FALSE;
			} else {
				sip->unidir=TRUE;
				sip = g_mem_chunk_alloc(smb_info_chunk);
				memcpy(sip, &si, sizeof(smb_info));
				sip->frame_req = pi.fd->num;
				sip->frame_res = 0;
				sip->src=g_malloc(sizeof(address));
				COPY_ADDRESS(sip->src, &pi.src);
				sip->dst=g_malloc(sizeof(address));
				COPY_ADDRESS(sip->dst, &pi.dst);
				g_hash_table_insert(smb_info_table, sip, sip);
			}
		}
	} else {
		sip->src = &pi.dst;
		sip->dst = &pi.src;
	}
	sip = g_hash_table_lookup(smb_info_table, sip);
	if(!sip){
		sip = &si;
	}
	/* need to redo these ones, might have changed if we got a new sip */
	sip->request = !(flags&SMB_FLAGS_DIRN);
	if(flags2 & 0x8000){
		sip->unicode = TRUE; /* Mark them as Unicode */
	} else {
		sip->unicode = FALSE;
	}

	if(sip->request){
		if(sip->frame_res){
			proto_tree_add_uint(smb_hdr_tree, hf_smb_response_in, NullTVB, 0, 0, sip->frame_res);
		}
	} else {
		if(!pi.fd->flags.visited){
			sip->frame_res=pi.fd->num;
		}
		if(sip->frame_req){
			proto_tree_add_uint(smb_hdr_tree, hf_smb_response_to, NullTVB, 0, 0, sip->frame_req);
		}
	}

	if (tree) {

	  proto_tree_add_uint_format(smb_hdr_tree, hf_smb_cmd, NullTVB, offset, 1, sip->cmd, "SMB Command: %s (0x%02x)", decode_smb_name(sip->cmd), sip->cmd);

	}

	offset += 1;

	/* Handle error code */

	if (!BYTES_ARE_IN_FRAME(SMB_offset + 10, 2))
	  return;

	if (GSHORT(pd, SMB_offset + 10) & 0x4000) {

	    /* handle NT 32 bit error code */
	    errcode = 0;	/* better than a random number */
	    status = GWORD(pd, offset); 

	    if (tree) {

		proto_tree_add_text(smb_hdr_tree, NullTVB, offset, 4, "Status: 0x%08x",
				    status);

	    }

	    offset += 4;

	}
	else {
	    /* handle DOS error code & class */

	    /* Next, look at the error class, SMB_RETCLASS */

	    errcls = pd[offset];

	    if (tree) {

		proto_tree_add_text(smb_hdr_tree, NullTVB, offset, 1, "Error Class: %s", 
				    val_to_str((guint8)pd[offset], errcls_types, "Unknown Error Class (%x)"));
	    }

	    offset += 1;

	    /* Error code, SMB_HEINFO ... */

	    errcode1 = pd[offset];

	    if (tree) {

		proto_tree_add_text(smb_hdr_tree, NullTVB, offset, 1, "Reserved: %i", errcode1); 

	    }

	    offset += 1;

	    errcode = GSHORT(pd, offset); 

	    if (tree) {

		proto_tree_add_text(smb_hdr_tree, NullTVB, offset, 2, "Error Code: %s",
				    decode_smb_error(errcls, errcode));

	    }

	    offset += 2;
	}

	if (tree) {

	  tf = proto_tree_add_text(smb_hdr_tree, NullTVB, offset, 1, "Flags: 0x%02x", flags);

	  flags_tree = proto_item_add_subtree(tf, ett_smb_flags);
	  proto_tree_add_text(flags_tree, NullTVB, offset, 1, "%s",
			      decode_boolean_bitfield(flags, 0x01, 8,
						      "Lock&Read, Write&Unlock supported",
						      "Lock&Read, Write&Unlock not supported"));
	  proto_tree_add_text(flags_tree, NullTVB, offset, 1, "%s",
			      decode_boolean_bitfield(flags, 0x02, 8,
						      "Receive buffer posted",
						      "Receive buffer not posted"));
	  proto_tree_add_text(flags_tree, NullTVB, offset, 1, "%s",
			      decode_boolean_bitfield(flags, 0x08, 8, 
						      "Path names caseless",
						      "Path names case sensitive"));
	  proto_tree_add_text(flags_tree, NullTVB, offset, 1, "%s",
			      decode_boolean_bitfield(flags, 0x10, 8,
						      "Pathnames canonicalized",
						      "Pathnames not canonicalized"));
	  proto_tree_add_text(flags_tree, NullTVB, offset, 1, "%s",
			      decode_boolean_bitfield(flags, 0x20, 8,
						      "OpLocks requested/granted",
						      "OpLocks not requested/granted"));
	  proto_tree_add_text(flags_tree, NullTVB, offset, 1, "%s",
			      decode_boolean_bitfield(flags, 0x40, 8, 
						      "Notify all",
						      "Notify open only"));

	  proto_tree_add_text(flags_tree, NullTVB, offset, 1, "%s",
			      decode_boolean_bitfield(flags, SMB_FLAGS_DIRN,
						      8, "Response to client/redirector", "Request to server"));

	}

	offset += 1;

	if (!BYTES_ARE_IN_FRAME(offset, 2))
	  return;

	if (tree) {

	  tf = proto_tree_add_text(smb_hdr_tree, NullTVB, offset, 2, "Flags2: 0x%04x", flags2);

	  flags2_tree = proto_item_add_subtree(tf, ett_smb_flags2);
	  proto_tree_add_text(flags2_tree, NullTVB, offset, 2, "%s",
			      decode_boolean_bitfield(flags2, 0x0001, 16,
						      "Long file names supported",
						      "Long file names not supported"));
	  proto_tree_add_text(flags2_tree, NullTVB, offset, 2, "%s",
			      decode_boolean_bitfield(flags2, 0x0002, 16,
						      "Extended attributes supported",
						      "Extended attributes not supported"));
	  proto_tree_add_text(flags2_tree, NullTVB, offset, 1, "%s",
			      decode_boolean_bitfield(flags2, 0x0004, 16,
						      "Security signatures supported",
						      "Security signatures not supported"));
	  proto_tree_add_text(flags2_tree, NullTVB, offset, 2, "%s",
			      decode_boolean_bitfield(flags2, 0x0800, 16,
						      "Extended security negotiation supported",
						      "Extended security negotiation not supported"));
	  proto_tree_add_text(flags2_tree, NullTVB, offset, 2, "%s",
			      decode_boolean_bitfield(flags2, 0x1000, 16, 
						      "Resolve pathnames with DFS",
						      "Don't resolve pathnames with DFS"));
	  proto_tree_add_text(flags2_tree, NullTVB, offset, 2, "%s",
			      decode_boolean_bitfield(flags2, 0x2000, 16,
						      "Permit reads if execute-only",
						      "Don't permit reads if execute-only"));
	  proto_tree_add_text(flags2_tree, NullTVB, offset, 2, "%s",
			      decode_boolean_bitfield(flags2, 0x4000, 16,
						      "Error codes are NT error codes",
						      "Error codes are DOS error codes"));
	  proto_tree_add_text(flags2_tree, NullTVB, offset, 2, "%s",
			      decode_boolean_bitfield(flags2, 0x8000, 16, 
						      "Strings are Unicode",
						      "Strings are ASCII"));

	}

	offset += 2;

	if (tree) {

	  /*
	   * The document at
	   *
	   *	http://www.samba.org/samba/ftp/specs/smbpub.txt
	   *
	   * (a text version of "Microsoft Networks SMB FILE SHARING
	   * PROTOCOL, Document Version 6.0p") says that:
	   *
	   *	the first 2 bytes of these 12 bytes are, for NT Create and X,
	   *	the "High Part of PID";
	   *
	   *	the next four bytes are reserved;
	   *
	   *	the next four bytes are, for SMB-over-IPX (with no
	   *	NetBIOS involved) two bytes of Session ID and two bytes
	   *	of SequenceNumber.
	   *
	   * If we ever implement SMB-over-IPX (which I suspect goes over
	   * IPX sockets 0x0550, 0x0552, and maybe 0x0554, as per the
	   * document in question), we'd probably want to have some way
	   * to determine whether this is SMB-over-IPX or not (which could
	   * be done by adding a PT_IPXSOCKET port type, having the
	   * IPX dissector set "pinfo->srcport" and "pinfo->destport",
	   * and having the SMB dissector check for a port type of
	   * PT_IPXSOCKET and for "pinfo->match_port" being either
	   * IPX_SOCKET_NWLINK_SMB_SERVER or IPX_SOCKET_NWLINK_SMB_REDIR
	   * or, if it also uses 0x0554, IPX_SOCKET_NWLINK_SMB_MESSENGER).
	   */
	  proto_tree_add_text(smb_hdr_tree, NullTVB, offset, 12, "Reserved: 6 WORDS");

	}

	offset += 12;

	/* Now the TID, tree ID */

	if (!BYTES_ARE_IN_FRAME(offset, 2))
	  return;

	if (tree) {

	  proto_tree_add_text(smb_hdr_tree, NullTVB, offset, 2, "Network Path/Tree ID (TID): %i (%04x)", sip->tid, sip->tid); 

	}

	offset += 2;

	/* Now the PID, Process ID */

	if (!BYTES_ARE_IN_FRAME(offset, 2))
	  return;

	if (tree) {

	  proto_tree_add_text(smb_hdr_tree, NullTVB, offset, 2, "Process ID (PID): %i (%04x)", sip->pid, sip->pid); 

	}

	offset += 2;

	/* Now the UID, User ID */

	if (tree) {

	  proto_tree_add_text(smb_hdr_tree, NullTVB, offset, 2, "User ID (UID): %i (%04x)", sip->uid, sip->uid); 

	}
	
	offset += 2;

	/* Now the MID, Multiplex ID */

	if (tree) {

	  proto_tree_add_text(smb_hdr_tree, NullTVB, offset, 2, "Multiplex ID (MID): %i (%04x)", sip->mid, sip->mid); 

	}

	offset += 2;


	new_dissector = (pi.match_port==pi.destport)?
		smb_dissector[sip->cmd].request:smb_dissector[sip->cmd].response;

	if(new_dissector){
		tvbuff_t *next_tvb;
		proto_item *cmd_item = NULL;
		proto_tree *cmd_tree = NULL;

		/* we need some data from the smb_info structure */
		pi.private = sip;

		proto_item_set_len(ti, offset-old_offset);

		/* when dissect_smb() is tvbuffified we can start with
		   creating a subset tvb starting at SMB_offset
                   and get rid of SMB_offset completely*/
		next_tvb = tvb_create_from_top(SMB_offset);
		cmd_item = proto_tree_add_text(smb_tree, next_tvb, offset-SMB_offset,
			END_OF_FRAME, "%s %s (0x%02x)",
			decode_smb_name(sip->cmd), 
			(pi.match_port==pi.destport)?"Request":"Response",
			sip->cmd);

		cmd_tree = proto_item_add_subtree(cmd_item, ett_smb_command);
		(*new_dissector)(next_tvb, &pi, cmd_tree, offset-SMB_offset, smb_tree);
	} else {
		/* call old dissector */
/* si is a fairly large structure and it is passed as value OOOUCH! */
		(dissect[sip->cmd])(pd, offset, fd, tree, smb_tree, si, max_data,
		SMB_offset, errcode);
	}
}

/*** External routines called during the registration process */

extern void register_proto_smb_browse( void);
extern void register_proto_smb_logon( void);
extern void register_proto_smb_mailslot( void);
extern void register_proto_smb_pipe( void);
extern void register_proto_smb_mailslot( void);


void
proto_register_smb(void)
{
	static hf_register_info hf[] = {
	{ &hf_smb_cmd,
		{ "SMB Command", "smb.cmd", FT_UINT8, BASE_HEX,
		VALS(smb_cmd_vals), 0x0, "SMB Command", HFILL }},

	{ &hf_smb_word_count,
		{ "Word Count (WCT)", "smb.wct", FT_UINT8, BASE_DEC,
		NULL, 0x0, "Word Count, count of parameter words", HFILL }},

	{ &hf_smb_byte_count,
		{ "Byte Count (BCC)", "smb.bcc", FT_UINT16, BASE_DEC,
		NULL, 0x0, "Byte Count, count of data bytes", HFILL }},

	{ &hf_smb_file_type,
		{ "File Type", "smb.file_type", FT_UINT16, BASE_DEC,
		VALS(filetype_vals), 0, "Type of file", HFILL }},

	{ &hf_smb_buffer_format,
		{ "Buffer Format", "smb.buffer_format", FT_UINT8, BASE_DEC,
		VALS(buffer_format_vals), 0x0, "Buffer Format, type of buffer", HFILL }},

	{ &hf_smb_search_pattern,
		{ "Search Pattern", "smb.search_pattern", FT_STRING, BASE_NONE,
		NULL, 0, "Search Pattern", HFILL }},

	{ &hf_smb_file_name,
		{ "File Name", "smb.file", FT_STRING, BASE_NONE,
		NULL, 0, "File Name", HFILL }},

	{ &hf_smb_old_file_name,
		{ "Old File Name", "smb.file", FT_STRING, BASE_NONE,
		NULL, 0, "Old File Name (When renaming a file)", HFILL }},

	{ &hf_smb_file_attr_read_only,
		{ "", "smb.file.attribute.read_only", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_read_only), FILE_ATTRIBUTE_READ_ONLY, "READ ONLY file attribute", HFILL }},

	{ &hf_smb_file_attr_hidden,
		{ "", "smb.file.attribute.hidden", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_hidden), FILE_ATTRIBUTE_HIDDEN, "HIDDEN file attribute", HFILL }},

	{ &hf_smb_file_attr_system,
		{ "", "smb.file.attribute.system", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_system), FILE_ATTRIBUTE_SYSTEM, "SYSTEM file attribute", HFILL }},

	{ &hf_smb_file_attr_volume,
		{ "", "smb.file.attribute.volume", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_volume), FILE_ATTRIBUTE_VOLUME, "VOLUME file attribute", HFILL }},

	{ &hf_smb_file_attr_directory,
		{ "", "smb.file.attribute.directory", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_directory), FILE_ATTRIBUTE_DIRECTORY, "DIRECTORY file attribute", HFILL }},

	{ &hf_smb_file_attr_archive,
		{ "", "smb.file.attribute.archive", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_archive), FILE_ATTRIBUTE_ARCHIVE, "ARCHIVE file attribute", HFILL }},

	{ &hf_smb_file_attr_device,
		{ "", "smb.file.attribute.device", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_device), FILE_ATTRIBUTE_DEVICE, "Is this file a device?", HFILL }},

	{ &hf_smb_file_attr_normal,
		{ "", "smb.file.attribute.normal", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_normal), FILE_ATTRIBUTE_NORMAL, "Is this a normal file?", HFILL }},

	{ &hf_smb_file_attr_temporary,
		{ "", "smb.file.attribute.temporary", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_temporary), FILE_ATTRIBUTE_TEMPORARY, "Is this a temporary file?", HFILL }},

	{ &hf_smb_file_attr_sparse,
		{ "", "smb.file.attribute.sparse", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_sparse), FILE_ATTRIBUTE_SPARSE, "Is this a sparse file?", HFILL }},

	{ &hf_smb_file_attr_reparse,
		{ "", "smb.file.attribute.reparse", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_reparse), FILE_ATTRIBUTE_REPARSE, "Does this file have an associated reparse point?", HFILL }},

	{ &hf_smb_file_attr_compressed,
		{ "", "smb.file.attribute.compressed", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_compressed), FILE_ATTRIBUTE_COMPRESSED, "Is this file compressed?", HFILL }},

	{ &hf_smb_file_attr_offline,
		{ "", "smb.file.attribute.offline", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_offline), FILE_ATTRIBUTE_OFFLINE, "Is this file offline?", HFILL }},

	{ &hf_smb_file_attr_not_content_indexed,
		{ "", "smb.file.attribute.not_content_indexed", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_not_content_indexed), FILE_ATTRIBUTE_NOT_CONTENT_INDEXED, "May this file be indexed by the content indexing service", HFILL }},

	{ &hf_smb_file_attr_encrypted,
		{ "", "smb.file.attribute.encrypted", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_encrypted), FILE_ATTRIBUTE_ENCRYPTED, "Is this file encrypted?", HFILL }},

	{ &hf_smb_search_attribute_read_only,
		{ "Read Only", "smb.search.attribute.read_only", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_read_only), FILE_ATTRIBUTE_READ_ONLY, "READ ONLY search attribute", HFILL }},

	{ &hf_smb_search_attribute_hidden,
		{ "Hidden", "smb.search.attribute.hidden", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_hidden), FILE_ATTRIBUTE_HIDDEN, "HIDDEN search attribute", HFILL }},

	{ &hf_smb_search_attribute_system,
		{ "System", "smb.search.attribute.system", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_system), FILE_ATTRIBUTE_SYSTEM, "SYSTEM search attribute", HFILL }},

	{ &hf_smb_search_attribute_volume,
		{ "Volume", "smb.search.attribute.volume", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_volume), FILE_ATTRIBUTE_VOLUME, "VOLUME search attribute", HFILL }},

	{ &hf_smb_search_attribute_directory,
		{ "Directory", "smb.search.attribute.directory", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_directory), FILE_ATTRIBUTE_DIRECTORY, "DIRECTORY search attribute", HFILL }},

	{ &hf_smb_search_attribute_archive,
		{ "Archive", "smb.search.attribute.archive", FT_BOOLEAN, 16,
		TFS(&tfs_file_attribute_archive), FILE_ATTRIBUTE_ARCHIVE, "ARCHIVE search attribute", HFILL }},

	{ &hf_smb_last_write_time,
		{ "Last Write Time", "smb.last_write.time", FT_ABSOLUTE_TIME, BASE_HEX,
		NULL, 0, "Last time this file was written to", HFILL }},

	{ &hf_smb_file_size,
		{ "File Size", "smb.file.size", FT_UINT32, BASE_DEC,
		NULL, 0, "File Size", HFILL }},

	{ &hf_smb_unknown,
		{ "Unknown Data", "smb.unknown", FT_BYTES, BASE_HEX,
		NULL, 0, "Unknown Data. Should be implemented by someone", HFILL }},

	{ &hf_smb_reserved,
		{ "Reserved", "smb.reserved", FT_BYTES, BASE_HEX,
		NULL, 0, "Reserved bytes, must be zero", HFILL }},

	{ &hf_smb_padding,
		{ "Padding", "smb.padding", FT_BYTES, BASE_HEX,
		NULL, 0, "Padding or unknown data", HFILL }},

	{ &hf_smb_dialect_name,
		{ "Name", "smb.dialect.name", FT_STRING, BASE_NONE,
		NULL, 0, "Name of dialect", HFILL }},

	{ &hf_smb_dialect_index,
		{ "Selected Index", "smb.dialect.index", FT_UINT16, BASE_DEC,
		NULL, 0, "Index of selected dialect", HFILL }},

	{ &hf_sm_mode16,
		{ "Mode", "smb.sm.mode", FT_BOOLEAN, 16,
		TFS(&tfs_sm_mode), SECURITY_MODE_MODE, "User or Share security mode?", HFILL }},

	{ &hf_sm_password16,
		{ "Password", "smb.sm.password", FT_BOOLEAN, 16,
		TFS(&tfs_sm_password), SECURITY_MODE_PASSWORD, "Encrypted or plaintext passwords?", HFILL }},

	{ &hf_sm_mode,
		{ "Mode", "smb.sm.mode", FT_BOOLEAN, 8,
		TFS(&tfs_sm_mode), SECURITY_MODE_MODE, "User or Share security mode?", HFILL }},

	{ &hf_sm_password,
		{ "Password", "smb.sm.password", FT_BOOLEAN, 8,
		TFS(&tfs_sm_password), SECURITY_MODE_PASSWORD, "Encrypted or plaintext passwords?", HFILL }},

	{ &hf_sm_signatures,
		{ "Signatures", "smb.sm.signatures", FT_BOOLEAN, 8,
		TFS(&tfs_sm_signatures), SECURITY_MODE_SIGNATURES, "Are security signatures enabled?", HFILL }},

	{ &hf_sm_sig_required,
		{ "Sig Req", "smb.sm.sig_required", FT_BOOLEAN, 8,
		TFS(&tfs_sm_sig_required), SECURITY_MODE_SIG_REQUIRED, "Are security signatures required?", HFILL }},

	{ &hf_max_buf_size,
		{ "Max Buffer", "smb.max_buf", FT_UINT16, BASE_DEC,
		NULL, 0, "Max client buffer size", HFILL }},

	{ &hf_max_trans_buf_size,
		{ "Max TBuffer", "smb.max_trans", FT_UINT32, BASE_DEC,
		NULL, 0, "Max transmit buffer size", HFILL }},

	{ &hf_max_mpx_count,
		{ "Max Mpx", "smb.max_mpx", FT_UINT16, BASE_DEC,
		NULL, 0, "Max pending multiplexed requests", HFILL }},

	{ &hf_vc_num,
		{ "VC Number", "smb.vc", FT_UINT16, BASE_DEC,
		NULL, 0, "VC Number", HFILL }},

	{ &hf_max_vcs_num,
		{ "Max VCs", "smb.max_vcs", FT_UINT16, BASE_DEC,
		NULL, 0, "Max VCs between client and server", HFILL }},

	{ &hf_rm_read,
		{ "Read", "smb.rm.read", FT_BOOLEAN, 16,
		TFS(&tfs_rm_read), RAWMODE_READ, "Is rawmode READ supported?", HFILL }},

	{ &hf_rm_write,
		{ "Write", "smb.rm.write", FT_BOOLEAN, 16,
		TFS(&tfs_rm_write), RAWMODE_WRITE, "Is rawmode WRITE supported?", HFILL }},

	{ &hf_session_key,
		{ "Session Key", "smb.session_key", FT_UINT32, BASE_HEX,
		NULL, 0, "Unique token identifying this session", HFILL }},

	{ &hf_current_server_date,
		{ "Current Date", "smb.server.current.date", FT_ABSOLUTE_TIME, BASE_NONE,
		NULL, 0, "Current date and time at server", HFILL }},

	{ &hf_current_server_smb_date,
		{ "Current Date", "smb.server.current.smb.date", FT_UINT16, BASE_HEX,
		NULL, 0, "Current date at server, SMB_DATE format", HFILL }},

	{ &hf_current_server_smb_time,
		{ "Current Time", "smb.server.current.smb.time", FT_UINT16, BASE_HEX,
		NULL, 0, "Current time at server, SMB_TIME format", HFILL }},

	{ &hf_server_timezone,
		{ "Time Zone", "smb.server.timezone", FT_INT16, BASE_DEC,
		NULL, 0, "Current timezone at server.", HFILL }},

	{ &hf_encryption_key_length,
		{ "Key Length", "smb.encryption.key.length", FT_UINT16, BASE_DEC,
		NULL, 0, "Encryption key length (must be 0 if not LM2.1 dialect)", HFILL }},

	{ &hf_encryption_key,
		{ "Encryption Key", "smb.encryption.key", FT_BYTES, BASE_HEX,
		NULL, 0, "Challange/Response Encryption Key (for LM2.1 dialect)", HFILL }},

	{ &hf_lanman,
		{ "LanMan", "smb.lanman", FT_STRING, BASE_NONE,
		NULL, 0, "Which LANMAN we are running", HFILL }},

	{ &hf_os,
		{ "OS", "smb.os", FT_STRING, BASE_NONE,
		NULL, 0, "Which OS we are running", HFILL }},

	{ &hf_domain,
		{ "Domain", "smb.domain", FT_STRING, BASE_NONE,
		NULL, 0, "The Server's Primary Domain", HFILL }},

	{ &hf_max_raw_buf_size,
		{ "Max Raw Buffer", "smb.max_raw", FT_UINT32, BASE_DEC,
		NULL, 0, "Max raw buffer size", HFILL }},

	{ &hf_server_cap_raw_mode,
		{ "Raw Mode", "smb.server.cap.raw_mode", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_raw_mode), SERVER_CAP_RAW_MODE, "Is RAW MODE capability supported?", HFILL }},

	{ &hf_server_cap_mpx_mode,
		{ "MPX Mode", "smb.server.cap.mpx_mode", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_mpx_mode), SERVER_CAP_MPX_MODE, "Is MPX MODE capability supported?", HFILL }},

	{ &hf_server_cap_unicode,
		{ "Unicode", "smb.server.cap.unicode", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_unicode), SERVER_CAP_UNICODE, "Is UNICODE capability supported?", HFILL }},

	{ &hf_server_cap_large_files,
		{ "Large Files", "smb.server.cap.large_files", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_large_files), SERVER_CAP_LARGE_FILES, "Is LARGE FILES capability supported?", HFILL }},

	{ &hf_server_cap_nt_smbs,
		{ "NT Smbs", "smb.server.cap.nt_smbs", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_nt_smbs), SERVER_CAP_NT_SMBS, "Is NT SMBS capability supported?", HFILL }},

	{ &hf_server_cap_rpc_remote_apis,
		{ "RPC Remote APIs", "smb.server.cap.rpc_remote_apis", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_rpc_remote_apis), SERVER_CAP_RPC_REMOTE_APIS, "Is RPC REMOTE APIS capability supported?", HFILL }},

	{ &hf_server_cap_status32,
		{ "Status32", "smb.server.cap.status32", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_status32), SERVER_CAP_STATUS32, "Is STATUS32 capability supported?", HFILL }},

	{ &hf_server_cap_level_ii_oplocks,
		{ "Level 2 Oplocks", "smb.server.cap.level_2_oplocks", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_level_ii_oplocks), SERVER_CAP_LEVEL_II_OPLOCKS, "Is LEVEL II OPLOCKS capability supported?", HFILL }},

	{ &hf_server_cap_lock_and_read,
		{ "Lock and Read", "smb.server.cap.lock_and_read", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_lock_and_read), SERVER_CAP_LOCK_AND_READ, "Is LOCK AND READ capability supported?", HFILL }},

	{ &hf_server_cap_nt_find,
		{ "NT Find", "smb.server.cap.nt_find", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_nt_find), SERVER_CAP_NT_FIND, "Is NT FIND capability supported?", HFILL }},

	{ &hf_server_cap_dfs,
		{ "DFS", "smb.server.cap.dfs", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_dfs), SERVER_CAP_DFS, "Is DFS capability supported?", HFILL }},

	{ &hf_server_cap_infolevel_passthru,
		{ "Infolevel Passthru", "smb.server.cap.infolevel_passthru", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_infolevel_passthru), SERVER_CAP_INFOLEVEL_PASSTHRU, "Is INFOLEVEL PASSTHRU capability supported?", HFILL }},

	{ &hf_server_cap_large_readx,
		{ "Large ReadX", "smb.server.cap.large_readx", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_large_readx), SERVER_CAP_LARGE_READX, "Is LARGE READX capability supported?", HFILL }},

	{ &hf_server_cap_large_writex,
		{ "Large WriteX", "smb.server.cap.large_writex", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_large_writex), SERVER_CAP_LARGE_WRITEX, "Is LARGE WRITEX capability supported?", HFILL }},

	{ &hf_server_cap_unix,
		{ "Unix", "smb.server.cap.unix", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_unix), SERVER_CAP_UNIX , "Is UNIX capability supported?", HFILL }},

	{ &hf_server_cap_reserved,
		{ "Reserved", "smb.server.cap.reserved", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_reserved), SERVER_CAP_RESERVED, "RESERVED", HFILL }},

	{ &hf_server_cap_bulk_transfer,
		{ "Bulk Transfer", "smb.server.cap.bulk_transfer", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_bulk_transfer), SERVER_CAP_BULK_TRANSFER, "Is BULK TRANSFER capability supported?", HFILL }},

	{ &hf_server_cap_compressed_data,
		{ "Compressed Data", "smb.server.cap.compressed_data", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_compressed_data), SERVER_CAP_COMPRESSED_DATA, "Is COMPRESSED DATA capability supported?", HFILL }},

	{ &hf_server_cap_extended_security,
		{ "Extended Security", "smb.server.cap.extended_security", FT_BOOLEAN, 32,
		TFS(&tfs_server_cap_extended_security), SERVER_CAP_EXTENDED_SECURITY, "Is EXTENDED SECURITY capability supported?", HFILL }},

	{ &hf_security_blob,
		{ "Security Blob", "smb.security.blob", FT_BYTES, BASE_HEX,
		NULL, 0, "Security blob", HFILL }},

	{ &hf_server_guid,
		{ "Server GUID", "smb.server.guid", FT_BYTES, BASE_HEX,
		NULL, 0, "Globally unique identifier for this server", HFILL }},

	{ &hf_time_low,
		{ "Time Low", "smb.time.low", FT_UINT32, BASE_HEX,
		NULL, 0, "Low part of TIME structure", HFILL }},

	{ &hf_time_high,
		{ "Time High", "smb.time.high", FT_UINT32, BASE_HEX,
		NULL, 0, "High part of TIME structure", HFILL }},

	{ &hf_smb_dir_name,
		{ "Directory", "smb.dir_name", FT_STRING, BASE_NONE,
		NULL, 0, "SMB Directory Name", HFILL }},

	{ &hf_smb_sid,
		{ "SID", "smb.sid", FT_UINT16, BASE_HEX,
		NULL, 0, "SID: Search ID, handle for find operations", HFILL }},

	{ &hf_smb_tid,
		{ "TID", "smb.tid", FT_UINT16, BASE_HEX,
		NULL, 0, "TID: TreeID", HFILL }},

	{ &hf_smb_fid,
		{ "FID", "smb.fid", FT_UINT16, BASE_HEX,
		NULL, 0, "FID: FileID", HFILL }},

	{ &hf_smb_count,
		{ "Count", "smb.count", FT_UINT32, BASE_DEC,
		NULL, 0, "Count, number of bytes/files", HFILL }},

	{ &hf_smb_maxcount,
		{ "MaxCount", "smb.maxcount", FT_UINT16, BASE_DEC,
		NULL, 0, "Maximum Count", HFILL }},

	{ &hf_smb_mincount,
		{ "MinCount", "smb.mincount", FT_UINT16, BASE_DEC,
		NULL, 0, "Minimum Count", HFILL }},

	{ &hf_smb_storage_type,
		{ "Storage Type", "smb.storage_type", FT_UINT32, BASE_DEC,
		NULL, 0, "Type of storage", HFILL }},

	{ &hf_smb_resume,
		{ "Resume Key", "smb.resume", FT_UINT32, BASE_DEC,
		NULL, 0, "Resume Key", HFILL }},

	{ &hf_smb_file_name_len,
		{ "File Name Len", "smb.file_name_len", FT_UINT32, BASE_DEC,
		NULL, 0, "Length of File Name", HFILL }},

	{ &hf_smb_offset,
		{ "Offset", "smb.offset", FT_UINT32, BASE_DEC,
		NULL, 0, "Offset in file", HFILL }},

	{ &hf_smb_high_offset,
		{ "High Offset", "smb.offset_high", FT_UINT32, BASE_DEC,
		NULL, 0, "High 32 Bits Of File Offset", HFILL }},

	{ &hf_smb_remaining,
		{ "Remaining", "smb.remaining", FT_UINT32, BASE_DEC,
		NULL, 0, "Remaining number of bytes", HFILL }},

	{ &hf_smb_change_time,
		{ "Change", "smb.change.time", FT_ABSOLUTE_TIME, BASE_NONE,
		NULL, 0, "Last Change Time", HFILL }},

	{ &hf_smb_access_time,
		{ "Last Access", "smb.access.time", FT_ABSOLUTE_TIME, BASE_NONE,
		NULL, 0, "Last Access Time", HFILL }},

	{ &hf_smb_access_dos_date,
		{ "Last Access Date", "smb.access.smb.date", FT_UINT16, BASE_HEX,
		NULL, 0, "Last Access Date, SMB_DATE format", HFILL }},

	{ &hf_smb_access_dos_time,
		{ "Last Access Time", "smb.access.smb.time", FT_UINT16, BASE_HEX,
		NULL, 0, "Last Access Time, SMB_TIME format", HFILL }},

	{ &hf_smb_create_time,
		{ "Created", "smb.create.time", FT_ABSOLUTE_TIME, BASE_NONE,
		NULL, 0, "Creation Time", HFILL }},

	{ &hf_smb_last_write_date,
		{ "Last Write", "smb.last_write.date", FT_ABSOLUTE_TIME, BASE_NONE,
		NULL, 0, "Last Write", HFILL }},

	{ &hf_smb_create_dos_date,
		{ "Create Date", "smb.create.smb.date", FT_UINT16, BASE_HEX,
		NULL, 0, "Create Date, SMB_DATE format", HFILL }},

	{ &hf_smb_create_dos_time,
		{ "Create Time", "smb.create.smb.time", FT_UINT16, BASE_HEX,
		NULL, 0, "Create Time, SMB_TIME format", HFILL }},

	{ &hf_smb_last_write_dos_date,
		{ "Last Write Date", "smb.last_write.smb.date", FT_UINT16, BASE_HEX,
		NULL, 0, "Last Write Date, SMB_DATE format", HFILL }},

	{ &hf_smb_last_write_dos_time,
		{ "Last Write Time", "smb.last_write.smb.time", FT_UINT16, BASE_HEX,
		NULL, 0, "Last Write Time, SMB_TIME format", HFILL }},

	{ &hf_smb_file_data,
		{ "File Data", "smb.file.data", FT_BYTES, BASE_HEX,
		NULL, 0, "Data read/written to the file", HFILL }},

	{ &hf_smb_data_len,
		{ "Data Len", "smb.data_len", FT_UINT16, BASE_DEC,
		NULL, 0, "Length of data", HFILL }},

	{ &hf_smb_data_disp,
		{ "Data Disp", "smb.data_disp", FT_UINT16, BASE_DEC,
		NULL, 0, "Data Displacement", HFILL }},

	{ &hf_smb_data_offset,
		{ "Data Offset", "smb.data_offset", FT_UINT16, BASE_DEC,
		NULL, 0, "Data Offset", HFILL }},

	{ &hf_smb_seek_mode,
		{ "Seek Mode", "smb.seek_mode", FT_UINT16, BASE_DEC,
		VALS(seek_mode_vals), 0, "Seek Mode, what type of seek", HFILL }},

	{ &hf_smb_timeout,
		{ "Timeout", "smb.timeout", FT_UINT32, BASE_DEC,
		NULL, 0, "Timeout in miliseconds", HFILL }},

	{ &hf_smb_dcm,
		{ "Data Compaction Mode", "smb.dcm", FT_UINT16, BASE_DEC,
		NULL, 0, "Data Compaction Mode", HFILL }},

	{ &hf_smb_units,
		{ "Total Units", "smb.units", FT_UINT16, BASE_DEC,
		NULL, 0, "Total number of units at server", HFILL }},

	{ &hf_smb_bpu,
		{ "Blocks Per Unit", "smb.bpu", FT_UINT16, BASE_DEC,
		NULL, 0, "Blocks per unit at server", HFILL }},

	{ &hf_smb_blocksize,
		{ "Blocksize", "smb.blocksize", FT_UINT16, BASE_DEC,
		NULL, 0, "Blocksize (in bytes) at server", HFILL }},

	{ &hf_smb_freeunits,
		{ "Free Units", "smb.free_units", FT_UINT16, BASE_DEC,
		NULL, 0, "Number of free units at server", HFILL }},

	{ &hf_smb_andxoffset,
		{ "AndXOffset", "smb.andxoffset", FT_UINT16, BASE_DEC,
		NULL, 0, "Offset to next command in this SMB packet", HFILL }},

	{ &hf_smb_password_len,
		{ "Password Len", "smb.pwlen", FT_UINT16, BASE_DEC,
		NULL, 0, "Length of password", HFILL }},

	{ &hf_smb_password,
		{ "Password", "smb.password", FT_BYTES, BASE_NONE,
		NULL, 0, "Password", HFILL }},

	{ &hf_smb_account,
		{ "Account", "smb.account", FT_STRING, BASE_NONE,
		NULL, 0, "Account, username", HFILL }},

	{ &hf_smb_connect_flags_dtid,
		{ "", "smb.connect.flags.dtid", FT_BOOLEAN, 16,
		TFS(&tfs_disconnect_tid), 0x0001, "Disconnect TID?", HFILL }},

	{ &hf_smb_path,
		{ "Path", "smb.path", FT_STRING, BASE_NONE,
		NULL, 0, "Path. Server name and share name", HFILL }},

	{ &hf_smb_trans_name,
		{ "Transaction Name", "smb.trans_name", FT_STRING, BASE_NONE,
		NULL, 0, "Name of transaction", HFILL }},

	{ &hf_smb_service,
		{ "Service", "smb.service", FT_STRING, BASE_NONE,
		NULL, 0, "Service name", HFILL }},

	{ &hf_smb_setup_action_guest,
		{ "", "smb.setup.action.guest", FT_BOOLEAN, 16,
		TFS(&tfs_setup_action_guest), 0x0001, "Client logged in as GUEST?", HFILL }},

	{ &hf_smb_connect_support_search,
		{ "", "smb.connect.support.search", FT_BOOLEAN, 16,
		TFS(&tfs_connect_support_search), 0x0001, "Search Bits supported?", HFILL }},

	{ &hf_smb_connect_support_in_dfs,
		{ "", "smb.connect.support.dfs", FT_BOOLEAN, 16,
		TFS(&tfs_connect_support_in_dfs), 0x0002, "Is this in a DFS tree?", HFILL }},

	{ &hf_smb_fs,
		{ "Native FS", "smb.fs", FT_STRING, BASE_NONE,
		NULL, 0, "Native FileSystem", HFILL }},

	{ &hf_smb_open_flags_add_info,
		{ "", "smb.open.flags.add_info", FT_BOOLEAN, 16,
		TFS(&tfs_open_flags_add_info), 0x0001, "Additional Information Requested?", HFILL }},

	{ &hf_smb_open_flags_ex_oplock,
		{ "", "smb.open.flags.ex_oplock", FT_BOOLEAN, 16,
		TFS(&tfs_open_flags_ex_oplock), 0x0002, "Exclusive Oplock Requested?", HFILL }},

	{ &hf_smb_open_flags_batch_oplock,
		{ "", "smb.open.flags.batch_oplock", FT_BOOLEAN, 16,
		TFS(&tfs_open_flags_batch_oplock), 0x0004, "Batch Oplock Requested?", HFILL }},

	{ &hf_smb_open_flags_ealen,
		{ "", "smb.open.flags.ealen", FT_BOOLEAN, 16,
		TFS(&tfs_open_flags_ealen), 0x0008, "Total EA Len Requested?", HFILL }},

	{ &hf_smb_access_mode,
		{ "", "smb.access.mode", FT_UINT16, BASE_HEX,
		VALS(da_access_vals), 0x0007, "Access Mode", HFILL }},

	{ &hf_smb_access_sharing,
		{ "", "smb.access.sharing", FT_UINT16, BASE_HEX,
		VALS(da_sharing_vals), 0x0070, "Sharing Mode", HFILL }},

	{ &hf_smb_access_locality,
		{ "", "smb.access.locality", FT_UINT16, BASE_HEX,
		VALS(da_locality_vals), 0x0700, "Locality of reference", HFILL }},

	{ &hf_smb_access_caching,
		{ "", "smb.access.caching", FT_BOOLEAN, 16,
		TFS(&tfs_da_caching), 0x1000, "Caching mode?", HFILL }},

	{ &hf_smb_access_writetru,
		{ "", "smb.access.writethrough", FT_BOOLEAN, 16,
		TFS(&tfs_da_writetru), 0x4000, "Writethrough mode?", HFILL }},

	{ &hf_smb_open_function_create,
		{ "", "smb.open.function.create", FT_BOOLEAN, 16,
		TFS(&tfs_of_create), 0x0010, "Create file if it doesnt exist?", HFILL }},

	{ &hf_smb_open_function_open,
		{ "", "smb.open.function.open", FT_UINT16, BASE_HEX,
		VALS(of_open), 0x0003, "Open Function, how to open the file", HFILL }},

	{ &hf_smb_last_name_offset,
		{ "Last Name Offset", "smb.last_name_offset", FT_UINT16, BASE_DEC,
		NULL, 0, "If non-0 this is the offset into the datablock for the file name of the last entry", HFILL }},

	{ &hf_smb_end_of_search,
		{ "End Of Search", "smb.end_of_search", FT_UINT16, BASE_DEC,
		NULL, 0, "Was last entry returned?", HFILL }},

	{ &hf_smb_end_of_file,
		{ "End Of File", "smb.end_of_file", FT_UINT32, BASE_DEC,
		NULL, 0, "Offset to the first free byte in the file", HFILL }},

	{ &hf_smb_alloc_size,
		{ "Allocation Size", "smb.alloc_size", FT_UINT32, BASE_DEC,
		NULL, 0, "Number of bytes to reserve on create or truncate", HFILL }},

	{ &hf_smb_open_action_open,
		{ "", "smb.open.action.open", FT_UINT16, BASE_HEX,
		VALS(oa_open_vals), 0x0003, "Open Action, how the file was opened", HFILL }},

	{ &hf_smb_open_action_lock,
		{ "", "smb.open.action.lock", FT_BOOLEAN, 16,
		TFS(&tfs_oa_lock), 0x8000, "Can this file be locked (now)?", HFILL }},

	{ &hf_smb_device_state,
		{ "Device State", "smb.device_state", FT_UINT16, BASE_HEX,
		NULL, 0, "Device State", HFILL }},

	{ &hf_smb_server_fid,
		{ "Server FID", "smb.server_fid", FT_UINT32, BASE_HEX,
		NULL, 0, "Server unique File ID", HFILL }},

	{ &hf_smb_move_flags_file,
		{ "", "smb.move.flags.file", FT_BOOLEAN, 16,
		TFS(&tfs_mf_file), 0x0001, "Must target be a file?", HFILL }},

	{ &hf_smb_move_flags_dir,
		{ "", "smb.move.flags.dir", FT_BOOLEAN, 16,
		TFS(&tfs_mf_dir), 0x0002, "Must target be a directory?", HFILL }},

	{ &hf_smb_move_flags_verify,
		{ "", "smb.move.flags.verify", FT_BOOLEAN, 16,
		TFS(&tfs_mf_verify), 0x0010, "Verify all writes?", HFILL }},

	{ &hf_smb_total_param_count,
		{ "Total Param Count", "smb.tpc", FT_UINT16, BASE_DEC,
		NULL, 0, "Total number of parameter bytes", HFILL }},

	{ &hf_smb_total_data_count,
		{ "Total Data Count", "smb.tdc", FT_UINT16, BASE_DEC,
		NULL, 0, "Total number of data bytes", HFILL }},

	{ &hf_smb_max_param_count,
		{ "Max Param Count", "smb.mpc", FT_UINT16, BASE_DEC,
		NULL, 0, "Maximum number of parameter bytes to return", HFILL }},

	{ &hf_smb_max_data_count,
		{ "Max Data Count", "smb.mdc", FT_UINT16, BASE_DEC,
		NULL, 0, "Maximum number of data bytes to return", HFILL }},

	{ &hf_smb_max_setup_count,
		{ "Max Setup Count", "smb.msc", FT_UINT8, BASE_DEC,
		NULL, 0, "Maximum number of setup words to return", HFILL }},

	{ &hf_smb_transaction_flags_dtid,
		{ "", "smb.transaction.flags.dtid", FT_BOOLEAN, 16,
		TFS(&tfs_tf_dtid), 0x0001, "Disconnect TID?", HFILL }},

	{ &hf_smb_transaction_flags_owt,
		{ "", "smb.transaction.flags.owt", FT_BOOLEAN, 16,
		TFS(&tfs_tf_owt), 0x0002, "One Way Transaction (no response)?", HFILL }},

	{ &hf_smb_param_disp,
		{ "Param Disp", "smb.pd", FT_UINT16, BASE_DEC,
		NULL, 0, "Displacement of these parameter bytes", HFILL }},

	{ &hf_smb_param_count,
		{ "Param Count", "smb.pc", FT_UINT16, BASE_DEC,
		NULL, 0, "Number of parameter bytes in this buffer", HFILL }},

	{ &hf_smb_param_offset,
		{ "Param Offset", "smb.po", FT_UINT16, BASE_DEC,
		NULL, 0, "Offset (from header start) to parameters", HFILL }},

	{ &hf_smb_data_count,
		{ "Data Count", "smb.dc", FT_UINT16, BASE_DEC,
		NULL, 0, "Number of data bytes in this buffer", HFILL }},

	{ &hf_smb_search_count,
		{ "Search Count", "smb.search_count", FT_UINT16, BASE_DEC,
		NULL, 0, "Maximum number of search entries to return", HFILL }},

	{ &hf_smb_setup_count,
		{ "Setup Count", "smb.sc", FT_UINT8, BASE_DEC,
		NULL, 0, "Number of setup words in this buffer", HFILL }},

	{ &hf_smb_trans2_subcmd,
		{ "Subcommand", "smb.trans2.cmd", FT_UINT16, BASE_DEC,
		VALS(trans2_cmd_vals), 0, "Subcommand for TRANSACTION2", HFILL }},

	{ &hf_smb_qpi_loi,
		{ "Level of Interest", "smb.loi", FT_UINT16, BASE_DEC,
		VALS(qpi_loi_vals), 0, "Level of interest for TRANSACTION[2] commands", HFILL }},

	{ &hf_smb_ff2_backup,
		{ "", "smb.find_first2.flags.backup", FT_BOOLEAN, 16,
		TFS(&tfs_ff2_backup), 0x0010, "Find with backup intent", HFILL }},

	{ &hf_smb_ff2_continue,
		{ "", "smb.find_first2.flags.continue", FT_BOOLEAN, 16,
		TFS(&tfs_ff2_continue), 0x0008, "Continue search from previous ending place", HFILL }},

	{ &hf_smb_ff2_resume,
		{ "", "smb.find_first2.flags.resume", FT_BOOLEAN, 16,
		TFS(&tfs_ff2_resume), 0x0004, "Return resume keys for each entry found", HFILL }},

	{ &hf_smb_ff2_close_eos,
		{ "", "smb.find_first2.flags.eos", FT_BOOLEAN, 16,
		TFS(&tfs_ff2_close_eos), 0x0002, "Close search if end of search reached", HFILL }},

	{ &hf_smb_ff2_close,
		{ "", "smb.find_first2.flags.close", FT_BOOLEAN, 16,
		TFS(&tfs_ff2_close), 0x0001, "Close search after this request", HFILL }},

	{ &hf_smb_ff2_information_level,
		{ "Level of Interest", "smb.ff2_loi", FT_UINT16, BASE_DEC,
		VALS(ff2_il_vals), 0, "Level of interest for FIND_FIRST2 command", HFILL }},

	{ &hf_smb_qfsi_information_level,
		{ "Level of Interest", "smb.qfi_loi", FT_UINT16, BASE_DEC,
		VALS(qfsi_vals), 0, "Level of interest for QUERY_FS_INFORMATION2 command", HFILL }},

	{ &hf_smb_sfi_information_level,
		{ "Level of Interest", "smb.sfi_loi", FT_UINT16, BASE_DEC,
		VALS(sfi_il_vals), 0, "Level of interest for SET_FILE_INFORMATION2 command", HFILL }},

	{ &hf_smb_max_referral_level,
		{ "Max Referral Level", "smb.max_referral_level", FT_UINT16, BASE_DEC,
		NULL, 0, "Latest referral version number understood", HFILL }},

	{ &hf_smb_data_size,
		{ "Data Size", "smb.data_size", FT_UINT32, BASE_DEC,
		NULL, 0, "Data Size", HFILL }},

	{ &hf_smb_ea_size,
		{ "EA Size", "smb.ea_size", FT_UINT32, BASE_DEC,
		NULL, 0, "Size of file's EA information", HFILL }},

	{ &hf_smb_list_length,
		{ "ListLength", "smb.list_len", FT_UINT32, BASE_DEC,
		NULL, 0, "Length of the remaining data", HFILL }},

	{ &hf_smb_number_of_links,
		{ "Link Count", "smb.link_count", FT_UINT32, BASE_DEC,
		NULL, 0, "Number of hard links to the file", HFILL }},

	{ &hf_smb_delete_pending,
		{ "Delete Pending", "smb.delete_pending", FT_UINT16, BASE_DEC,
		VALS(delete_pending_vals), 0, "Is this object about to be deleted?", HFILL }},

	{ &hf_smb_is_directory,
		{ "Is Directory", "smb.is_directory", FT_UINT16, BASE_DEC,
		VALS(is_directory_vals), 0, "Is this object a directory?", HFILL }},

	{ &hf_smb_current_offset,
		{ "Current Offset", "smb.current_offset", FT_UINT32, BASE_DEC,
		NULL, 0, "Current offset in the file", HFILL }},

	{ &hf_smb_index_number,
		{ "Index Number", "smb.index_number", FT_UINT32, BASE_DEC,
		NULL, 0, "File system unique identifier", HFILL }},


	{ &hf_smb_t2_access_read_data,
		{ "", "smb.t2.access.read", FT_BOOLEAN, 32,
		TFS(&tfs_t2_access_read_data), 0x00000001, "This file can be read from", HFILL }},

	{ &hf_smb_t2_access_write_data,
		{ "", "smb.t2.access.write", FT_BOOLEAN, 32,
		TFS(&tfs_t2_access_write_data), 0x00000002, "This file can be written to", HFILL }},

	{ &hf_smb_t2_access_append_data,
		{ "", "smb.t2.access.append", FT_BOOLEAN, 32,
		TFS(&tfs_t2_access_append_data), 0x00000004, "Data can be appended to this file", HFILL }},

	{ &hf_smb_t2_access_read_ea,
		{ "", "smb.t2.access.read_ea", FT_BOOLEAN, 32,
		TFS(&tfs_t2_access_read_ea), 0x00000008, "Extended Attributes for this filehandle can be read", HFILL }},

	{ &hf_smb_t2_access_write_ea,
		{ "", "smb.t2.access.write_ea", FT_BOOLEAN, 32,
		TFS(&tfs_t2_access_write_ea), 0x00000010, "Extended Attributes for this filehandle can be written", HFILL }},

	{ &hf_smb_t2_access_execute,
		{ "", "smb.t2.access.execute", FT_BOOLEAN, 32,
		TFS(&tfs_t2_access_execute), 0x00000020, "This filehandle supports demand paging and can be executed", HFILL }},

	{ &hf_smb_t2_access_read_attributes,
		{ "", "smb.t2.access.read_attr", FT_BOOLEAN, 32,
		TFS(&tfs_t2_access_read_attributes), 0x00000080, "Attributes for this filehandle can be read", HFILL }},

	{ &hf_smb_t2_access_write_attributes,
		{ "", "smb.t2.access.write_attr", FT_BOOLEAN, 32,
		TFS(&tfs_t2_access_write_attributes), 0x00000100, "Attributes for this filehandle can be written", HFILL }},

	{ &hf_smb_t2_access_delete,
		{ "", "smb.t2.access.delete", FT_BOOLEAN, 32,
		TFS(&tfs_t2_access_delete), 0x00010000, "This file can be deleted", HFILL }},

	{ &hf_smb_t2_access_read_control,
		{ "", "smb.t2.access.read_control", FT_BOOLEAN, 32,
		TFS(&tfs_t2_access_read_control), 0x00020000, "ACL and ownership can be read from this filehandle", HFILL }},

	{ &hf_smb_t2_access_write_dac,
		{ "", "smb.t2.access.write_dac", FT_BOOLEAN, 32,
		TFS(&tfs_t2_access_write_dac), 0x00040000, "ACL and owner can be written(updated) for this filehandle", HFILL }},

	{ &hf_smb_t2_access_write_owner,
		{ "", "smb.t2.access.write_owner", FT_BOOLEAN, 32,
		TFS(&tfs_t2_access_write_owner), 0x00080000, "Ownership information can be written(changed)", HFILL }},

	{ &hf_smb_t2_access_synchronize,
		{ "", "smb.t2.access.sync", FT_BOOLEAN, 32,
		TFS(&tfs_t2_access_synchronize), 0x00100000, "The file handle supports synchronous updates", HFILL }},

	{ &hf_smb_t2_mode_write_through,
		{ "", "smb.t2.mode.write_through", FT_BOOLEAN, 32,
		TFS(&tfs_t2_mode_write_through), 0x00000002, "Whether file is to be opened in write-through mode or not?", HFILL}},

	{ &hf_smb_t2_mode_sequential_only,
		{ "", "smb.t2.mode.seq_only", FT_BOOLEAN, 32,
		TFS(&tfs_t2_mode_sequential_only), 0x00000004, "Is file going to be sequentially accessed?", HFILL}},

	{ &hf_smb_t2_mode_sync_io_alert,
		{ "", "smb.t2.mode.sync_io_alert", FT_BOOLEAN, 32,
		TFS(&tfs_t2_mode_sync_io_alert), 0x00000010, "All operations are performed synchronous", HFILL}},

	{ &hf_smb_t2_mode_sync_io_nonalert,
		{ "", "smb.t2.mode.sync_io_nonalert", FT_BOOLEAN, 32,
		TFS(&tfs_t2_mode_sync_io_nonalert), 0x00000020, "All operations are synchronous and may block", HFILL}},

	{ &hf_smb_t2_alignment,
		{ "Alignment", "smb.alignment", FT_UINT32, BASE_DEC,
		VALS(alignment_vals), 0, "What alignment do we require for buffers", HFILL }},

	{ &hf_smb_t2_next_entry_offset,
		{ "Next Entry Offset", "smb.next_entry_offset", FT_UINT32, BASE_DEC,
		NULL, 0, "Offset to next entry", HFILL }},

	{ &hf_smb_t2_stream_name_length,
		{ "Stream Name Length", "smb.stream_name_len", FT_UINT32, BASE_DEC,
		NULL, 0, "Length of stream name", HFILL }},

	{ &hf_smb_t2_stream_size,
		{ "Stream Size", "smb.stream_size", FT_UINT32, BASE_DEC,
		NULL, 0, "Size of the stream in number of bytes", HFILL }},

	{ &hf_smb_t2_compressed_file_size,
		{ "Compressed Size", "smb.compressed.file_size", FT_UINT32, BASE_DEC,
		NULL, 0, "Size of the compressed file", HFILL }},

	{ &hf_smb_t2_compressed_format,
		{ "Compression Format", "smb.compressed.format", FT_UINT16, BASE_DEC,
		NULL, 0, "Compression algorithm used", HFILL }},

	{ &hf_smb_t2_compressed_unit_shift,
		{ "Unit Shift", "smb.compressed.unit_shift", FT_UINT8, BASE_DEC,
		NULL, 0, "Size of the stream in number of bytes", HFILL }},

	{ &hf_smb_t2_compressed_chunk_shift,
		{ "Chunk Shift", "smb.compressed.chunk_shift", FT_UINT8, BASE_DEC,
		NULL, 0, "Allocated size of the stream in number of bytes", HFILL }},

	{ &hf_smb_t2_compressed_cluster_shift,
		{ "Cluster Shift", "smb.compressed.cluster_shift", FT_UINT8, BASE_DEC,
		NULL, 0, "Allocated size of the stream in number of bytes", HFILL }},

	{ &hf_smb_dfs_path_consumed,
		{ "Path Consumed", "smb.dfs.path_consumed", FT_UINT16, BASE_DEC,
		NULL, 0, "Number of RequestFilename bytes client", HFILL }},

	{ &hf_smb_dfs_num_referrals,
		{ "Num Referrals", "smb.dfs.num_referrals", FT_UINT16, BASE_DEC,
		NULL, 0, "Number of referrals in this pdu", HFILL }},

	{ &hf_smb_get_dfs_server_hold_storage,
		{ "", "smb.dfs.flags.server_hold_storage", FT_BOOLEAN, 16,
		TFS(&tfs_get_dfs_server_hold_storage), 0x02, "The servers in referrals should hold storage for the file", HFILL }},

	{ &hf_smb_get_dfs_fielding,
		{ "", "smb.dfs.flags.fielding", FT_BOOLEAN, 16,
		TFS(&tfs_get_dfs_fielding), 0x01, "The servers in referrals are capable of fielding", HFILL }},

	{ &hf_smb_dfs_referral_version,
		{ "Version", "smb.dfs.referral.version", FT_UINT16, BASE_DEC,
		NULL, 0, "Version of referral element", HFILL }},

	{ &hf_smb_dfs_referral_size,
		{ "Size", "smb.dfs.referral.size", FT_UINT16, BASE_DEC,
		NULL, 0, "Size of referral element", HFILL }},

	{ &hf_smb_dfs_referral_server_type,
		{ "Server Type", "smb.dfs.referral.server.type", FT_UINT16, BASE_DEC,
		VALS(dfs_referral_server_type_vals), 0, "Type of referral server", HFILL }},

	{ &hf_smb_dfs_referral_flags_strip,
		{ "", "smb.dfs.referral.flags.strip", FT_BOOLEAN, 16,
		TFS(&tfs_dfs_referral_flags_strip), 0x01, "Should we strip off pathconsumed characters before submitting?", HFILL }},

	{ &hf_smb_dfs_referral_node,
		{ "Node", "smb.dfs.referral.node", FT_STRING, BASE_NONE,
		NULL, 0, "Name of entity to visit next", HFILL }},

	{ &hf_smb_dfs_referral_proximity,
		{ "Proximity", "smb.dfs.referral.proximity", FT_UINT32, BASE_DEC,
		NULL, 0, "Hint describing proximity of this server to the client", HFILL }},

	{ &hf_smb_dfs_referral_ttl,
		{ "TTL", "smb.dfs.referral.ttl", FT_UINT32, BASE_DEC,
		NULL, 0, "Number of seconds the client can cache this referral", HFILL }},

	{ &hf_smb_dfs_referral_path,
		{ "Path", "smb.dfs.referral.path", FT_STRING, BASE_NONE,
		NULL, 0, "DFS Path that matched pathconsumed", HFILL }},

	{ &hf_smb_dfs_referral_alt_path,
		{ "Alt Path", "smb.dfs.referral.alt_path", FT_STRING, BASE_NONE,
		NULL, 0, "Alternative(8.3) Path that matched pathconsumed", HFILL }},

	{ &hf_smb_response_to,
		{ "Response to", "smb.response_to", FT_UINT32, BASE_DEC,
		NULL, 0, "This packet is a response to the packet in this frame", HFILL }},

	{ &hf_smb_response_in,
		{ "Response in", "smb.response_in", FT_UINT32, BASE_DEC,
		NULL, 0, "The response to this packet is in this packet", HFILL }},

	{ &hf_smb_ea_error_offset,
		{ "EA Error offset", "smb.ea.error_offset", FT_UINT16, BASE_DEC,
		NULL, 0, "Offset into EA list if EA error", HFILL }},

	{ &hf_smb_ea_length,
		{ "EA Length", "smb.ea.length", FT_UINT32, BASE_DEC,
		NULL, 0, "Total EA length for opened file", HFILL }},

	{ &hf_smb_fs_id,
		{ "FS Id", "smb.fs.id", FT_UINT32, BASE_DEC,
		NULL, 0, "File System ID (NT Server always returns 0)", HFILL }},

	{ &hf_smb_sector_unit,
		{ "Sectors/Unit", "smb.fs.sector_per_unit", FT_UINT32, BASE_DEC,
		NULL, 0, "Sectors per allocation unit", HFILL }},

	{ &hf_smb_fs_units,
		{ "Total Units", "smb.fs.units", FT_UINT32, BASE_DEC,
		NULL, 0, "Total number of units on this filesystem", HFILL }},

	{ &hf_smb_fs_sector,
		{ "Bytes per Sector", "smb.fs.bytes_per_sector", FT_UINT32, BASE_DEC,
		NULL, 0, "Bytes per sector", HFILL }},

	{ &hf_smb_volume_serial_num,
		{ "Volume Serial Number", "smb.volume.serial", FT_UINT32, BASE_HEX,
		NULL, 0, "Volume serial number", HFILL }},

	{ &hf_smb_fs_name_len,
		{ "Label Length", "smb.fs.name.len", FT_UINT32, BASE_DEC,
		NULL, 0, "Length of filesystem name in bytes", HFILL }},

	{ &hf_smb_fs_name,
		{ "FS Name", "smb.fs.name", FT_STRING, BASE_DEC,
		NULL, 0, "Name of filesystem", HFILL }},

	{ &hf_smb_volume_label_len,
		{ "Label Length", "smb.volume.label.len", FT_UINT32, BASE_DEC,
		NULL, 0, "Length of volume label", HFILL }},

	{ &hf_smb_volume_label,
		{ "Label", "smb.volume.label", FT_STRING, BASE_DEC,
		NULL, 0, "Volume label", HFILL }},

	{ &hf_smb_free_alloc_units,
		{ "Free Units", "smb.free_alloc_units", FT_UINT32, BASE_DEC,
		NULL, 0, "Number of free allocation units", HFILL }},

	{ &hf_smb_device_type,
		{ "Device Type", "smb.device.type", FT_UINT32, BASE_HEX,
		VALS(device_type_vals), 0, "Type of device", HFILL }},

	{ &hf_smb_device_char_removable,
		{ "", "smb.device.removable", FT_BOOLEAN, 32,
		TFS(&tfs_device_char_removable), 0x00000001, "Is this a removable device", HFILL }},

	{ &hf_smb_device_char_read_only,
		{ "", "smb.device.read_only", FT_BOOLEAN, 32,
		TFS(&tfs_device_char_read_only), 0x00000002, "Is this a read-only device", HFILL }},

	{ &hf_smb_device_char_floppy,
		{ "", "smb.device.floppy", FT_BOOLEAN, 32,
		TFS(&tfs_device_char_floppy), 0x00000004, "Is this a floppy disk", HFILL }},

	{ &hf_smb_device_char_write_once,
		{ "", "smb.device.write_once", FT_BOOLEAN, 32,
		TFS(&tfs_device_char_write_once), 0x00000008, "Is this a write-once device", HFILL }},

	{ &hf_smb_device_char_remote,
		{ "", "smb.device.remote", FT_BOOLEAN, 32,
		TFS(&tfs_device_char_remote), 0x00000010, "Is this a remote device", HFILL }},

	{ &hf_smb_device_char_mounted,
		{ "", "smb.device.mounted", FT_BOOLEAN, 32,
		TFS(&tfs_device_char_mounted), 0x00000020, "Is this a mounted device", HFILL }},

	{ &hf_smb_device_char_virtual,
		{ "", "smb.device.virtual", FT_BOOLEAN, 32,
		TFS(&tfs_device_char_virtual), 0x00000040, "Is this a virtual device", HFILL }},

	{ &hf_smb_fs_attr_css,
		{ "", "smb.fs.attr.css", FT_BOOLEAN, 32,
		TFS(&tfs_fs_attr_css), 0x00000001, "Does this FS support Case Sensitive Search?", HFILL }},

	{ &hf_smb_fs_attr_cpn,
		{ "", "smb.fs.attr.cpn", FT_BOOLEAN, 32,
		TFS(&tfs_fs_attr_cpn), 0x00000002, "Will this FS Preserve Name Case?", HFILL }},

	{ &hf_smb_fs_attr_pacls,
		{ "", "smb.fs.attr.pacls", FT_BOOLEAN, 32,
		TFS(&tfs_fs_attr_pacls), 0x00000004, "Does this FS support Persistent ACLs?", HFILL }},

	{ &hf_smb_fs_attr_fc,
		{ "", "smb.fs.attr.fc", FT_BOOLEAN, 32,
		TFS(&tfs_fs_attr_fc), 0x00000008, "Does this FS support File Compression?", HFILL }},

	{ &hf_smb_fs_attr_vq,
		{ "", "smb.fs.attr.vq", FT_BOOLEAN, 32,
		TFS(&tfs_fs_attr_vq), 0x00000010, "Does this FS support Volume Quotas?", HFILL }},

	{ &hf_smb_fs_attr_dim,
		{ "", "smb.fs.attr.dim", FT_BOOLEAN, 32,
		TFS(&tfs_fs_attr_dim), 0x00000020, "Is this FS a Mounted Device?", HFILL }},

	{ &hf_smb_fs_attr_vic,
		{ "", "smb.fs.attr.vic", FT_BOOLEAN, 32,
		TFS(&tfs_fs_attr_vic), 0x00008000, "Is this FS Compressed?", HFILL }},

	{ &hf_smb_max_name_len,
		{ "Max name length", "smb.fs.max_name_len", FT_UINT32, BASE_DEC,
		NULL, 0, "Maximum length of each file name component in number of bytes", HFILL }},

	{ &hf_smb_echo_count,
		{ "Echo Count", "smb.echo.count", FT_UINT16, BASE_DEC,
		NULL, 0, "Number of times to echo data back", HFILL }},

	{ &hf_smb_echo_data,
		{ "Echo Data", "smb.echo.data", FT_BYTES, BASE_HEX,
		NULL, 0, "Data for SMB Echo Request/Response", HFILL }},

	{ &hf_smb_echo_seq_num,
		{ "Echo Seq Num", "smb.echo.seq_num", FT_UINT16, BASE_DEC,
		NULL, 0, "Sequence number for this echo response", HFILL }},

	};
	static gint *ett[] = {
		&ett_smb,
		&ett_smb_hdr,
		&ett_smb_fileattributes,
		&ett_smb_file_attributes,
		&ett_smb_capabilities,
		&ett_smb_aflags,
		&ett_smb_dialects,
		&ett_smb_dialect,
		&ett_smb_mode,
		&ett_smb_rawmode,
		&ett_smb_flags,
		&ett_smb_flags2,
		&ett_smb_desiredaccess,
		&ett_smb_search,
		&ett_smb_file,
		&ett_smb_openfunction,
		&ett_smb_filetype,
		&ett_smb_writemode,
		&ett_smb_lock_type,
		&ett_smb_ssetupandxaction,
		&ett_smb_optionsup,
		&ett_smb_command,
		&ett_smb_time_date,
		&ett_smb_connect_flags,
		&ett_smb_setup_action,
		&ett_smb_connect_support_bits,
		&ett_smb_open_flags,
		&ett_smb_open_action,
		&ett_smb_move_flags,
		&ett_smb_transaction_flags,
		&ett_smb_transaction_params,
		&ett_smb_transaction_open_flags,
		&ett_smb_find_first2_flags,
		&ett_smb_transaction_data,
		&ett_smb_64bit_time,
		&ett_smb_get_dfs_flags,
		&ett_smb_dfs_referrals,
		&ett_smb_dfs_referral,
		&ett_smb_dfs_referral_flags,
		&ett_smb_device_characteristics,
		&ett_smb_fs_attributes,
	};

	proto_smb = proto_register_protocol("SMB (Server Message Block Protocol)",
	    "SMB", "smb");

	proto_register_subtree_array(ett, array_length(ett));
	proto_register_field_array(proto_smb, hf, array_length(hf));
	register_init_routine(&smb_info_init);
	
	register_proto_smb_browse();
	register_proto_smb_logon();
	register_proto_smb_mailslot();
	register_proto_smb_pipe();
}
